#include "BinarizerPluginController.hpp"

#include <QFileDialog>
#include <QMessageBox>


#include <DoQuBook/Binarizer.hpp>
#include <DoQuBook/PluginLoader.hpp>

#include <QTimer>


ByteArrayReply::ByteArrayReply(const QByteArray& ba, const char* mimeType, QObject * parent)
  : QNetworkReply(parent)
{
  setOpenMode(QIODevice::ReadOnly | QIODevice::Unbuffered);
  setHeader(QNetworkRequest::ContentTypeHeader, QVariant(mimeType));

  dataBuffer_.setData(ba);
  dataBuffer_.open(QIODevice::ReadOnly);

  QTimer::singleShot(0, this, SLOT(dataReady()));
}

void ByteArrayReply::abort() {
}


void ByteArrayReply::close() {
}


qint64 ByteArrayReply::bytesAvailable() const {
  return dataBuffer_.bytesAvailable();
}


qint64 ByteArrayReply::readData(char *data, qint64 len) {
  return dataBuffer_.read(data, len);
}


void ByteArrayReply::setReadBufferSize(qint64 size) {
  Q_UNUSED(size);
  Q_ASSERT(0); //should never happened
}


ByteArrayReply::~ByteArrayReply() {
}


void ByteArrayReply::dataReady() {

  qint64 totalSize = dataBuffer_.bytesAvailable();

  if (!totalSize) { //not in IconServer or (internet) access denied
    setError(QNetworkReply::ContentAccessDenied, "Access Denied");
    emit error(QNetworkReply::ContentAccessDenied);
    emit finished();
    return;
  }

  emit downloadProgress(0, totalSize);
  emit downloadProgress(totalSize, totalSize);
  emit readyRead();
  emit finished();
}


bool ByteArrayReply::open(OpenMode mode) {
  Q_UNUSED(mode);
  Q_ASSERT(0); //should never happened
  return true;
}


qint64 ByteArrayReply::size() const {
  return bytesAvailable(); //well, should never happended too
}





//////////





BinarizerPluginController::BinarizerPluginController(QWidget *parent) :
  QWebView(parent)
, _accessManager (NULL)
{
  _accessManager = new NAM (this);
  this->page()->setNetworkAccessManager(_accessManager);

  connect(this->page()->mainFrame(), SIGNAL(javaScriptWindowObjectCleared()),
          this, SLOT(populateJavaScriptWindowObject()));

  QFile source(":/view/index");
  source.open(QIODevice::ReadOnly);
  setHtml(source.readAll());
  populateJavaScriptWindowObject();
}


QImage BinarizerPluginController::getImage()
{
  QImage img (getAttr("file","value"));
  return img;
}

QImage BinarizerPluginController::getResult()
{
  return _result;
}

void BinarizerPluginController::openFile()
{

  QString fileName = QFileDialog::getOpenFileName(this,tr("Open Image"), "/Users", tr("Image Files (*.png *.jpg *.bmp)"));

  setAttr("file", "value", fileName);
  setAttr("originalImg", "src", "file://"+fileName);
}

QString timeConversion(int msecs)
{
  QString formattedTime;

  int hours = msecs/(1000*60*60);
  int minutes = (msecs-(hours*1000*60*60))/(1000*60);
  int seconds = (msecs-(minutes*1000*60)-(hours*1000*60*60))/1000;
  int milliseconds = msecs-(seconds*1000)-(minutes*1000*60)-(hours*1000*60*60);

  formattedTime.append(QString("%1").arg(hours, 2, 10, QLatin1Char('0')) + ":" +
                       QString( "%1" ).arg(minutes, 2, 10, QLatin1Char('0')) + ":" +
                       QString( "%1" ).arg(seconds, 2, 10, QLatin1Char('0')) + ":" +
                       QString( "%1" ).arg(milliseconds, 3, 10, QLatin1Char('0')));

  return formattedTime;
}

void BinarizerPluginController::run()
{
  if(_pluginName.isEmpty()) {
    QMessageBox::critical(this, "Error", "Please select an algorithm");
    return;
  }

  QString s = getAttr("file","value");
  while(s.isEmpty()) {
    openFile();
    s = getAttr("file","value");
  }

  DoQuBook::PluginLoader *loader = DoQuBook::PluginLoader::instance();
  BinarizerPtr bin = loader->getPlugin<Binarizer>(_pluginName);

  QImage img = getImage();
  QTime myTimer;
  myTimer.start();
  _result = bin->binarize(img);
  _elapseMmilliSecond = myTimer.elapsed();

  setAttr("result", "src", "convert:"+s+_pluginName);
  setAttr("resultDiv", "style", "");

  int ms = _elapseMmilliSecond;
  QString elapsedString = timeConversion(ms);
  setInner("elapsed",elapsedString);

  emit hasFinished();
}

void BinarizerPluginController::checkForm()
{
  emit formValidated();
}



void BinarizerPluginController::setPlugin(QString pluginName)
{
  _pluginName=pluginName;
  setInner("selectedPlugin", pluginName);
}

void BinarizerPluginController::populateJavaScriptWindowObject()
{
  this->page()->mainFrame()->addToJavaScriptWindowObject("view", this);
  DoQuBook::PluginLoader *loader = DoQuBook::PluginLoader::instance();
  QStringList binarizers = loader->listPlugins<Binarizer>();

  setInner("plugins", "");
  foreach(QString bin, binarizers) {
    QString inner = "<li><a  href='#' onclick='view.setPlugin(\""+bin+"\")'>"+bin+"</a></li>";
    appendInner("plugins",inner);
  }
}

QString BinarizerPluginController::getAttr(const QString &id, const QString &attrName)
{
  QWebFrame *frame = this->page()->mainFrame();
  QWebElement elem = frame->findFirstElement("#"+id);
  return elem.attribute(attrName);
}

void BinarizerPluginController::setAttr(const QString &id, const QString &attrName, const QString &val)
{
  QWebFrame *frame = this->page()->mainFrame();
  QWebElement elem = frame->findFirstElement("#"+id);
  elem.setAttribute(attrName, val);
}

void BinarizerPluginController::setInner(const QString &id, const QString &val)
{
  QWebFrame *frame = this->page()->mainFrame();
  QWebElement elem = frame->findFirstElement("#"+id);
  elem.setInnerXml(val);
}

void BinarizerPluginController::appendInner(const QString &id, const QString &val)
{
  QWebFrame *frame = this->page()->mainFrame();
  QWebElement elem = frame->findFirstElement("#"+id);
  elem.appendInside(val);
}

void NAM::loadIcon(QByteArray &bytes, const QString &path, const char *format)
{
  QImage image = _view->getResult();

  //QImageReader reader(path);
  //reader.read(&image);

  QBuffer buffer(&bytes);
  buffer.open(QIODevice::WriteOnly);
  image.save(&buffer, format);

}
