#ifndef CLASSBUILDER_HPP
#define CLASSBUILDER_HPP
#include <QList>
#include <QFile>
#include <QTextStream>

#include "Entity.hpp"

/**
  * Writes a xml entity to a class in different languages :
  *  cpp \see CppClassBuilder.
  *  java \see JavaClassBuilder.
  * \code
  * <entity name="Personne">
        <field type="int" name="age" id="true" />
        <list type="Page" name="pages" marshallable="true"/>
        <list type="String" name="prenoms" marshallable="false" />
        <field type="int" name="version" />
    </entity>
   \endcode
  *
  */
class ClassBuilder
{
public:
    ClassBuilder();

    virtual QString lang() = 0;

    /// class name = Personne
    void setClassName(const QString &name);
    /// Filed of type <field type="int" name="age"/>
    void addBasicField (const Field &basicField);
    /// Field of type <list type="int" name="age"/>
    void addListField (const Field &listField);

    /// Returns the field type to String. (example list of Page => java = List<Page>, cpp QList<PagePtr>)
    virtual QString getFieldType(const Field&f);
    /// Returns the field annotations for marshalling or hibernate
    virtual QString getFieldAnnotations(const Field&f);

    /// Returns the declaration of a field (public int id;) and its annotations (@Id)
    QString fieldDeclaration(const Field &f);
    QString fieldDeclarationGettersAndSetters(const Field &f);
    QString fieldImplementationGettersAndSetters(const Field &f);

    /// Writes the class to files
    virtual void toFile ();
    /// Writes the class implemantation (.cpp for example)
    virtual void writeImplementationFile();
    /// Writes the class declaration (.hpp for example)
    virtual void writeDeclarationFile();

    virtual bool hasDeclarationAndImplementationFiles() = 0;
    virtual  QFile * declarationFile() = 0;
    virtual QFile * implementationFile() = 0;

    /// Re-implement this function for language specific stuff
    virtual void moreTemplatesOnDeclarationFile(QString & templ);
    /// Re-implement this function for language specific stuff
    virtual void moreTemplatesOnImplementationFile(QString & templ);

//////// Templates Methods
    QString getTemplate();
    virtual QString classTemplate() = 0;

    QString getImplementationFileTemplate();

    QString getAnnotationsTemplate();

    QString getInitArrayListTemplate();
    QString getFieldDeclarationTemplate();
    QString getBasicFieldAnnotation();
    QString getIdFieldAnnotation();
    QString getIntIdFieldAnnotation();
    QString getListFieldAnnotations();

    QString getGetterTemplate();


    QString getGetterDeclarationTemplate();

    QString getGetterImplementationTemplate();

    QString getSetterTemplate();

    QString getSetterDeclarationTemplate();

    QString getSetterImplementationTemplate();

    QString getConstructorTemplate();
    QString getConstructorImplementationTemplate();
    QString getConstructorDeclarationTemplate();

    QString openTemplate(const QString &templateName);

//// Utility functions
    QString classNameUpper();

protected :
    virtual QString imports() = 0;

protected :
    QString _name;
    QList<Field> _fields;
    QList<Field> _listFields;
};

#endif // CLASSBUILDER_HPP
