/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "DBWebBookService.hpp"

#include <QtNetwork>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QNetworkAccessManager>
#include <QBuffer>

#include <DoQuBook/MarshallableFormater.hpp>
#include <DoQuBook/Logger.hpp>
#include <DoQuBook/AttachmentServices.hpp>
#include <DoQuBook/RESTObjectManager.hpp>

#include <DoQuBook/Config.hpp>

#include <DoQuBook/AnnotatedPolygon.hpp>

#include <DBWebImageService.hpp>

#include <MarshallableFormaterFactory.hpp>

#include <assert.h>

using namespace DoQuBook;



DBWebBookService::DBWebBookService (const QString & serverAdress)
    : server_adress (serverAdress)
    , m_marshaller (DBWebRepo::MarshallableFormaterFactory::instance()->create())
    , m_bookManager (WebClient::JSON, m_marshaller)
    , m_bookListManager (WebClient::JSON, m_marshaller)
    , m_pageImageManager (RESTImageManager::JPEG)
    , m_existsImageManager (WebClient::JSON, m_marshaller)
{

    rest_service_url = "DoQuBookWeb/rest";
    book_ressource = "book";
    image_ressource = "image";

    QString postUrl (server_adress + "/" + rest_service_url + "/" + book_ressource);
    m_bookManager.addRessourcePath (WebClient::POST, postUrl);

    QString getUrl (server_adress + "/" + rest_service_url + "/" + book_ressource + "/(id)");
    m_bookManager.addRessourcePath (WebClient::GET, getUrl);

    QString putUrl (server_adress + "/" + rest_service_url + "/" + book_ressource);
    m_bookManager.addRessourcePath (WebClient::PUT, putUrl);

    QString delUrl (server_adress + "/" + rest_service_url + "/" + book_ressource + "/(id)");
    m_bookManager.addRessourcePath (WebClient::DEL, delUrl);

    QString bookListGet (server_adress + "/" + rest_service_url + "/" + book_ressource + "/");
    m_bookListManager.addRessourcePath (WebClient::GET, bookListGet);

    QString getImageUrl (server_adress + "/" + rest_service_url + "/" + image_ressource + "/(id)" + "?width=(width)");
    m_pageImageManager.addRessourcePath (WebClient::GET, getImageUrl);

    QString postImageUrl (server_adress + "/" + rest_service_url + "/" + image_ressource + "/(id)");
    m_pageImageManager.addRessourcePath (WebClient::POST, postImageUrl);

    QString imageExistsUrl (server_adress + "/" + rest_service_url + "/" + image_ressource + "/exists/(id)?width=(width)");
    m_existsImageManager.addRessourcePath (WebClient::GET, imageExistsUrl);

}


QList<int> DBWebBookService::getAllBooks ()
{
    WebClient c;

    if (!c.isAvailable (server_adress) ) {
        throw InternetConnectionException (server_adress);
    }

    QMap<QString, QString> noPrams;
    IntlistPtr b = m_bookListManager.get (noPrams);
    return b->getIds();
}

void DBWebBookService::del (const int & id)
{
    WebClient c;

    if (!c.isAvailable (server_adress) ) {
        throw InternetConnectionException (server_adress);
    }

    QMap<QString, QString> params;
    params["id"] = QString::number (id);

    m_bookManager.del (params);
}

void DBWebBookService::del (BookPtr book)
{
    return del (book->getId () );
}

void DBWebBookService::update (BookPtr book)
{
    WebClient c;

    if (!c.isAvailable (server_adress) ) {
        throw InternetConnectionException (server_adress);
    }

    BookPtr b = m_bookManager.post (book);

    foreach (PagePtr page, book->getPages() ) {
        //Upload Image
        if (!imageExistsOnServer (page) ) {
            uploadImage (page);
        }
        else {
            qDebug() << "Yes image exists !";
        }

    }

    injectImageServiceToPages (book);

    book->merge (b);
}

/*PointPtr DBWebBookService::findPointEquals (AnnotatedPolygonPtr poly, PointPtr point)
{
    bool found = false;
    QListIterator<PointPtr> it (poly->getPoints() );
    PointPtr current;
    while (!found && it.hasNext() ) {
        current = it.next();
        found = current->equals (*point);

    }
    if (found) {
        return current;
    }
    else {
        return PointPtr();
    }
}

AnnotationPtr DBWebBookService::findAnnotationEquals (PagePtr page, AnnotationPtr autre)
{
    bool found = false;
    QListIterator<AnnotationPtr> it (page->getAnnotations() );
    AnnotationPtr current;
    while (!found && it.hasNext() ) {
        current = it.next();
        found = current->equals (*autre);
    }
    if (found) {
        return current;
    }
    else {
        return AnnotationPtr();
    }
}

AnnotatedPolygonPtr DBWebBookService::findPolyEquals (AnnotationLocalPtr ann, AnnotatedPolygonPtr poly)
{
    bool found = false;
    QListIterator<AnnotatedPolygonPtr> it (ann->getAnnotatedPolygons() );
    AnnotatedPolygonPtr current;
    while (!found && it.hasNext() ) {
        current = it.next();
        found = current->equals (*poly);
    }
    if (found) {
        return current;
    }
    else {
        return AnnotatedPolygonPtr();
    }
}*/


BookPtr DBWebBookService::get (const int & id)
{
    WebClient c;

    if (!c.isAvailable (server_adress) ) {
        throw InternetConnectionException (server_adress);
    }

    QMap<QString, QString> params;
    params["id"] = QString::number (id);

    BookPtr b = m_bookManager.get (params);
    injectImageServiceToPages (b);

    return b;
}

bool DoQuBook::DBWebBookService::put (DoQuBook::BookPtr b)
{
    BookPtr pushedBook = m_bookManager.put (b);

    bool uploadedBook = b->getName () == pushedBook->getName ();

    int n = 0;
    BookPtr oldBook = b;


    foreach (PagePtr page, pushedBook->getPages() ) {
        PagePtr oldPage = oldBook->findPageByName (page);
        assert (!oldPage.isNull () );
        page->setImage (oldPage->getImageFilePath () );
        // Upload Image
        //if(!imageExistsOnServer(page)) {
        uploadImage (oldPage);

        emit progress (b->getPages ().size (), n);

        n++;
    }

    injectImageServiceToPages (b);

    b->setId (pushedBook->getId() );
    b->merge (pushedBook);

    return uploadedBook;
}

bool DBWebBookService::imageExistsOnServer (PagePtr page, int w )
{
    return imageExistsOnServer (page->getId (), w);
}

bool  DoQuBook::DBWebBookService::imageExistsOnServer (const  int & id, int w)
{
    QMap<QString, QString> params;
    params["id"] = QString::number (id);
    params["width"] = QString::number (w);

    BooleanPtr imageExists = m_existsImageManager.get (params);

    return imageExists->getValue();
}

#include <iostream>

bool DoQuBook::DBWebBookService::uploadImage (PagePtr page)
{

    QMap<QString, QString> params;
    params["id"] = QString::number (page->getId () );

    QImage image = page->getImage();

    assert (!image.isNull() );

    m_pageImageManager.post (image, "original.jpeg", params);

    return true;
}

QImage DoQuBook::DBWebBookService::getImage (PagePtr page, int w)
{
    return getImage (page->getId(), w);
}

QImage DoQuBook::DBWebBookService::getImage (const int & id, int w)
{
    QMap<QString, QString> params;
    params["id"] = QString::number (id);
    params["width"] = QString::number (w);

    return m_pageImageManager.get (params);
}

bool DoQuBook::DBWebBookService::isAvailable()
{
    WebClient wc;
    QUrl url (server_adress);
    return wc.isAvailable (url);
}


void DoQuBook::DBWebBookService::injectImageServiceToPages (const DoQuBook::BookPtr & book)
{
    foreach (PagePtr page, book->getPages() ) {
        page->installImageService (new DBWebImageService (server_adress) );
    }
}
