#include "DBWebTaskService.hpp"
#include <DoQuBook/Exceptions.hpp>
#include <DoQuBook/StringList.hpp>

#include <MarshallableFormaterFactory.hpp>

using namespace DoQuBook;

DBWebTaskService::DBWebTaskService (const QString & serverAdress)
    : m_marshaller (DoQuBook::DBWebRepo::MarshallableFormaterFactory::instance()->create())
    , m_serverAdress (serverAdress)
    , m_taskManager (WebClient::JSON, m_marshaller)
    , m_asyncTaskManager (WebClient::JSON, m_marshaller)
    , m_taskIdsManager (WebClient::JSON, m_marshaller)
    , m_newTaskIdsManager (WebClient::JSON, m_marshaller)
    , m_tempImageService (serverAdress)
    , m_taskImageService (serverAdress)
    , m_pageService (serverAdress)
    , m_annotationService (serverAdress)
{
    m_restServicesUrl = "DoQuBookWeb/rest/task";

    QString postUrl (m_serverAdress + "/" + m_restServicesUrl);
    m_taskManager.addRessourcePath (WebClient::POST, postUrl);

    QString getUrl (m_serverAdress + "/" + m_restServicesUrl + "/(id)");
    m_taskManager.addRessourcePath (WebClient::GET, getUrl);

    QString getAll (m_serverAdress + "/" + m_restServicesUrl);
    m_taskIdsManager.addRessourcePath (WebClient::GET, getAll);

    QString getNewTasks (m_serverAdress + "/" + m_restServicesUrl + "/new");
    m_newTaskIdsManager.addRessourcePath (WebClient::GET, getNewTasks);

    QString postNewTasks (m_serverAdress + "/" + m_restServicesUrl + "/newForExecutors");
    m_newTaskIdsManager.addRessourcePath (WebClient::POST, postNewTasks);

    QString putUrl (m_serverAdress + "/" + m_restServicesUrl);
    m_taskManager.addRessourcePath (WebClient::PUT, putUrl);

    QString delUrl (m_serverAdress + "/" + m_restServicesUrl + "/(id)");
    m_taskManager.addRessourcePath (WebClient::DEL, delUrl);

    QString getWhenDone (m_serverAdress + "/" + m_restServicesUrl + "/(id)/doneOrFailed");
    m_asyncTaskManager.addRessourcePath (WebClient::GET, getWhenDone);


    RemoteQVariantConverterPtr imageConverter (new RemoteQVariantImageConverter (m_taskImageService) );
    RemoteQVariantConverterPtr imageListConverter (new RemoteQVariantImageListConverter (m_tempImageService) );
    RemoteQVariantConverterPtr histoGramConverter (new RemoteQVariantHistoGramConverter (m_serverAdress) );
    RemoteQVariantConverterPtr numberMapConverter (new RemoteQVariantNumberMapConverter() );
    RemoteQVariantConverterPtr stringConverter (new RemoteQVariantStringConverter() );
    RemoteQVariantConverterPtr numberConverter (new RemoteQVariantNumberConverter() );
    RemoteQVariantConverterPtr booleanConverter (new RemoteQVariantBooleanConverter() );


    // The order of converters is important :
    // First try the number converterd then at last we can use the String converter.
    m_factos << imageConverter  << imageListConverter << histoGramConverter << numberConverter << booleanConverter << numberMapConverter << stringConverter;
}

TaskPtr DBWebTaskService::get (int id)
{
    /// \todo Refactor check internet availability
    WebClient c;

    if (!c.isAvailable (m_serverAdress) ) {
        throw InternetConnectionException (m_serverAdress);
    }

    QMap<QString, QString> params;
    params["id"] = QString::number (id);

    TaskPtr task;
    try {
        task = m_taskManager.get (params);
        try {
            convertTaskParametersToQVariant (task);
        }
        catch (MessageException e) {
            task->setStatus (Task::Fail);
            task->setErrorMessage (e.message() );
            notifyHasFailed (task);
            throw e;
        }
        return task;
    }
    catch (CouldNotParseJSONException e) {
        qCritical()  << "DBWebAnnotationService::get->CouldNotParseJSONException";
        return task;
    }
    catch (MessageException e) {
        task->setErrorMessage (e.message() );
        notifyHasFailed (task);
        throw e;
    }
    return task;
}

TaskPtr DBWebTaskService::getTaskWhenDone (int id)
{
    /// \todo Refactor check internet availability
    WebClient c;

    if (!c.isAvailable (m_serverAdress) ) {
        throw InternetConnectionException (m_serverAdress);
    }

    QMap<QString, QString> params;
    params["id"] = QString::number (id);

    try {
        TaskPtr task = m_asyncTaskManager.get (params);
        if (task->getStatus() != Task::Fail) {
            convertResultToQVariant (task);
        }
        else {
            throw MessageException ("Task has failed " + task->getErrorMessage() );
        }
        return task;
    }
    catch (CouldNotParseJSONException e) {
        qCritical()  << "DBWebAnnotationService::get->CouldNotParseJSONException";
        return TaskPtr();
    }
}

void DBWebTaskService::convertResultToQVariant (TaskPtr t)
{
    QVariant resultUrl = t->getResult();
    QString type = t->getResourceType();
    QListIterator<RemoteQVariantConverterPtr> factosIt (m_factos);
    bool found = false;
    QVariant outVariant;
    while (!found && factosIt.hasNext() ) {
        RemoteQVariantConverterPtr facto = factosIt.next();
        if (facto->accept (type) ) {
            found = true;
            outVariant = facto->toVariant (resultUrl.toString(), t);
        }
    }
    if (!found ) {
        throw MessageException ("Unsuported resource result for task " + QString::number(t->getId()) );
    }

    t->setPureResult (outVariant);
}

void DBWebTaskService::convertResultQVariantToString (TaskPtr & t)
{
    QVariant result = t->getResult();
    QString typeName = "Unknown";
    if (!result.isNull() && result != "" ) {
        QListIterator<RemoteQVariantConverterPtr> it (m_factos);
        bool found = false;
        QString resultAsString;
        while (!found && it.hasNext() ) {
            RemoteQVariantConverterPtr facto = it.next();
            if (facto->accept (result) ) {
                found = true;
                resultAsString = facto->toString (result, t);
                typeName = facto->typeName();
            }
        }
        if (resultAsString.size() <= 0) {
            typeName = "String";
            resultAsString = result.toString();
            qWarning() << "Task " + QString::number (t->getId() ) + " : unsupported result type " + result.typeName() + " res :" + result.toString();
        }

        t->setPureResult (resultAsString);
        t->setResourceType (typeName);
    }
}

void DBWebTaskService::convertTaskParametersToQVariant (TaskPtr task)
{
    foreach (TaskParameterPtr param, task->getParams() ) {
        QVariant v = convertTaskParameterToQVariant (param, task);
        param->setValue (v);
    }
}

void DBWebTaskService::convertTaskParametersToString (TaskPtr & t)
{
    QList<TaskParameterPtr> params = t->getParams();
    foreach (TaskParameterPtr param, params) {
        QString typeName;
        QVariant value = param->getValue();
        if (value.type() != QVariant::String) {
            QListIterator<RemoteQVariantConverterPtr> it (m_factos);
            bool found = false;
            QString valueAsString;
            while (!found && it.hasNext() ) {
                RemoteQVariantConverterPtr facto = it.next();
                qDebug() << value.type();
                if (facto->accept (value) ) {
                    found = true;
                    valueAsString = facto->toString (value, t);
                    typeName = facto->typeName();
                    qDebug() << facto->typeName();
                }
            }
            if (valueAsString.size() <= 0) {
                throw MessageException ("Parameter " + param->getName() + " : unsupported result type " + param->getParameterType() );
            }
            param->setValue (valueAsString);
            param->setParameterType (typeName);
        }
    }
}


QVariant DBWebTaskService::convertTaskParameterToQVariant (TaskParameterPtr taskParam, TaskPtr task)
{
    if (taskParam.isNull() ) {
        throw ParamParsingException (taskParam->getName() );
    }

    QString type = taskParam->getParameterType();
    QVariant inputVariant = taskParam->getValue();

    QListIterator<RemoteQVariantConverterPtr> it (m_factos);
    bool found = false;
    QVariant outVariant;
    while (!found && it.hasNext() ) {
        RemoteQVariantConverterPtr facto = it.next();
        if (facto->accept (type) ) {
            found = true;
            outVariant = facto->toVariant (inputVariant.toString(), task);
        }
    }

    if (!found) {
        outVariant = inputVariant;
#if DEBUG_LOG
        qWarning() << "Could not find a proper variant conversion method. Using default";
#endif
    }
    if (outVariant.isNull() ) {
        throw MessageException ("QVariant converted taskparameter is null");
    }

    return outVariant;
}

#include <DoQuBook/AnnotationGlobal.hpp>

void DBWebTaskService::add (TaskPtr & task)
{
    task->setStatus (Task::Building);
    QList<TaskParameterPtr> oldParams = task->getParams();
    QList<TaskParameterPtr> emptyParams;
    task->setParams (emptyParams);
    task = m_taskManager.post (task);

    task->setParams (oldParams);

    replaceClassByURLs (task);
    QList<TaskParameterPtr> params = task->getParams();
    foreach (TaskParameterPtr param, params) {
        if (param->getValue().canConvert<AnnotationGlobalPtr>() ) {
          AnnotationGlobalPtr annGlobal =  param->getValue().value<AnnotationGlobalPtr>();
            param->setValue (annGlobal->getName() );
        }
    }
    task->setStatus (Task::New);
    task = m_taskManager.post (task);
    task = get (task->getId() );
}

void DBWebTaskService::notifyAccepted (TaskPtr & task)
{
    WebClient c;
    c.setContentType (WebClient::JSON);
    QString url (m_serverAdress + "/" + m_restServicesUrl + "/" + QString::number (task->getId() ) + "/setAccepted");

    QByteArray response = c.sendSynchronous (url, WebClient::PUT, "");
    if (response == "OK") {
        task->setStatus (Task::Accepted);
    }
}

void DBWebTaskService::notifyHasFailed (TaskPtr & task)
{
    WebClient c;
    c.setContentType (WebClient::JSON);
    QString url (m_serverAdress + "/" + m_restServicesUrl + "/" + QString::number (task->getId() ) + "/hasFailed");

    qDebug() << "notifyHasFailed " << url;
    QByteArray content (task->getErrorMessage().toStdString().c_str() );
    qDebug() << "content " << content;
    QByteArray response = c.sendSynchronous (url, WebClient::PUT, content);
    qDebug() << "response " << response;

    if (response == "OK") {
        task->setStatus (Task::Fail);
    }
}

void DBWebTaskService::setFinalResult (TaskPtr & task)
{
    convertResultQVariantToString (task);
    WebClient c;
    c.setContentType (WebClient::JSON);

    QString url (m_serverAdress + "/" + m_restServicesUrl + "/" + QString::number (task->getId() ) + "/setFinalResult");
    qDebug() << "PUT AT " << url;

    QByteArray content (task->getResult().toString().toStdString().c_str() );
    QByteArray response = c.sendSynchronous (url, WebClient::PUT, content);
    qDebug() << "RESPONSE is " << response;
    if (response == "OK") {
        task->setStatus (Task::Done);
    }
}

void DBWebTaskService::refreshForm(TaskPtr &task)
{

    QMap<QString, QString> params;
    params["id"] = QString::number (task->getId());

    TaskPtr existing = m_taskManager.get (params);
    existing->setStatus(Task::RefreshForm);
    QList<TaskParameterPtr> oldParams = existing->getParams();
    foreach(TaskParameterPtr param, oldParams) {
        if(!param->getValue().toString().startsWith("http")) {
            QString paramName = param->getName();
            QString newVal = task->getParam(paramName)->getValue().toString();
            qDebug() << "newVal " << newVal;
            QString oldVal = param->getValue().toString();
            qDebug() << "oldVal " << oldVal;
            if(newVal != oldVal) {
                param->setValue(newVal);
            }
        }
    }

    task = m_taskManager.post(existing);
}


bool DBWebTaskService::del (TaskPtr & task)
{
    WebClient c;

    if (!c.isAvailable (m_serverAdress) ) {
        throw InternetConnectionException (m_serverAdress);
    }

    QMap<QString, QString> params;
    params["id"] = QString::number (task->getId() );

    m_taskManager.del (params);
    /// \todo Parse the boolean response;
    return true;
}

QList<int> DBWebTaskService::getAll()
{
    WebClient c;

    if (!c.isAvailable (m_serverAdress) ) {
        throw InternetConnectionException (m_serverAdress);
    }

    QMap<QString, QString> params;
    IntlistPtr ids = m_taskIdsManager.get (params);
    if (!ids.isNull() ) {
        return ids->getIds();
    }
    else {
        QList<int> ids;
        return ids;
    }
}

QList<int> DBWebTaskService::getNewTasks()
{
    WebClient c;

    if (!c.isAvailable (m_serverAdress) ) {
        throw InternetConnectionException (m_serverAdress);
    }

    QMap<QString, QString> params;
    IntlistPtr ids = m_newTaskIdsManager.get (params);
    if (!ids.isNull() ) {
        return ids->getIds();
    }
    else {
        QList<int> ids;
        return ids;
    }
}


QList<int> DBWebTaskService::getNewTasksForExecutors (QStringList & executors)
{
    QUrl url = m_serverAdress + "/" + m_restServicesUrl + "/newForExecutors";
#ifdef DEBUG_LOG
    qDebug() << "POST : " << url;
#endif
    WebClient wc;
    wc.setContentType (WebClient::JSON);

    StringListPtr executorsPtr = StringList::create();
    executorsPtr->setStrings (executors);

    MarshallableFormaterPtr formater = DBWebRepo::MarshallableFormaterFactory::instance()->create();
    QString s;
    formater->marshall (executorsPtr, &s);

    QByteArray jsonBytes (s.trimmed().toStdString().c_str() );
    QByteArray array = wc.sendSynchronous (url, WebClient::POST, jsonBytes);
#ifdef DEBUG_LOG
    qDebug() << "Content " << jsonBytes;
    qDebug() << "Response " << array;
#endif

    IntlistPtr ids = Intlist::create();
    formater->unMarshall (array, ids);

    return ids->getIds();
}


void DBWebTaskService::runExecutor (const QString & name, PagePtr & page, const QVariantMap & paramsMap)
{
    TaskPtr t = Task::create (name);

    QMapIterator<QString, QVariant> it (paramsMap);
    while (it.hasNext() ) {
        it.next();
        t->addParam (it.key(), it.value() );
    }

    add (t);

    TaskPtr onServerUpdatedTask = getTaskWhenDone (t->getId() );

    // The task is done : we delete it
    //del (t);
    if (onServerUpdatedTask->getStatus() == Task::Fail) {
        qCritical() << onServerUpdatedTask->getErrorMessage();
        throw MessageException (onServerUpdatedTask->getErrorMessage() );
    }
    QVariant resultUrl = onServerUpdatedTask->getResult();
    if (onServerUpdatedTask->getResourceType() == "Page") {
        page = m_pageService.get (resultUrl.toInt() );
    }
    else {
        qDebug() << "ResourceType : " <<  onServerUpdatedTask->getResourceType();
        qDebug() << onServerUpdatedTask->getResult();
        qCritical() << "Unsuported resource result for task " << onServerUpdatedTask->getId();
    }
}

void DBWebTaskService::replaceClassByURLs (TaskPtr & t)
{
    convertTaskParametersToString (t);
    convertResultQVariantToString (t);
}


/*
 *if(param->getValue().type() == QVariant::Image) {
            QUrl urlP (m_tempImageService.add(param->getValue().value<QImage>()));
            param->setValue(urlP);
        } else if(param->getParameterType() == "Page") {
            PagePtr p = param->getValue().value<DoQuBook::PagePtr>();
            m_pageService.add(p);
            /// The value is the id
            param->setValue(p->getId());

        } else if(param->getParameterType() == "AnnotationGlobal") {
            qDebug() << "Annotation !!!";
            AnnotationGlobalPtr p = param->getValue().value<DoQuBook::AnnotationGlobalPtr>();
            assert(!p.isNull());
            AnnotationPtr ann = p;
            assert(!ann.isNull());
            m_annotationService.add(ann);
            assert(!ann.isNull());
            qDebug() << "Annotation !!! " << ann->getId();
            /// The value is the id
            param->setValue(ann->getId());

        } if(! param->getValue().canConvert(QVariant::String)){
            throw MessageException ("Task "+QString::number(t->getId())+ " : unsupported param type for " + param->getName() );
        } */


