#include "TestAnnotationService.hpp"

#include <QtTest>

#include "Utils.hpp"

#include <DoQuBook/AnnotationService.hpp>
#include <DoQuBook/DBWebAnnotationService.hpp>
#include <DoQuBook/DBWebPageService.hpp>
#include <QList>

using namespace DoQuBook;

class RESTEntityDescription;
typedef QSharedPointer<RESTEntityDescription> RESTEntityDescriptionPtr;

class RESTEntityDescription : public SimpleMarshallable {
    Q_OBJECT

    Q_PROPERTY(QList<QString> attributes READ getAttributes WRITE setAttributes)
    QList<QString> m_attributes;

  public :
    MARSHALLABLE_CREATE_FONCTIONS (RESTEntityDescription)
    MARSHALLABLE_QTLIST (QList, QString, attributes)
    MARSHALLABLE_ACTIV_1LIST (attributes, m_attributes)

    RESTEntityDescription(){}

    QList<QString> getAttributes() const
    {
      return m_attributes;
    }
  public slots:
    void setAttributes(QList<QString> arg)
    {
      m_attributes = arg;
    }
};

template <class T,class IdType>
class GenericTest {
  public :

    GenericTest(DBWebCRUDService<T, IdType > &srv)
      : m_service (srv) {}

    void testGetAddAndDel (QSharedPointer<T> &e){
      {
        QSharedPointer<T>  entityOnServer = m_service.get (e->getId());
        QVERIFY (entityOnServer.isNull());
      }
      // add the entity
      m_service.add (e);
      {
        QSharedPointer<T> entityOnServer = m_service.get (e->getId());
        QVERIFY (!entityOnServer.isNull());
        QVERIFY (entityOnServer->equals(*e));
      }
      bool deleted = m_service.del(e->getId());
      QVERIFY(deleted);
      // Check if really deleted
      {
        QSharedPointer<T> entityOnServer = m_service.get (e->getId());
        QVERIFY(entityOnServer.isNull());
      }
    }

    void testFields() {
      RESTObjectManager<RESTEntityDescription> manager (WebClient::JSON);

      T a;
      const QMetaObject * o = a.metaObject();

      QString realClassName  = QString (o->className() ).toLower();
      QStringList namespaces = realClassName.split ("::");
      QString className = namespaces.at (namespaces.length()-1);

      QString getUrl ("http://localhost:8080/DoQuBookWeb/rest/"+className+"/description/entity");
      manager.addRessourcePath(WebClient::GET, getUrl);

      QMap<QString, QString> params;
      QSharedPointer<RESTEntityDescription> entityDescription = manager.get(params);
      qDebug() << entityDescription->getAttributes();


      QList<QString> attrs = entityDescription->getAttributes();
      foreach(QString attr, attrs) {
        bool found = false;
        for(int i = 0; i < o->propertyCount() && !found; i++) {
          found = attr == o->property(i).name();
        }
        QVERIFY(found);
      }
      for(int i = 0; i < o->propertyCount(); i++) {
        bool found = false;
        QString prop = o->property(i).name();
        if(prop != "objectName" && prop != "type") {
          for(int j = 0; j < attrs.size() && !found; j++) {
            found = attrs[j] == prop;
          }
          if(!found) qCritical() << prop << " not found";
          QVERIFY(found);
        }
      }
    }


  private :
    //void modify(QObject *o)

  private :
    DBWebCRUDService<T, IdType> &m_service;
};




void TestAnnotationService::testAddGetDelAnnotationLocal()
{
  QVERIFY (!service.isNull());
  AnnotationPtr annotationLocal = createAnnotationLocal();
  // Verify that no annotation of annLocal name exists on server
  {
    AnnotationPtr annotationOnServer = service->get (annotationLocal->getId());
    QVERIFY (annotationOnServer.isNull());
  }
  // add the annotation
  service->add (annotationLocal);
  // Verify that annotationLocal has be added
  {
    AnnotationPtr annotationOnServer = service->get (annotationLocal->getId());
    QVERIFY (!annotationOnServer.isNull());
    QVERIFY (annotationOnServer->metaObject()->className() == QString("DoQuBook::AnnotationLocal"));
    QVERIFY (annotationOnServer->equals(*annotationLocal));
  }

  // Second add ! (Shoudl receive an exception)
  try {
    service->add (annotationLocal);
    QVERIFY(false);
  } catch (CouldNotParseJSONException e) {
    QVERIFY(true);
  } catch (NetworkConnectionException e) {
    QVERIFY(true);
  }

  // clean up
  service->del (annotationLocal);
  // Verify that the annotation has been deleted
  {
    AnnotationPtr annotationOnServer = service->get (annotationLocal->getId());
    QVERIFY (annotationOnServer.isNull());
  }
}


void TestAnnotationService::testSupportedTypesOfValue()
{
  AnnotationPtr ann = AnnotationGlobal::create();
  ann->setName("annGlobal");

  // test a boolean
  {
    AnnotationGlobalPtr annGlobal = ann.dynamicCast<AnnotationGlobal>();
    annGlobal->setValue(true);
    service->add(ann);
    AnnotationGlobalPtr annotationOnServer = service->getAnnotationGlobal (ann->getId());
    QVERIFY (!annotationOnServer.isNull());
    QVERIFY (annotationOnServer->getValue().toBool());
    service->del(annotationOnServer);
  }
  ann->setId(-1);
  // test a double
  {
    AnnotationGlobalPtr annGlobal = ann.dynamicCast<AnnotationGlobal>();
    annGlobal->setValue(2.9);
    service->add(ann);
    AnnotationGlobalPtr annotationOnServer = service->getAnnotationGlobal (ann->getId());
    QVERIFY (!annotationOnServer.isNull());
    QVERIFY (annotationOnServer->getValue().toDouble() == 2.9);
    service->del(annotationOnServer);
  }
  ann->setId(-1);
  // test a QString
  {
    AnnotationGlobalPtr annGlobal = ann.dynamicCast<AnnotationGlobal>();
    annGlobal->setValue("IMPORTANT");
    service->add(ann);
    AnnotationGlobalPtr annotationOnServer = service->getAnnotationGlobal (ann->getId());
    QVERIFY (!annotationOnServer.isNull());
    QVERIFY (annotationOnServer->getValue().toString() == QString("IMPORTANT"));
    service->del(annotationOnServer);
  }
  ann->setId(-1);
  // test a QList
  {
    AnnotationGlobalPtr annGlobal = ann.dynamicCast<AnnotationGlobal>();

    //QString s ("2.9;2.6");
    QList<QVariant> measures;
    measures.append(2.9); measures.append(2.6);

    QList<QVariant> measures2;
    measures2.append(2.6);

    annGlobal->setValue(measures);
    service->add(ann);
    AnnotationGlobalPtr annotationOnServer = service->getAnnotationGlobal (ann->getId());
    QVERIFY (!annotationOnServer.isNull());
    qDebug() << annotationOnServer->getValue().toStringList();
    QVERIFY (annotationOnServer->getValueAsList() == measures);
    QVERIFY (annotationOnServer->getValueAsList() != measures2);
    service->del(annotationOnServer);
  }
  ann->setId(-1);
  // test a QDate
  {
    AnnotationGlobalPtr annGlobal = ann.dynamicCast<AnnotationGlobal>();
    QDate p(1982, 7, 12);
    annGlobal->setValue(p);
    MarshallableFormaterJSon<AnnotationGlobal> jsonFormater;
    QString dest;
    jsonFormater.marshall(annGlobal, &dest);
    service->add(ann);
    AnnotationGlobalPtr annotationOnServer = service->getAnnotationGlobal (ann->getId());
    QVERIFY (!annotationOnServer.isNull());
    QVERIFY (annotationOnServer->getValue().toDate()==p);
    service->del(annotationOnServer);
  }
  ann->setId(-1);
  // test a QMap
  {

    AnnotationGlobalPtr annGlobal = ann.dynamicCast<AnnotationGlobal>();

    QMap<QString, QVariant> measures;
    measures["m0"] = 2.9; measures["m1"] = 2.6;

    QMap<QString, QVariant> measures2;
    measures2["m1"] = 2.6;

    QMap<QString, QVariant> measures3;
    measures3["m8"] = 2.9; measures3["m1"] = 2.6;

    QMap<QString, QVariant> measures4;
    measures4["m0"] = 2.97; measures4["m1"] = 2.6;

    QMap<QString, QVariant> measures5;
    measures5["m0"] = 2.90; measures5["m1"] = 2.6;

    annGlobal->setValue(measures);
    qDebug() << annGlobal->getValue();
    service->add(ann);
    AnnotationGlobalPtr annotationOnServer = service->getAnnotationGlobal (ann->getId());
    QVERIFY (!annotationOnServer.isNull());
    qDebug() << annotationOnServer->getValue().toStringList();
    QVERIFY (annotationOnServer->getValueAsMap() == measures);
    QVERIFY (annotationOnServer->getValueAsMap() != measures2);
    QVERIFY (annotationOnServer->getValueAsMap() != measures3);
    QVERIFY (annotationOnServer->getValueAsMap() != measures4);
    QVERIFY (annotationOnServer->getValueAsMap() == measures5);
    service->del(annotationOnServer);
  }
  ann->setId(-1);
}

void TestAnnotationService::testGetAllNames()
{
  DBWebPageService pageSrv ("http://localhost:8080");
  PagePtr page = pageSrv.get(2);
  QVERIFY(!page.isNull());

  /// Create a global annotation
  DoQuBook::AnnotationGlobalPtr annGlobal = DoQuBook::AnnotationGlobal::create();
  annGlobal->setName("annGlobal-testGetAllNames");
  annGlobal->setValue(10);

  /// Add it
  annGlobal = service->addGlobalAnnotationToPage(page, annGlobal);

  /// Create a local annotation
  DoQuBook::AnnotationLocalPtr annLocal = createAnnotationLocal();
  annLocal->setName("annLocal-testGetAllNames");
  annLocal = service->addLocalAnnotationToPage(page, annLocal);


  /// Create a global annotation with same name to check if doublons
  DoQuBook::AnnotationGlobalPtr annGlobal2 = DoQuBook::AnnotationGlobal::create();
  annGlobal2->setName("annGlobal-testGetAllNames");
  annGlobal2->setValue(12);
  // And add it to an other page
  PagePtr page2 = pageSrv.get(3);
  annGlobal = service->addGlobalAnnotationToPage(page2, annGlobal2);

  QStringList names = service->getAllAnnotationNames();
  QVERIFY(names.size() == 2);
  QVERIFY(names.contains(annGlobal->getName()));
  QVERIFY(names.contains(annLocal->getName()));
  QVERIFY(names.contains(annGlobal2->getName()));

  cleanUpAllAddedAnnotationOfPage(page);
  cleanUpAllAddedAnnotationOfPage(page2);
}

void TestAnnotationService::testGetPagesByAnnotationName()
{

  DBWebPageService pageSrv ("http://localhost:8080");
  PagePtr page = pageSrv.get(2);
  QVERIFY(!page.isNull());

  /// Create a global annotation
  DoQuBook::AnnotationGlobalPtr annGlobal = DoQuBook::AnnotationGlobal::create();
  annGlobal->setName("annGlobal-testGetPagesByAnnotationName");
  annGlobal->setValue(10);

  /// Add it
  annGlobal = service->addGlobalAnnotationToPage(page, annGlobal);

  /// Create a local annotation
  DoQuBook::AnnotationLocalPtr annLocal = createAnnotationLocal();
  annLocal->setName("annLocal-testGetPagesByAnnotationName");
  annLocal = service->addLocalAnnotationToPage(page, annLocal);


  /// Create a global annotation with same name to check if doublons
  DoQuBook::AnnotationGlobalPtr annGlobal2 = DoQuBook::AnnotationGlobal::create();
  annGlobal2->setName("annGlobal-testGetPagesByAnnotationName");
  annGlobal2->setValue(12);
  // And add it to an other page
  PagePtr page2 = pageSrv.get(3);
  annGlobal2 = service->addGlobalAnnotationToPage(page2, annGlobal2);


  PageIteratorPtr it = service->getPagesWithAnnotationName(annGlobal->getName());
  QList<int> ids;
  while(it->hasNext()) {
    PagePtr p = it->next();
    ids << p->getId();
  }
  QVERIFY(ids.size()==2);
  QVERIFY(ids.contains(page->getId()));
  QVERIFY(ids.contains(page2->getId()));

  PageIteratorPtr itAnnLocal = service->getPagesWithAnnotationName(annLocal->getName());
  QList<int> idsAnnLocal;
  while(itAnnLocal->hasNext()) {
    PagePtr p = itAnnLocal->next();
    idsAnnLocal << p->getId();

  }
  QVERIFY(idsAnnLocal.size()==1);
  QVERIFY(idsAnnLocal.contains(page->getId()));

  cleanUpAllAddedAnnotationOfPage(page);
  cleanUpAllAddedAnnotationOfPage(page2);
}

void TestAnnotationService::testGetAnnotationByPageAndByName()
{
  DBWebPageService pageSrv ("http://localhost:8080");
  PagePtr page = pageSrv.get(2);
  QVERIFY(!page.isNull());

  /// Create a global annotation
  DoQuBook::AnnotationGlobalPtr annGlobal = DoQuBook::AnnotationGlobal::create();
  annGlobal->setName("annGlobal-testGetPagesByAnnotationName");
  annGlobal->setValue(10);

  /// Add it
  annGlobal = service->addGlobalAnnotationToPage(page, annGlobal);

  AnnotationGlobalPtr annOnServerGlobal = service->getAnnotationGlobalOfPageByName(page, annGlobal->getName());
  QVERIFY(annOnServerGlobal->getId() == annGlobal->getId());

  cleanUpAllAddedAnnotationOfPage(page);

  /// Test if annotation does not exists
  AnnotationGlobalPtr annGlobalNotExists = service->getAnnotationGlobalOfPageByName(page, "xxxxx");
  QVERIFY(annGlobalNotExists.isNull());
}

void TestAnnotationService::init()
{
  AnnotationPtr annotationLocal = createAnnotationLocal();
  AnnotationPtr annotationOnServer = service->get (annotationLocal->getId());
  if(!annotationOnServer.isNull()) {
    service->del (annotationOnServer);
  }
}

void TestAnnotationService::cleanup()
{
}


void TestAnnotationService::qDebugAnnotation(const AnnotationPtr& ann)
{
    qDebug() << "--------------------------";
    qDebug() << ann->getId() << " " << ann->getName();
    AnnotationLocalPtr annLocal = ann.dynamicCast<AnnotationLocal>();

    if(!annLocal.isNull()) {
      qDebug() << "\t Type : " << "AnnotationLocal";
    }
    AnnotationGlobalPtr annGlobal = ann.dynamicCast<AnnotationGlobal>();
    if(!annGlobal.isNull()) {
      qDebug() << "\t Type : " << "AnnotationGlobal";
    }
}

void TestAnnotationService::cleanUpAllAddedAnnotationOfPage(const PagePtr& page)
{
  /// clean up all added annotations :
  AnnotationIteratorPtr it2 = service->getAnnotationOfPage(page);
  while(it2->hasNext()) {
    AnnotationPtr ann = it2->next();
    service->del(ann);
  }
}

void TestAnnotationService::testSimpleAddToPage()
{
  /// Create the needed services
  DBWebPageService pageSrv ("http://localhost:8080");
  PagePtr page = pageSrv.get(2);
  QVERIFY(!page.isNull());

  /// Create a global annotation
  DoQuBook::AnnotationGlobalPtr annGlobal = DoQuBook::AnnotationGlobal::create();
  annGlobal->setName("annGlobal-testAdd");
  annGlobal->setValue(10);

  /// Add it
  annGlobal = service->addGlobalAnnotationToPage(page, annGlobal);
  qDebug() << annGlobal->getId();

  /// Create a local annotation
  DoQuBook::AnnotationLocalPtr annLocal = createAnnotationLocal();
  annLocal = service->addLocalAnnotationToPage(page, annLocal);
  qDebug() << annLocal->getId();

  ///List all annotations of Page page
  AnnotationIteratorPtr it = service->getAnnotationOfPage(page);
  while(it->hasNext()) {
    AnnotationPtr ann = it->next();
    qDebugAnnotation(ann);
  }

  cleanUpAllAddedAnnotationOfPage(page);

}

void TestAnnotationService::testAttributes()
{
  {
    DBWebAnnotationService srv ("http://localhost:8080");
    GenericTest<Annotation, int> t (srv);
    t.testFields();
    AnnotationPtr ptr = createAnnotationGlobal();
    t.testGetAddAndDel(ptr);
  }

  {
    DBWebPageService srv ("http://localhost:8080");
    GenericTest<Page, int> t (srv);
    t.testFields();
    PagePtr ptr = createPage();
    t.testGetAddAndDel(ptr);
  }
}

QTEST_MAIN (TestAnnotationService)

