/*!
 * \file AnnotatedPolygon.hpp
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#ifndef ANNOTATEDPOLYGON_HPP
#define ANNOTATEDPOLYGON_HPP

#include <DoQuBook/DoQuBook_global.hpp>
#include <QPolygon>
#include <QObject>
#include <DoQuBook/Marshallable.hpp>
#include <DoQuBook/SimpleMarshallable.hpp>
#include <QSharedPointer>
#include <DoQuBook/Point.hpp>

namespace DoQuBook
{

  class AnnotatedPolygon;
  typedef QSharedPointer<AnnotatedPolygon> AnnotatedPolygonPtr;

  /*! \class AnnotatedPolygon
         * \brief A polygon with an QVariant value.
         * \author Maxime Daymard
         *
         *  The class provides access to the points of polygon and its value
         *  This class is mainly used by AnnnotationLocal.
         *
         *  \code
         *    PointPtr p1 = Point::create(100.0, 20,0);
         *    AnnotatedPolygonPtr annPoly = AnnotatedPolygon::create();
         *    annPoly->addPoint(p1);
         * //
         *    annPoly->setValue("super");
         * // Or
         *    annPoly->setValue(true);
         * //Or
         *    annPoly->setValue(1.0);
         *  \endcode
         *
         *  \attention - QVariant can be of type (bool, int, float, string).
         *
         *  \attention - AnnotatedPolygon is Marshallable with an id.
         *          An annotatedPolygon can equal an other in two different way :
         *              - a.id == b.id
         *              - a == b (a.value == b.value && a.points = b.points)
         *              - a.equals(b) <=> a==b
         *
         */
  class DoQuBookSHARED_EXPORT AnnotatedPolygon : public SimpleMarshallable
  {
      Q_OBJECT

      /*!
              * \brief id set by the server or a book provider.
              *  - -1 = not saved yet.
              *  - else = the real id of the AnnotatedPolygon
              */
      Q_PROPERTY (int id READ getId WRITE setId NOTIFY idChanged)
      /*!
              * \brief The value of the annotation at the location defined by the points.
              *
              * This property is a QVariant but is saved on DoQuBookWeb with a String :
              * The convertion from the string to the QVariant and vice-versa is tested with :
              *      - bool,
              *      - int,
              *      - float,
              *      - string.
              */
      Q_PROPERTY (QVariant value READ getValue WRITE setValue)
      /// \brief Location of the AnnotatedPolygon.
      Q_PROPERTY (QList<PointPtr> points READ getPoints WRITE setPoints)

    public:

      MARSHALLABLE_CREATE_FONCTIONS (AnnotatedPolygon)

      MARSHALLABLE_MLIST (QList, Point, points)
      MARSHALLABLE_ACTIV_1LIST (points, _points)

      QList<PointPtr> getPoints() const;

      QPolygonF toQPolygonF() const;

      QVariant getValue() const;

      QString valueQString() const;

      int valueInt() const;

      template<typename T>
      T value() const {
        return _value.value<T> ();
      }

      void setValue (QVariant value);

      PointPtr getPoint (int i) const;

      void setPoints (QList<PointPtr> p);

      void addPoint (PointPtr p);

      bool operator== (AnnotatedPolygon const & b);

      /*! An annotatedPolygon can equal an other in two different way :
             *              - a.id == b.id
             *              - a == b (a.value == b.value && a.points = b.points)
             *              - a.equals(b) <=> a==b
             */
      bool equals (const AnnotatedPolygon & autre);

      bool pointsMatches (const AnnotatedPolygon & autre);

      int getId() const;

      //bool merge (MarshallablePtr marsh);

      PointPtr findPoint (PointPtr p);

      virtual void merge (MarshallablePtr other);

    public slots:
      void setId (int arg);

    signals:
      void idChanged (int arg);

    public :
      AnnotatedPolygon();

    private:
      /// See corresponding property
      QList<PointPtr> _points;
      /// See corresponding property
      QVariant _value;
      /// See corresponding property
      int m_id;
  };

}

#endif // ANNOTATEDPOLYGON_HPP
