/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#ifndef ANNOTATION_HPP
#define ANNOTATION_HPP

#include <QString>
#include <DoQuBook/DoQuBook_global.hpp>
#include <DoQuBook/AnnotationVisitor.hpp>
#include <QSharedPointer>

#include <QVariant>
#include <DoQuBook/SimpleMarshallable.hpp>
#include <DoQuBook/Singleton.hpp>

namespace DoQuBook
{

  class Annotation;
  typedef QSharedPointer<Annotation> AnnotationPtr;

  /*!
   * \brief Ann Annotation refers to an information concerning a Page.
   * For now, two kinds Annotations exists :
   *  - AnnotationGlobal : a name and a value.
   *  - AnnotationLocal : a name and several annotated polygons; each polygon has its own value.
   *
   * @see AnnotationService for more details on linking annotations and pages.
   */
  class DoQuBookSHARED_EXPORT Annotation  : public SimpleMarshallable
  {
      Q_OBJECT

      Q_PROPERTY (int id READ getId WRITE setId)
      Q_PROPERTY (QString name READ getName WRITE setName)
      Q_PROPERTY (QString description READ getDescription WRITE setDescription)
    public:

      MARSHALLABLE_DECLARE_SEE_ALSO()

      MARSHALLABLE_CREATE_FONCTIONS (Annotation)

      QString getName() const {
        return _name;
      }
      void setName (const QString & name) {
        _name = name;
        emitModified();
      }

      virtual void accept (AnnotationVisitor &/*visitor*/) {}

      virtual bool equals (const Annotation & autre) {
        return (_name.toStdString() == autre._name.toStdString() );
      }

      virtual void merge (MarshallablePtr ann);

    public slots:

      void setDescription (const QString & arg) {
        m_description = arg;
      }

      void setId (int arg) {
        m_id = arg;
      }



    signals:
      void modified();

    public :
      Annotation (const QString & name);
      Annotation ();

      QString getDescription() const {
        return m_description;
      }

      int getId() const {
        return m_id;
      }

    protected:
      int m_id;
      QString _name;
      QString m_description;
  };

} // namespace

Q_DECLARE_METATYPE ( DoQuBook::AnnotationPtr )

#endif // ANNOTATION_HPP
