/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "AnnotationLocal.hpp"

#include <DoQuBook/AnnotatedPolygon.hpp>

using namespace DoQuBook;

AnnotationLocal::AnnotationLocal (const QString & name)
    : Annotation (name)
    , m_type ("annotationLocal")
{}

void AnnotationLocal::addAnnotatedPolygon (AnnotatedPolygonPtr annotatedPolygon)
{
    // Do not add a polygon with id != -1
    if (annotatedPolygon->getId() != -1) {

        AnnotatedPolygonPtr existingPoly = findAnnotatedPolygonById (annotatedPolygon->getId() );
        if (!existingPoly.isNull() ) {
            // found the coresponding polygon
            // replacing its content
            existingPoly->setValue (annotatedPolygon->getValue() );
            existingPoly->setPoints (annotatedPolygon->getPoints() );
        }
        else {   /* Did not find a polygon with same id*/
            AnnotatedPolygonPtr existingPoly = findAnnotatedByPointsMatching (annotatedPolygon);
            /// try to find polygon with same location
            if (existingPoly.isNull() ) {
                // Not found ? final add
                _annotatedPolygons.push_back (annotatedPolygon);
                connect (annotatedPolygon.data(), SIGNAL (modified() ),
                         this, SLOT (emitModified() ) );
            }
            else {
                // Found with same location ? replace existing poly id and value
                existingPoly->setId (annotatedPolygon->getId() );
                existingPoly->setValue (annotatedPolygon->getValue() );
            }
        }

    }
    else {
        // if the polygon id is -1
        AnnotatedPolygonPtr existingPoly = findAnnotatedByPointsMatching (annotatedPolygon);
        /// try to find polygon with same location
        if (existingPoly.isNull() ) {
            // Not found
            // add it
            _annotatedPolygons.push_back (annotatedPolygon);
            connect (annotatedPolygon.data(), SIGNAL (modified() ),
                     this, SLOT (emitModified() ) );

        }
        else {   /* Did find polygon with same location */
            // replace its value
            existingPoly->setValue (annotatedPolygon->getValue() );
            // keep the existing id
            annotatedPolygon->setId (existingPoly->getId() );
        }
    }

    emitModified();
}

void AnnotationLocal::accept (AnnotationVisitor & visitor)
{
    visitor.visit (this);
}

AnnotationLocalPtr AnnotationLocal::create (const QString & name)
{
    AnnotationLocalPtr  ptr = AnnotationLocalPtr (new AnnotationLocal (name) );
    return ptr;
}

void AnnotationLocal::setAnnotatedPolygons (QList<AnnotatedPolygonPtr> ann)
{
    _annotatedPolygons = ann;
    foreach (AnnotatedPolygonPtr poly, ann)
    connect (poly.data(), SIGNAL (modified() ),
             this, SLOT (emitModified() ) );

    emitModified();

}


void DoQuBook::AnnotationLocal::merge (MarshallablePtr ann)
{
    Annotation::merge (ann);
    AnnotationLocalPtr remote = ann.dynamicCast<AnnotationLocal>();
    foreach (AnnotatedPolygonPtr poly, getAnnotatedPolygons () ) {
        AnnotatedPolygonPtr remoteFoundPoly;
        if (poly->getId() == -1) {
            remoteFoundPoly = remote->findAnnotatedPolygon (poly);
        }
        else {
            remoteFoundPoly = remote->findAnnotatedPolygonById (poly->getId() );
        }
        if (!remoteFoundPoly.isNull () ) {
            poly->merge (remoteFoundPoly);
        }
    }
    foreach (AnnotatedPolygonPtr poly, remote->getAnnotatedPolygons () ) {
        AnnotatedPolygonPtr sameAnnPoly;
        sameAnnPoly = findAnnotatedPolygon (poly);
        if (sameAnnPoly.isNull () ) {
            addAnnotatedPolygon (poly);
        }
    }
    emitModified();
}

DoQuBook::AnnotatedPolygonPtr DoQuBook::AnnotationLocal::findAnnotatedPolygon (DoQuBook::AnnotatedPolygonPtr poly)
{
    bool found = false;

    QListIterator<AnnotatedPolygonPtr> it (_annotatedPolygons);

    AnnotatedPolygonPtr foundAnnPolygon;
    while (!found && it.hasNext () ) {
        foundAnnPolygon = it.next ();
        found = foundAnnPolygon->equals (* (poly) );
    }
    if (found) {
        return foundAnnPolygon;
    }
    else {
        return AnnotatedPolygonPtr();
    }
}

bool DoQuBook::AnnotationLocal::equals (const DoQuBook::Annotation & autre)
{
    bool annEqual = Annotation::equals (autre);
    const AnnotationLocal * local = dynamic_cast<const AnnotationLocal *> (&autre);
    if (local) {
        bool equal = local->_annotatedPolygons.size() == _annotatedPolygons.size();
        int i = 0;
        while (equal && i < _annotatedPolygons.size() ) {
            bool found = false;
            int j = 0;
            while (!found && j < local->_annotatedPolygons.size() ) {
                found = _annotatedPolygons.at (i)->equals (* (local->_annotatedPolygons.at (j) ) );
                j++;
            }
            equal = found;
            i++;
        }

        return equal && annEqual;
    }
    else {
        return annEqual;
    }
}

DoQuBook::AnnotatedPolygonPtr DoQuBook::AnnotationLocal::findAnnotatedByPointsMatching (DoQuBook::AnnotatedPolygonPtr poly)
{
    bool found = false;
    QListIterator<AnnotatedPolygonPtr> it (_annotatedPolygons);

    AnnotatedPolygonPtr foundAnnPolygon;
    while (!found && it.hasNext () ) {
        foundAnnPolygon = it.next ();
        found = foundAnnPolygon->pointsMatches (*poly);
    }
    if (found) {
        return foundAnnPolygon;
    }
    else {
        return AnnotatedPolygonPtr();
    }
}

AnnotatedPolygonPtr AnnotationLocal::findAnnotatedPolygonById (int id)
{
    bool found = false;
    QListIterator<AnnotatedPolygonPtr> it (_annotatedPolygons);

    AnnotatedPolygonPtr foundAnnPolygon;
    while (!found && it.hasNext () ) {
        foundAnnPolygon = it.next ();
        qDebug() << foundAnnPolygon->getId() << "==" << id;
        found = foundAnnPolygon->getId() == id;
    }
    if (found) {
        return foundAnnPolygon;
    }
    else {
        return AnnotatedPolygonPtr();
    }
}

void AnnotationLocal::convert_annotatedPolygons (QVariantList * list)
{

    foreach (AnnotatedPolygonPtr p, _annotatedPolygons) {
        *list << p->serialize();
    }

}

void AnnotationLocal::from_annotatedPolygons (QVariant v)
{
    MarshallablePtr factory = AnnotatedPolygon::create();
    listFrom<AnnotatedPolygonPtr, AnnotatedPolygon> (v, factory , _annotatedPolygons);
}

void DoQuBook::AnnotationLocal::removeAnnotatedPolygon (DoQuBook::AnnotatedPolygonPtr annotatedPolygon)
{
    _annotatedPolygons.removeOne (annotatedPolygon);
}
