#ifndef ANNOTATIONSERVICE_HPP
#define ANNOTATIONSERVICE_HPP


#include <DoQuBook/DoQuBook_global.hpp>

#include <QSharedPointer>

#include <DoQuBook/AnnotationLocal.hpp>
#include <DoQuBook/AnnotationGlobal.hpp>

#include <DoQuBook/Page.hpp>

#include <DoQuBook/Page.hpp>
#include <DoQuBook/CRUDService.hpp>

namespace DoQuBook
{
   typedef meta_iterator<AnnotationPtr>::IteratorPtr AnnotationIteratorPtr;


    class AnnotationService;
    typedef QSharedPointer<AnnotationService> AnnotationServicePtr;


    /**
     * @brief The AnnotationService class manages annotations stored on a DoQuBookWeb repo.
     * This class also alows to link (and manage the link) annotations and pages.
     * @see Annotation
     * \attention Only one annotation of specific name per page !
     *
     * Usage :
\code
DoQuBook::RepositoryPtr repo = DoQuBook::ApplicationContext::instance()->getDefaultRepository();
PageAnnotationServicePtr annotationSrv = repo->getAnnotationService();
// annotationSrv->[...method...]
\endcode
     */
    class DoQuBookSHARED_EXPORT AnnotationService : virtual public CRUDService<Annotation, int>
    {
        public:
            AnnotationService();
            virtual ~AnnotationService();

            /** Returns all annotation names stored on the repo.
            *     - Usefull if you need to check wether an annotation exist or not.
            *     - Or to populate a combobox for example.
            **/
            virtual QStringList getAllAnnotationNames() = 0;

            /**
             * Returns all pages annotated by annotationName.
             * Usage :
             \code
  PageIteratorPtr it = service->getPagesWithAnnotationName("test");
  while(it->hasNext()) {
    PagePtr p = it->next();
    // -- do some with p.
  }
             \endcode
             * @brief getPagesWithAnnotationName
             * @param annotationName
             * @return
             */
            virtual PageIteratorPtr getPagesWithAnnotationName(const QString& annotationName) = 0;

            /**
             * Adds a global annotation (name + value) to an existing page.
             *   \attention the annotation does not need to be added first
             *   \attention the page must exists on the repo (and attached (id != -1))
             * @param p
             * @param ann
             * @return
             */
            virtual AnnotationGlobalPtr addGlobalAnnotationToPage(PagePtr &p, AnnotationGlobalPtr &ann) = 0;

            /**
             * Adds a global annotation (name + value) to an existing page.
             *   \attention the annotation does not need to be added first
             *   \attention the page must exists on the repo (and attached (id != -1))
             * @param p
             * @param ann
             * @return
             */
            virtual AnnotationGlobalPtr addGlobalAnnotationToPage(PagePtr &p, const QString &name, const QVariant &value) = 0;

            /**
             * Adds a local annotation (name + value) to an existing page.
             *   \attention the annotation does not need to be added first
             *   \attention the page must exists on the repo (and attached (id != -1))
             * @param p
             * @param ann
             * @return
             */
            virtual AnnotationLocalPtr addLocalAnnotationToPage(PagePtr &p, AnnotationLocalPtr &ann) = 0;

            /**
             * Returns all annotation of a specific page.
             * \attention the page must exists on the repo.
             Usage :
             \code
AnnotationIteratorPtr it = service->getAnnotationOfPage(page);
while(it->hasNext()) {
   AnnotationPtr ann = it->next();
}
             \endcode
             * @param p
             * @return
             */
            virtual AnnotationIteratorPtr getAnnotationOfPage(const PagePtr &p) = 0;

            /**
             * @brief returns the one and only annotation of a specific name for a specific page !
             *  \attention if the annotation does not exists or if it is not a local annotation will return null.
             * @param p
             * @param name
             * @return
             */
            virtual AnnotationLocalPtr getAnnotationLocalOfPageByName(const PagePtr &p, const QString &name) = 0;

            /**
             * @brief returns the one and only annotation of a specific name for a specific page !
             *  \attention if the annotation does not exists or if it is not a global annotation will return null.
             * @param p
             * @param name
             * @return
             */
            virtual AnnotationGlobalPtr getAnnotationGlobalOfPageByName(const PagePtr &p, const QString &name) = 0;

            // convenient methods to avoid the dynamicCast of a smart pointer folowing the get method
            virtual AnnotationLocalPtr getAnnotationLocal (const int & id) = 0;
            // convenient methods to avoid the dynamicCast of a smart pointer folowing the get method
            virtual AnnotationGlobalPtr getAnnotationGlobal (const int & id) = 0;
    };

}

#endif // ANNOTATIONSERVICE_HPP
