/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "ApplicationContext.hpp"
#include <QSettings>

#include <QDir>

#include <QNetworkDiskCache>

#include <DoQuBook/PluginLoader.hpp>
#include <QApplication>


#ifdef HAVE_QT5
  #include <QJsonDocument>
#else
  #include <qjson/parser.h>
#endif

using namespace DoQuBook;


//const QString ApplicationContext::DEFAULT_REPOSITORY_NAME = "Gallifrey";

//const QString ApplicationContext::DEFAULT_REPOSITORY_NAME = "LocalhostDev";

const QString ApplicationContext::DEFAULT_REPOSITORY_NAME = "digidoc";

ApplicationContext::ApplicationContext()
  : _cacheActivated (true)
  , _internetPluginsActivated (true)
{
  organisationName = "labri";
  appName = "DoQuBook";
  initConfiguredReposFromConfigurationFile();
  initPluginKeysAliasesFromConfigurationFile();
}

void ApplicationContext::initConfiguredReposFromConfigurationFile()
{
  QList<QString> dqbResourceFolderPaths = getReposConfFilesPaths();
  foreach (const QString &path, dqbResourceFolderPaths) {  
    QFile file(path);
    file.open(QIODevice::ReadOnly);
    QByteArray content = file.readAll();
    file.close();

    QVariant v;
#ifdef HAVE_QT5
    QJsonDocument doc = QJsonDocument::fromJson(content);
    v = doc.toVariant();
#else
    QJson::Parser parser;
    v = parser.parse(content);
#endif
    QVariantList repos = v.toList();
    foreach (const QVariant &repo, repos) {
      _configuredRepos.append(repo);
    }
  }
  qDebug() << "Found " <<  _configuredRepos.size( ) << "repos ";
}

void ApplicationContext::initPluginKeysAliasesFromConfigurationFile()
{
  QList<QString> aliasesPaths = getPluginKeysAliasesConfFilesPaths();
  foreach (const QString &path, aliasesPaths) {
    QFile file(path);
    file.open(QIODevice::ReadOnly);
    QByteArray content = file.readAll();
    file.close();

    QVariant v;
#ifdef HAVE_QT5
    QJsonDocument doc = QJsonDocument::fromJson(content);
    v = doc.toVariant();
#else
    QJson::Parser parser;
    v = parser.parse(content);
#endif
    foreach (QVariant alias, v.toList()) {
      QVariantMap map = alias.toMap();
      _aliases[map["alias"].toString()] = map["value"].toString();
    }
  }
  qDebug() << "Plugin Aliases " << _aliases;
}

QList<DoQuBook::RepositoryPtr> DoQuBook::ApplicationContext::getRepositories()
{
  PluginLoader *loader = PluginLoader::instance();

  QList<DoQuBook::RepositoryPtr> repositories;

  foreach (QVariant repoV, _configuredRepos) {
    QVariantMap repoMap = repoV.toMap();
    RepositoryPtr repo = loader->getPlugin<Repository>(repoMap["plugin"].toString())->clone();
    repo->setName (repoMap["name"].toString());
    repo->setParam (repoMap["conf"]);
    repositories.append (repo);
  }

  /// For RC - 001 we programatically add Gallifrey repository :
  //  RepositoryPtr gallifrey = loader->getPlugin<Repository>("labri.DBWebRepository.0.0.1")->clone();
  //  gallifrey->setName ("Gallifrey");
  //  QVariant url = "http://sd-22392.dedibox.fr:8080";

  //  repositories.append (gallifrey);


  /// And LocalhostDev
  /*RepositoryPtr local = DoQuBookWebRepository::create();
    local->setName ("LocalhostDev");
    QVariant urlLocal = "http://localhostdev:8080";
    local->setParam (urlLocal);
    repositories.append (local);
    */

  /// And DIGIDOC
  //  RepositoryPtr digidoc = loader->getPlugin<Repository>("labri.DBWebRepository.0.0.1")->clone();
  //  digidoc->setName ("digidoc");
  //  QVariant urldigidoc = "http://digidoc-repo.labri.fr";
  //  digidoc->setParam (urldigidoc);
  //  repositories.append (digidoc);

  /// And SomeOther
  //    RepositoryPtr someOther = DoQuBookWebRepository::create();
  //    someOther->setName("SomeOther");
  //    QVariant urlLocal2 = "http://192.168.0.14:8080";
  //    someOther->setParam(urlLocal2);
  //    repositories.append(someOther);
  return repositories;
}

RepositoryPtr ApplicationContext::getDefaultRepository()
{
  return getRepository (DEFAULT_REPOSITORY_NAME);
}

void DoQuBook::ApplicationContext::addRepository (RepositoryPtr repo)
{
  QList<DoQuBook::RepositoryPtr> allRepos = getRepositories();
  allRepos.append (repo);
  saveRepos (allRepos);
}

void DoQuBook::ApplicationContext::saveRepos (const QList<DoQuBook::RepositoryPtr> &repos)
{
  QSettings settings (organisationName, appName);
  settings.beginWriteArray ("repos");
  int i = 0;
  foreach (RepositoryPtr repoPtr, repos) {
    settings.setArrayIndex (i);
    settings.setValue ("repoName", repoPtr->getName() );
    settings.setValue ("providerParam",  repoPtr->getParam() );
    i++;
  }
  settings.endArray();
  settings.sync();
}

QStringList DoQuBook::ApplicationContext::getRepositoriesNameList()
{
  QStringList nameList;
  QList<RepositoryPtr> repos = getRepositories();
  foreach (RepositoryPtr repo, repos) {
    nameList.append (repo->getName() );
  }
  return nameList;
}

DoQuBook::RepositoryPtr DoQuBook::ApplicationContext::getRepository (const QString & name)
{
  QList<RepositoryPtr> repos = getRepositories();

  bool found = false;
  QListIterator<RepositoryPtr> reposIt (repos);
  RepositoryPtr foundRepo;

  while (!found && reposIt.hasNext() ) {
    foundRepo = reposIt.next();
    found = foundRepo->getName() == name;
  }
  if (!found) {
    throw MessageException ("Could not find repository : " + name);
  }

  return foundRepo;

}

QString DoQuBook::ApplicationContext::getUserName()
{
  QSettings settings (organisationName , appName);
  if (!settings.contains ("username") ) {
    //qCritical() << "No user name set";
    // TODO: REACTIVATE
    //throw QString ("No user name set");
  }

  return settings.value ("username").toString();
}

void DoQuBook::ApplicationContext::setUserName (const QString & username)
{
  QSettings settings (organisationName , appName);
  settings.setValue ("username", username);
  settings.sync();
}

QString ApplicationContext::getDoQuBookDataPath()
{
#if QT_VERSION >= 0x050000
  return QStandardPaths::writableLocation (QStandardPaths::HomeLocation) +"/.DoQuBook";
#else
  return QDesktopServices::storageLocation (QDesktopServices::HomeLocation)+"/.DoQuBook";
#endif
}

QString ApplicationContext::getAppDataPath()
{
#if QT_VERSION >= 0x050000
  return QStandardPaths::writableLocation (QStandardPaths::DataLocation);
#else
  return QDesktopServices::storageLocation (QDesktopServices::DataLocation);
#endif
}

QString DoQuBook::ApplicationContext::getTempRepoPath()
{

#if QT_VERSION >= 0x050000
  return QStandardPaths::writableLocation (QStandardPaths::TempLocation);
#else
  return QDesktopServices::storageLocation (QDesktopServices::TempLocation);
#endif
}

QNetworkDiskCache * ApplicationContext::getCache()
{
  QNetworkDiskCache * diskCache = new QNetworkDiskCache;

  diskCache->setCacheDirectory (pathToCacheDirectory() );
  diskCache->setMaximumCacheSize (getCacheSize() );
  return diskCache;
}

void ApplicationContext::clearCache()
{
  QNetworkDiskCache * diskCache = new QNetworkDiskCache;
  diskCache->setCacheDirectory (pathToCacheDirectory() );
  diskCache->setMaximumCacheSize (getCacheSize() );
  diskCache->clear();
  delete diskCache;

}

bool ApplicationContext::cacheActivated()
{
  return _cacheActivated;
}

void ApplicationContext::activateCache (bool a)
{
  _cacheActivated = a;
}

void ApplicationContext::deactivateInternetPlugins()
{
  _internetPluginsActivated = false;
}

void ApplicationContext::activateInternetPlugins()
{
  _internetPluginsActivated = true;
}

bool ApplicationContext::isInternetPluginsActivated()
{
  return _internetPluginsActivated;
}

QList<QString> ApplicationContext::getResourcesLookUpDirs()
{
  QList<QString> lookUpDirs;
  lookUpDirs << ".";
  lookUpDirs << qApp->applicationDirPath();
#if defined(Q_OS_WIN)
  lookUpDirs << "C:/Program Files/DoQuBook_resources/";
  lookUpDirs << "C:/Program Files (x86)/DoQuBook_resources/";
  lookUpDirs << qApp->applicationDirPath() + "/DoQuBook_resources/";
#elif defined(Q_OS_MAC)
  QDir rDir = QDir (qApp->applicationDirPath() );
  if (rDir.dirName() == "MacOS") {
    rDir.cdUp();
    if (rDir.cd ("Resources") ) {
      lookUpDirs << rDir.absolutePath();
    }
  }
  QString pathToResources = "/Library/Frameworks/DoQuBook.framework/Resources/";
  lookUpDirs << pathToResources;
#elif defined(Q_OS_LINUX)
  lookUpDirs << "/usr/local/share/DoQuBook/Resources");
#endif
  return lookUpDirs;
}

QList<QString> ApplicationContext::getReposConfFilesPaths()
{
  return searchForResourceFile("_repos.json");
}

QList<QString> ApplicationContext::getPluginKeysAliasesConfFilesPaths()
{
  return searchForResourceFile("_plugins_aliases.json");
}

QList<QString> ApplicationContext::searchForResourceFile(const QString& suffix)
{
  QList<QString> filePaths;
  QList<QString> lookUpDirs = getResourcesLookUpDirs();
  qDebug() << lookUpDirs;
  bool found = false;
  QListIterator<QString> it(lookUpDirs);
  while(!found && it.hasNext()) {
    QString possibleDir = it.next();
    QDir dir (possibleDir);
    QStringList files = dir.entryList();
    foreach(const QString &filePath , files) {
      if(filePath.endsWith(suffix)) {
        found = true;
        filePaths.append(dir.absoluteFilePath(filePath));
      }
    }
  }
  return filePaths;
}

QString ApplicationContext::getPluginKey(const QString& pluginKey)
{
  if(_aliases.contains(pluginKey))
    return _aliases[pluginKey];
  else
    return pluginKey;
}

QString ApplicationContext::getDefaultMarshallableFormaterID()
{
  return "labri.QJSONMarshallableFormater.0.0.1";
}


