/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include <DoQuBook/Book.hpp>

#include <QVariantMap>

#include <QDebug>


using namespace DoQuBook;

Book::Book (QString name)
    : _name (name)
    , m_version (0)
    , m_id (-1)
{

}

void Book::addPage (PagePtr page)
{
    _pages.push_back (page);
    connect (page.data(), SIGNAL (modified() ),
             this, SLOT (emitModified() ) );
    emitModified();
}

PagePtr Book::getPage (int pageNum) const
{
    if (! (pageNum >= 0 && pageNum < _pages.size() ) ) {
        throw "pageNum < 0 or pageNum > page.size() is incorrect";
    }

    return _pages[pageNum];
}

void Book::removePage (PagePtr page)
{
    _pages.removeOne (page);
    emitModified();
}

void Book::removePage (int pageNum)
{
    _pages.removeAt (pageNum);
    emitModified();
}

void Book::setPages (QList<PagePtr> pages)
{
    if (!_pages.empty() ) {
        pages.clear();
    }

    foreach (PagePtr page, pages) {
        connect (page.data(), SIGNAL (modified() ), this, SLOT (emitModified() ) );
        addPage (page);
    }

    emitModified();
    emit pagesChanged (pages);
}

void DoQuBook::Book::merge (BookPtr book)
{
    if (!book.isNull() && book->getId() == getId() ) {

        this->setName (book->getName () );
        this->setVersion (book->getVersion () );

        foreach (PagePtr page, book->getPages () ) {
            PagePtr samePage = findPage (page);
            if (samePage.isNull () ) {
                addPage (page);
            }
            else {
                samePage->merge (page);
            }
        }
    }

    emit bookMerged ();
    emitModified();
}

DoQuBook::PagePtr DoQuBook::Book::findPage (const DoQuBook::PagePtr page) const
{
    bool found = false;

    QListIterator<PagePtr> it (_pages);
    PagePtr foundPage;
    while (!found && it.hasNext () ) {
        foundPage = it.next ();
        found = foundPage->getId() == page->getId ();
    }
    if (found) {
        return foundPage;
    }
    else {
        return PagePtr();
    }

}

DoQuBook::PagePtr DoQuBook::Book::findPageByName (const DoQuBook::PagePtr page) const
{
    bool found = false;

    QListIterator<PagePtr> it (_pages);
    PagePtr foundPage;
    while (!found && it.hasNext () ) {
        foundPage = it.next ();
        found = foundPage->getName () == page->getName ();
    }
    if (found) {
        return foundPage;
    }
    else {
        return PagePtr();
    }
}

int DoQuBook::Book::findPageIndex (const DoQuBook::PagePtr page) const
{
    bool found = false;

    QListIterator<PagePtr> it (_pages);
    PagePtr foundPage;
    int index = -1;
    while (!found && it.hasNext () ) {
        foundPage = it.next ();
        index++;
        found = foundPage->getName() == page->getName();
    }
    return index;

}

DoQuBook::PagePtr DoQuBook::Book::findPageById (int id) const
{
    bool found = false;

    QListIterator<PagePtr> it (_pages);
    PagePtr foundPage;
    while (!found && it.hasNext () ) {
        foundPage = it.next ();
        found = foundPage->getId() == id;
    }
    if (found) {
        return foundPage;
    }
    else {
        return PagePtr();
    }
}

// namespace
