/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "DoQuBookApp.hpp"
#include <QTimer>
#include <QDebug>
#include <iostream>
#include <QSettings>
#include <QDir>
#include <QTime>
#include <QFile>

#include <assert.h>

#include <DoQuBook/Book.hpp>
#include <DoQuBook/ApplicationContext.hpp>

using namespace DoQuBook;

DoQuBookApp::DoQuBookApp (QCoreApplication & app,
                          const QString & orgName,
                          const QString & appName,
                          const QString & exeName)
    : QObject (NULL)
    , _app (app)
    , _orgName (orgName)
    , _appName (appName)
    , _exeName (exeName)
{
    if (_exeName.isEmpty() ) {
        _exeName = _appName;
    }
}

void myMessageOutput (QtMsgType type, const char * msg);


#if defined(Q_OS_WIN)

#define COLOR_DEBUG ""
#define COLOR_WARN ""
#define COLOR_CRITICAL ""
#define COLOR_FATAL ""
#define COLOR_RESET ""

#define LOG_WRITE(OUTPUT, COLOR, LEVEL, MSG) OUTPUT << \
    QTime::currentTime().toString("hh:mm:ss").toStdString()  << \
    " " LEVEL " : " << MSG << "\n"

#else

#define COLOR_DEBUG "\033[32;1m"
#define COLOR_WARN "\033[33;1m"
#define COLOR_CRITICAL "\033[31;1m"
#define COLOR_FATAL "\033[33;1m"
#define COLOR_RESET "\033[0m"


#define LOG_WRITE(OUTPUT, COLOR, LEVEL, MSG) OUTPUT << COLOR << \
    QTime::currentTime().toString("hh:mm:ss").toUtf8().constData()  << \
    " " LEVEL " : " << COLOR_RESET << MSG << "\n"

#endif




void myCustomLogWriterQt4 (QtMsgType type, const char * msg)
{
    switch (type) {
        case QtWarningMsg:
            LOG_WRITE (std::cout, COLOR_WARN, "WARNING ", msg);
            break;
        case QtCriticalMsg:
            LOG_WRITE (std::cerr, COLOR_CRITICAL, "CRITICAL", msg);
            break;
        case QtFatalMsg:
            LOG_WRITE (std::cout, COLOR_FATAL, "FATAL  ", msg);
            break;
        case QtDebugMsg:
            LOG_WRITE (std::cout, COLOR_DEBUG, "DEBUG   ", msg);
            break;
    }
}
#if QT_VERSION >= 0x050000
void myCustomLogWriterQt5 (QtMsgType type, const QMessageLogContext&, const QString& msg)
{
  myCustomLogWriterQt4(type, msg.toUtf8().constData());
}
#endif
void installMessageHandler() {
#if QT_VERSION >= 0x050000
     qInstallMessageHandler (myCustomLogWriterQt5);
#else
     qInstallMsgHandler (myCustomLogWriterQt4);
#endif

}

int DoQuBookApp::exec()
{
  installMessageHandler();
    if (_app.arguments().contains ("-help")
        || _app.arguments().contains ("--help") ) {
        showHelpMessage();
        return -1;
    }

    if (QString (_app.metaObject()->className() ) != "QApplication") {
        return execConsoleApp();
    }
    else {

        initParams();
        runApp (_params);
        return _app.exec();
    }
}

int DoQuBook::DoQuBookApp::execConsoleApp()
{
    initParams();
    QObject::connect ( this, SIGNAL ( done() ), &_app, SLOT (quit() ) );
    QTimer::singleShot (0, this, SLOT (runSlot() ) );
    return _app.exec();
}



void DoQuBookApp::runSlot()
{
    installMessageHandler();

    qDebug() << "Running " + _orgName + "." + _appName;
    runApp (_params);
    qDebug() << "Finished";
    emit done();
}

void DoQuBook::DoQuBookApp::installAsDaemon()
{

#if defined(Q_OS_WIN)
    /// \todo create a deamon for windows

#elif defined(Q_OS_MAC)
    QString homePath = QDir::homePath();
    QString launchAgentPlistPath = homePath + "/Library/LaunchAgents/com." + _orgName + "." + _appName + ".plist";
    qDebug() << "Installing Launch Agent in :" ;
    qDebug() << "=== " << launchAgentPlistPath;

    QSettings settings (launchAgentPlistPath, QSettings::NativeFormat);
    settings.setValue ("Disabled", false);
    settings.setValue ("Label", "com." + _orgName + "." + _appName);
    settings.setValue ("WorkingDirectory", "/tmp");
    settings.setValue ("RunAtLoad", true);
    //settings.setValue("StartInterval", 360);

    settings.setValue ("Program", _app.applicationDirPath() + "/" + _exeName);

    settings.sync();

#elif defined(Q_OS_LINUX)
    ///  For Linux http://www.commentcamarche.net/faq/6858-demarrer-un-programme-en-tant-que-service-demon
    /// Use upstart
    QString fullTemplate;
    fullTemplate.append ("# @exe_name@ - set system hostname \n");

    fullTemplate.append ("description     \"DoQuBookAPP\" \n");

    fullTemplate.append ("start on runlevel 2 \n");
    fullTemplate.append ("stop on runlevel [^2345] \n");

    fullTemplate.append ("# Set any Environmental variables you need \n");
    fullTemplate.append ("env DISPLAY=:0.0 \n");
    fullTemplate.append ("exec @exe_path@ > /tmp/@exe_name@.log");

    fullTemplate.replace ("@exe_name@", _exeName);
    fullTemplate.replace ("@exe_path@",  _app.applicationDirPath() + "/" + _exeName);

    qDebug() << "Installing Upstart script in :" ;
    qDebug() << "=== " << "/etc/init/" + _exeName + ".conf";

    QFile confFile ("/etc/init/" + _exeName + ".conf");

    if (!confFile.open (QIODevice::WriteOnly | QIODevice::Text) ) {
        qCritical() << "Could not open file (did you forget to use sudo)";
        return;
    }

    QTextStream out (&confFile);
    out << fullTemplate;

#endif

}

void DoQuBook::DoQuBookApp::uninstallDaemon()
{
    QString homePath = QDir::homePath();
    QDir dir (homePath + "/Library/LaunchAgents/");
    dir.remove ("com." + _orgName + "." + _appName + ".plist");
}

QList<DoQuBookApp::ProgramParameter> DoQuBook::DoQuBookApp::paramsAndDescriptions()
{
    QList<DoQuBookApp::ProgramParameter> paramsAndDesc;
    ProgramParameter page ("page", false, "Page = existing file or exiting id on repo");
    paramsAndDesc.append (page);
    return paramsAndDesc;
}


void DoQuBook::DoQuBookApp::showHelpMessage()
{
    std::cout << "--------------------------------------------------------------------------------" << std::endl;
    std::cout << "--------------------------------------------------------------------------------" << std::endl;
    std::cout << "AppExeName " << "  -[option0] [optionValue0] -[option1] [optionValue1] ..." << std::endl;
    std::cout << "--------------------------------------------------------------------------------" << std::endl;
    std::cout << "Options : " << std::endl;
    QListIterator<ProgramParameter> it (paramsAndDescriptions() );
    while (it.hasNext() ) {
        ProgramParameter param = it.next();
        std::cout << param._name.toStdString() << ": \t " << "Required : " << param._required << " Description : " << param._description.toStdString() << std::endl;
    }
    std::cout << "--------------------------------------------------------------------------------" << std::endl;
    std::cout << "--------------------------------------------------------------------------------" << std::endl;
}

void DoQuBook::DoQuBookApp::initParams()
{
    QStringList args = _app.arguments();
    foreach (ProgramParameter param, paramsAndDescriptions() ) {
        QString option = param._name;
        QString realOption = "-" + option;
        if (args.contains (realOption) ) {
            int i = args.indexOf (realOption);
            if (args.size() > i + 1) {
                _params[option] = args.at (i + 1);
            }
            else if (param._required) {
                showHelpMessage();
                qCritical() << "Missing parameter on command line";
                assert (false);
            }
        }
        else if (param._required) {
            showHelpMessage();
            qCritical() << "Missing parameter on command line";
            assert (false);
        }
    }
}

PagePtr DoQuBook::DoQuBookApp::getPageFromParameters (DoQuBook::RepositoryPtr repo)
{
    PagePtr page;
    QFile f (_params["page"]);

    if (f.exists() ) {
        page = Page::create();
        page->setImage (_params["page"]);
    }
    else  {
        // get page 0 in book 1
        /// \todo retrive page from page service.
        if (!repo) {
            repo = getRepoFromParameters();
        }

        BookServicePtr bookservice = repo->getBookService();
        BookPtr b = bookservice->get (4);
        page = b->getPages() [0];
    }
    if (page.isNull() ) {
        qCritical() << "Page not found";
        throw "Page not found";
    }
    return page;
}

DoQuBook::RepositoryPtr DoQuBook::DoQuBookApp::getRepoFromParameters()
{
    if (_params.contains ("repo") ) {
        DoQuBook::RepositoryPtr repo = DoQuBook::ApplicationContext::instance()->getRepository (_params["repo"]);
        return repo;
    }
    else {
        throw "No parameters name repo found. Please use -repo \"Gallifrey\" for example";
    }
}

