/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#ifndef EXCEPTIONS_HPP
#define EXCEPTIONS_HPP

#include <QString>
#include <QDebug>
#include <QUrl>

#include <DoQuBook/DoQuBook_global.hpp>
#include <DoQuBook/Config.hpp>

namespace DoQuBook
{

  class DoQuBookSHARED_EXPORT Exception
  {
    public:
      Exception();
  };


  class DoQuBookSHARED_EXPORT MessageException : public Exception
  {
    public :
      MessageException (const QString & s)
        : _msg (s) {
#ifdef DEBUG_LOG
        qCritical() << "Exception::MessageException" << _msg;
#endif
      }

      QString message() {
        return _msg;
      }

    private :
      QString _msg;
  };

  class DoQuBookSHARED_EXPORT ConcurencyException : public MessageException
  {
    public :
      ConcurencyException (const QString & e)
        : MessageException (e) {}
  };

  class DoQuBookSHARED_EXPORT MissingMarameterException : public MessageException
  {
    public :
      MissingMarameterException (const QString & paramName)
        : MessageException ("Executor cannot find parameter named " + paramName) {}
  };

  class DoQuBookSHARED_EXPORT ParamParsingException : public MessageException
  {
    public :
      ParamParsingException (const QString & paramName)
        : MessageException ("ParamParsingException " + paramName) {}
  };

  class DoQuBookSHARED_EXPORT CouldNotParseJSONException : public MessageException
  {
    public :
      CouldNotParseJSONException (const QString & e, const QString & failedJSONString)
        : MessageException (e)
        , _failedJSONString (failedJSONString)
      {}

      QString getFailedJSONString() {
        return _failedJSONString;
      }
    private :
      QString _failedJSONString;
  };

  class DoQuBookSHARED_EXPORT ImageNotFoundException
  {
    public :
      ImageNotFoundException (int pageId, QString m)
        : _pageId (pageId)
        , message (m)
      {}

      int getPageId() {
        return _pageId;
      }

      QString getMessage() {
        return message;
      }

    private :
      int _pageId;
      QString message;

  };

  class DoQuBookSHARED_EXPORT AnnotationNotFoundException
  {
    public :
      AnnotationNotFoundException (int id, QString annotationName)
        : _id (id)
        , message ("Annotation of name " + annotationName + " does not exists in page " + id)
      {}

      int getPageId() {
        return _id;
      }

      QString getMessage() {
        return message;
      }

    private :
      int _id;
      QString message;

  };


  class DoQuBookSHARED_EXPORT AnnotationCastException
  {
    public :
      AnnotationCastException (const QString & annName)
        : message ("Annotation of name " + annName + " is of type xxx")
      {}

      QString getMessage() {
        return message;
      }

    private :
      QString message;
  };

  class DoQuBookSHARED_EXPORT InternetConnectionException
  {
    public :
      InternetConnectionException (const QUrl & url)
        : message ("Check Internet connection [ URL :" + url.toString() + " is not reachable] ") {
        qCritical() << "Exception::InternetConnectionException";
        qCritical() << message;
      }

      QString getMessage() {
        return message;
      }

    private :
      QString message;
  };

  class DoQuBookSHARED_EXPORT NetworkConnectionException : public MessageException
  {
    public :
      NetworkConnectionException (const QString & msg) : MessageException (msg) {
        qCritical() << "Exception::NetworkConnectionException" << msg;
      }
  };



  class DoQuBookSHARED_EXPORT ContentNotFoundException : public MessageException
  {
    public :
      ContentNotFoundException (const QUrl & url)
        : MessageException (url.toString() ) {
        qCritical() << "NoContent on url :" + url.toString();
      }
  };

  class DoQuBookSHARED_EXPORT NeedToRegister : public MessageException
  {
    public :
      NeedToRegister (const QUrl & url)
        : MessageException (url.toString() ) {
        qCritical() << "Need to register said url " + url.toString();
      }
  };


  class DoQuBookSHARED_EXPORT CouldNotFindPluginException : public MessageException
  {
    public :
      CouldNotFindPluginException (const QString &pluginId)
        : MessageException ("Could not find a plugin of id : "+pluginId) {
        qCritical() << "Could not find a plugin of id : "+pluginId;
      }
  };

  class DoQuBookSHARED_EXPORT ModelNotTrainedException : public MessageException
  {
    public :
      ModelNotTrainedException (const QString &msg)
        : MessageException (msg) {
        qCritical() << msg;
      }
  };

}

#endif // EXCEPTIONS_HPP
