#ifndef FUTURE_HPP
#define FUTURE_HPP

#include <QList>
#include <QObject>

#include <DoQuBook/Page.hpp>
#include <DoQuBook/MarshallableFormater.hpp>
#include <QThread>

#include <QNetworkAccessManager>
#include <QNetworkReply>

#include <QMutex>
#include <iostream>

template<class T>
class Functor
{

public :
    virtual ~Functor() {}
    virtual void process (QSharedPointer<T>  object) = 0;
};

template<class T>
class Future
{

public :
    virtual ~Future() {}

    virtual QSharedPointer<T> get() = 0;

    virtual void addObserver (Functor<T> *f) = 0;

protected :
    virtual void notifyObservers (QSharedPointer<T>  &t) = 0;

};



/**
  http://blog.qt.digia.com/blog/2010/06/17/youre-doing-it-wrong/
  http://qt-project.org/wiki/Threads_Events_QObjects
  */
class HttpCallManager  : public QObject
{

    Q_OBJECT

  public :
    HttpCallManager (const QString & url) :
      m (NULL)
    , _url (url)
    {}

    void initAll() {
      m = new QNetworkAccessManager (this);
      connect (m, SIGNAL (finished (QNetworkReply *) ), this, SLOT (replyFinished (QNetworkReply *) ) );
    }

    ~HttpCallManager () {
      //m->deleteLater();
    }

    void run();

  signals :
    void contentReady (QByteArray content);

  public slots:
    void replyFinished (QNetworkReply * r) {
      if (r->error() != QNetworkReply::NoError) {
        error (r->error() );
      }
      else {

        QByteArray array = r->readAll();
        if (array.size() > 0) {
          emit contentReady (array);
        }
        else {
          qCritical() << "HttpCallManager response is empty";
        }
      }
      run();

    }

    void readyRead() {

    }

    void error (QNetworkReply::NetworkError /*error*/) {
      std::cout << "error" << std::endl;
    }


  private :
    QNetworkAccessManager * m;
    QString _url;
};
class TThread : public QThread
{
public :
    TThread() {}
    virtual ~TThread() {}
    void run() {
        connect (this, SIGNAL (finished() ), this, SLOT (deleteLater() ) );
        http->run();
        exec();
    }
    void setWorker (HttpCallManager * c) {
        http = c;
    }

private:
    HttpCallManager * http;
};


/**
 * \brief Run a list of Functor<T> in a different thread
 */
template<class T>
class FunctorsThread : public QThread
{
public :
    FunctorsThread() {}
    virtual ~FunctorsThread() {}

    void run() {
        connect (this, SIGNAL (finished() ), this, SLOT (deleteLater() ) );
        foreach (Functor<T> *f, observers) {
            f->process (t);
        }
    }
    void addFunctor (Functor<T> *f) {
        observers.append (f);
    }

    void setObject (QSharedPointer<T> o) {
        t = o;
    }

protected :
    QList<Functor<T> *> observers;
    QSharedPointer<T> t;
};



class HttpResponseHandler : public QObject
{
    Q_OBJECT
public :
    HttpResponseHandler() {}
    virtual ~HttpResponseHandler () {}

    virtual void processResult (QByteArray content) = 0;

public slots:

    void setResult (QByteArray content) {
        processResult (content);
    }
};


template<class T>
class EventHandler
{

public :
    virtual ~EventHandler() {}

    virtual void addObserver (Functor<T> *f) = 0;

protected :
    virtual void notifyObservers (QSharedPointer<T>  t) = 0;

};

/**
 * \brief Creates a long-polling connection to an url, when server sends a response, unmarshalls the content to an instance of type T, calls any notify method of any observer and re-connects to the same url.
 *  An HttpEventHandler is both a HttpResponseHandler and an EventHandler of type T.
 *  - HttpResponseHandler is a simple Q_OBJECT that has a slot setResult that calls a [to be] overloaded method  processResult (QByteArray content) => this is where to unmarshall the object !
 *  - An EventHandler <T> has just two methods addObserver and notifyObserver tha later is called when the event is sent !
 *
 * Basic Usage :
 \code
 HttpEventHandler<Page> eventhandler (url);
 eventhandler.addObserver(aFunctor);
 eventhandler.listen();
 \endcode
  Usage example  :
 \code
  // An observer/functor interface for any Page event (T == Page)
  class PageEvent : public QObject , public Functor<Page> {

      Q_OBJECT

    public :
      PageEvent (QObject *parent);
      virtual void process(DoQuBook::PagePtr object) {
        postUnMarshall(object);
        emit processSignal(object);
      }

    protected :
      virtual void postUnMarshall(const PagePtr &) {}

    signals :
      void processSignal(DoQuBook::PagePtr page);
  };

  // An obeserver/functor for any HTTP Page event.
  // Will also create a HttpEventHandler !
  class DBWebPageEvent : public PageEvent {
    public :
      DBWebPageEvent (const QString &serveradress, const QString subscriptionUrl, QObject *parent = NULL);
      virtual void postUnMarshall(const DoQuBook::PagePtr &object);

      void addFunctor(Functor<Page> *f);

    private :
      QString _serveradress;
      HttpEventHandler<Page> _eventHandler;
  };
  // Impl :
DBWebPageEvent::DBWebPageEvent (const QString &serveradress, const QString subscriptionUrl, QObject *parent)
: PageEvent(parent)
, _serveradress (serveradress)
, _eventHandler (subscriptionUrl)
{
    _eventHandler.addObserver (this);
    _eventHandler.listen();
}

void DBWebPageEvent::postUnMarshall(const DoQuBook::PagePtr &object)
{
    object->installImageService (new DBWebImageService (_serveradress) );
}

void DoQuBook::DBWebPageEvent::addFunctor(Functor<Page> *f)
{
    _eventHandler.addObserver(f);
}


 \endcode
 @see PageService::onPageTaggedEvent
 */
template <class T>
class HttpEventHandler :  public HttpResponseHandler, public EventHandler <T>
{

public :
    HttpEventHandler (const QString & url, MarshallableFormaterPtr marshaller)
        : m_formater (marshaller),
          m_isReady (false),
          manager (NULL),
          http (NULL)
    {
        manager = new TThread;

        http = new HttpCallManager (url);
        connect (http, SIGNAL (contentReady (QByteArray) ), this, SLOT (setResult (QByteArray) ) );
    }

    virtual  ~HttpEventHandler () {
        if (manager)  {
            manager->wait();
            delete manager;
        }
        if (http) {
            disconnect (http, SIGNAL (contentReady (QByteArray) ), this, SLOT (setResult (QByteArray) ) );
            delete http;
            http = NULL;
        }
    }


    void initAll() {
        m_isReady = false;

        this->moveToThread (manager);
        http->moveToThread (manager);
        manager->setWorker (http);
    }

    void listen() {
        initAll();
        startHttpManager();
    }

    void startHttpManager() {
        if (!manager->isFinished() ) {
            manager->terminate();
        }
        manager->start();
    }


    virtual void processResult (QByteArray content) {

        try {
            QSharedPointer<T> object = T::create();
            if (!content.isEmpty() ) {
                m_formater->unMarshall (content, object);
            }

            m_isReady = true;
            notifyObservers (object);
        }
        catch (DoQuBook::CouldNotParseJSONException e) {
            qCritical() << e.getFailedJSONString();
        }
    }

    void addObserver (Functor<T> *f) {
        functors.append (f);
    }

protected :
    void notifyObservers (QSharedPointer<T>  t) {
        FunctorsThread<T> *functorsThreads = new FunctorsThread<T>;
        foreach (Functor<T> *f, functors) {
            functorsThreads->addFunctor (f);
        }
        functorsThreads->setObject (t);
        functorsThreads->start();
    }


protected :
    MarshallableFormaterPtr m_formater;

    bool m_isReady;
    TThread * manager;
    HttpCallManager * http;

    QList< Functor<T>* > functors;
};



template <class T>
class HttpFuture :  public HttpResponseHandler, public Future <T>
{

public :
    HttpFuture(MarshallableFormaterPtr marshaller)
        :  m_isReady (false)
        , m_formater (marshaller)
    {}

    virtual  ~HttpFuture () {}


    void initAll() {
        m_isReady = false;
        manager = new TThread;

        http = new HttpCallManager ("http://uneurl");
        connect (http, SIGNAL (contentReady (QByteArray) ), this, SLOT (setResult (QByteArray) ) );
        this->moveToThread (manager);
        http->moveToThread (manager);
        manager->setWorker (http);
    }

    void run() {
        initAll();
        startHttpManager();
    }

    void startHttpManager() {

        if (!manager->isFinished() ) {
            qWarning() << "Terminate ThreadManager";
            manager->terminate();
        }

        manager->start();

        qDebug() << "Manager started" << QThread::currentThreadId();
        // Will block until a unlock is done in setResult.
        m_mut.lock();
    }

    QSharedPointer<T> get() {
        // Will block until a unlock is done in setResult.
        m_mut.lock();

        m_mut.unlock();
        return object;
    }

    virtual void processResult (QByteArray content) {
        m_content = content;

        object = T::create();
        if (!m_content.isEmpty() ) {
            m_formater->unMarshall (m_content, object);
        }
        m_isReady = true;

        qDebug() << "Future result ready in thread " << QThread::currentThreadId();
        notifyObservers (object);
        m_mut.unlock();
    }

    void addObserver (Functor<T> *f) {
        observers.append (f);
    }

protected :
    void notifyObservers (QSharedPointer<T>  t) {
        foreach (Functor<T> *f, observers) {
            f->process (t);
        }
    }

    QList<Functor<T> *> observers;

protected :
    MarshallableFormaterPtr m_formater;
    QSharedPointer<T> object;
    bool m_isReady;
    TThread * manager;
    QMutex m_mut;
    QByteArray m_content;
    HttpCallManager * http;

};


#endif // FUTURE_HPP
