/*!
 * \file Marshallable.hpp
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#ifndef MARSHALLABLE_HPP
#define MARSHALLABLE_HPP

#include <QVariant>
#include <QMetaProperty>
#include <QStringList>
#include <QPolygon>
#include <QSharedPointer>
#include <DoQuBook/DoQuBook_global.hpp>

class Marshallable;
typedef QSharedPointer<Marshallable> MarshallablePtr;

/*!
 * \author Vincent Rabeux & Maxime Daymard
 * \version 1.0a
 * \brief An abstract class to serialize objects
 *
 * \attention Version 1.0a only supports :
 *     - Serialisation to JSON of primitive type, QString, other Marshallable, and list of Marshallables (ie QList<MarshallablePtr>).
 *     - For polymorphic marshallable we currently support only one level in the hierachy (ie : AbstractClass, ImplClass1, ImplClass2 see DoQuBook::Annotation for an example).
 *
 *
 * The class provides an abstract converter and abstract methods to create marshallable objects. It is based on Qt Property System
 *  (http://doc.qt.nokia.com/latest/properties.html) wich provides a simple reflexion of C++ objects, QVariant
 *  (and QVariantList and QVariantMap) wich acts like a void*, and some other patterns (implementation is simplified by macros.)
 *
 * For more info see : \ref pageMarshallable.
 *
 * \todo move this class to DoQuBook namespace
 *
*/
class DoQuBookSHARED_EXPORT Marshallable : public QObject
{
        Q_OBJECT

    public:
        Marshallable (QObject * parent = 0);
        virtual ~Marshallable() {}

        virtual QVariantMap serialize() = 0;
        virtual void from (QVariantMap map) = 0;
        virtual MarshallablePtr createFromVariantMap (QVariantMap m) = 0;

        virtual QList<MarshallablePtr> seeAlso();
        /**
          * for class Namespace::MyClass default is myclass.
          */
        virtual QString entityName();

    public :

        virtual MarshallablePtr createPtr() = 0;

        static QStringList ignoredProperties();
        static void addIgnoredProperty (const QString propName);

    protected:



        static QStringList ignoredProps;

        QVariantMap serialize (const QPoint & p) const;
        QPoint from (QVariant v);
};


Q_DECLARE_METATYPE ( QList<Marshallable *>)

Q_DECLARE_METATYPE (Marshallable *)

/*!
  Why do we need a Marshallable framework :
Q_DECLARE_METATYPE(DoQuBook::PointPtr)


/// Not possibble see http://www.qtcentre.org/threads/21195-Streaming-custom-QVariant
QDataStream &operator<<(QDataStream &out, const PointPtr &p)
{
    qDebug() << "out";
    QVariant value (QVariant::fromValue<PointPtr>(p));

    out << value;
    return out;
}
QDataStream &operator>>(QDataStream &in, PointPtr &p)
{
    qDebug() << "in";
    double x;
    double y;
    QVariant v;
    in >> v;

    PointPtr pointV = v.value<PointPtr>();

    p->setX(pointV->getX());
    p->setY(pointV->getY());
    return in;
}


int main(int argc, char *argv[])
{
    qRegisterMetaTypeStreamOperators<DoQuBook::PointPtr>("DoQuBook::PointPtr");

    AnnotatedPolygonPtr poly = AnnotatedPolygon::create();
    poly->setValue(2);
    PointPtr p = Point::create();
    p->setX(2);
    p->setX(2);


    PointPtr p1 = Point::create();
    p1->setX(3);
    p1->setY(3);

    poly->addPoint(p);
    poly->addPoint(p1);

    Point p3;
    p3.setX(1);
    p3.setY(3);

    QVariant value (QVariant::fromValue<PointPtr>(p1));
    qDebug() << value;
    qDebug() << value.value<PointPtr>()->getX();

    QVariantMap variant = QJson::QObjectHelper::qobject2qvariant(poly.data());
    qDebug() << variant;


    QJson::Serializer serializer;
    QByteArray array = serializer.serialize(value);
    qDebug() << array;

    QFile file("file.dat");
    file.open(QIODevice::WriteOnly);
    QDataStream d(&file);
    d << value;
    file.close();

    qDebug() << d;

    PointPtr pM = PointPtr::create();
    d >> pM;

    qDebug() << pM->getX();

    return 0;
    }
    Not possibble see http://www.qtcentre.org/threads/21195-Streaming-custom-QVariant
  */


#endif // MARSHALLABLE_HPP
