/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include <DoQuBook/Page.hpp>
#include <QVariantMap>
#include <QFileInfo>
#include <DoQuBook/Logger.hpp>
#include <DoQuBook/ApplicationContext.hpp>
#include <DoQuBook/AttachmentServices.hpp>

#include <DoQuBook/ImageService.hpp>
#include <DoQuBook/Config.hpp>


using namespace DoQuBook;

Page::Page (QObject * parent)
    : SimpleMarshallable (parent)
    , m_imageService (NULL)
    , m_id (-1)
{
}

Page::~Page()
{
    if (m_imageService) {
        delete m_imageService;
        m_imageService = NULL;
    }
}


QImage Page::getImage ( int w) const
{
    if (!m_imageService) {
        /// Image is on disk.
        QImage image;
        image.load (getImageFilePath() );
        return image;
    }
    else {
        return m_imageService->get (_this, w);
    }
}

void Page::setImage (QString fileName)
{
    _imageFilePath = fileName;

    QFileInfo fileInfo (fileName);
    _imageFileName = fileInfo.fileName();

    emitModified();
}

bool Page::equals (const DoQuBook::Page & autre)
{
    return _name == autre._name;
}

QString Page::getImageFilePath() const
{
    return _imageFilePath;
}

QString Page::getImageFileName() const
{
    return _imageFileName;
}

QString Page::getFilePath() const
{
    return _filePath;
}


void Page::setName (const QString & name)
{
    if (_name != name) {
        oldName = _name;
        _name = name;
        emit nameChanged();
    }
}

QString Page::getName() const
{
    return _name;
}


QString DoQuBook::Page::getOldName() const
{
    return oldName;
}

void DoQuBook::Page::merge (MarshallablePtr other)
{
    PagePtr page = other.dynamicCast<Page>();
    if (!page.isNull() ) {
        setName (page->getName () );
    }
}


void DoQuBook::Page::installImageService (ImageService * srv)
{
    if (m_imageService) {
        delete m_imageService;
        m_imageService = NULL;
    }
    m_imageService = srv;
}

void Page::addFeatureSet(const QString& setName, QMap<QString, double> pageFeats)
{
  featureSets[setName] = pageFeats;
}

QMap<QString, double> Page::getFeatureSet(const QString& setName)
{
  return featureSets[setName];
}

bool Page::hasFeatureSet(const QString& setName)
{
  return featureSets.contains(setName);
}

