/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#ifndef PAGE_HPP
#define PAGE_HPP

#include <DoQuBook/DoQuBook_global.hpp>

#include <string>

#include <QObject>
#include <QString>
#include <QList>
#include <QSharedPointer>
#include <QImage>

#include <DoQuBook/SimpleMarshallable.hpp>
#include <DoQuBook/Exceptions.hpp>

#include <DoQuBook/Config.hpp>

#include <DoQuBook/Iterator.hpp>

#include <QUrl>


namespace DoQuBook
{

  class ImageService;

  class Page;
  typedef QSharedPointer<Page> PagePtr;

  typedef meta_iterator<PagePtr>::IteratorPtr PageIteratorPtr;

  /*! \class Page
   * \brief A page with a name and a list of annotations
   *
   * The class provides access to the name and the annotation list of the page
   */
  class DoQuBookSHARED_EXPORT Page : public SimpleMarshallable
  {
      Q_OBJECT

    public:
      friend class Book;
      MARSHALLABLE_CREATE_FONCTIONS (Page)

      explicit Page (QObject * parent = NULL);


      virtual ~Page();

      /// Image (jpeg, ...)
      QString getImageFilePath() const;

      /// \todo quel diff entre getImageFileName  et getImageFilePath
      QString getImageFileName() const;

      /// JSon, XML..
      QString getFilePath() const;

      /// \throws ImageNotFoundException
      Q_INVOKABLE QImage getImage (int w = -1) const;


      void setImage (QString fileName);

      bool equals (const Page & autre);

      /// Page number in the book
      int getNumber() const;

      /// Page number in the book
      void setNumber (int number);

      /*!
              * Page name in the book
              *  \attention  On DoQuBookWeb Page::name must be unique, so please provide a name like bookName-pageName;
              */
      QString getName() const;

      /// Used by an image provider to find the old image.
      QString getOldName() const;


      /// Page name in the book
      void setName (const QString & name);

      virtual void merge (MarshallablePtr page);


      int getId() const {
        return m_id;
      }


    public slots:
      void setId (int arg) {
        if (m_id != arg) {
          m_id = arg;
          emit idChanged (arg);
        }
      }

      void installImageService (ImageService * srv);


      // To be refactored (Not saved by services)
      void addFeatureSet(const QString &setName, QMap<QString, double> pageFeats);
      QMap<QString, double> getFeatureSet(const QString &setName);
      bool hasFeatureSet(const QString &setName);

    signals:
      void idChanged (int arg);

      void nameChanged();

    private :


    protected :

      Q_PROPERTY (int id READ getId WRITE setId NOTIFY idChanged)

      Q_PROPERTY (QString name READ getName WRITE setName NOTIFY nameChanged)




    private:
      QString _name;



      /// example : image.jpeg
      QString _imageFileName;

      /// example : c:/.../image.jpeg
      QString _imageFilePath;

      /// example : c:/.../page.json
      QString _filePath;


      ImageService * m_imageService;

      /// Used by an image provider to find the old image.
      QString oldName;


      int m_id;

      QMap< QString, QMap<QString, double> > featureSets;

  };

} // namespace

Q_DECLARE_METATYPE ( DoQuBook::PagePtr )

#endif // PAGE_HPP
