#ifndef PAGEEVENT_HPP
#define PAGEEVENT_HPP

#include <QObject>
#include <DoQuBook/SimpleMarshallable.hpp>
#include <DoQuBook/Future.hpp>
#include <DoQuBook/Page.hpp>

namespace DoQuBook {


  /**
  * @brief The PageEvent class is a call back interface for any event containing a Page (@see PageService#onPageTaggedEvent for an example).
  *
  * Their is two different ways to use a PageEvent Object :

  * First by using a Qt signals :
  * \code
qRegisterMetaType<PagePtr>("PagePtr");
qRegisterMetaType<DoQuBook::PagePtr>("DoQuBook::PagePtr");
qRegisterMetaType<PagePtr&>("PagePtr&");
qRegisterMetaType<DoQuBook::PagePtr&>("DoQuBook::PagePtr&");

PageEvent *e = pageService->onPageTaggedEvent("mytag");

QObject::connect(e, SIGNAL(processSignal(DoQuBook::PagePtr)),
            &object, SLOT(onPageTagged(DoQuBook::PagePtr)));

    \endcode

    Or by specifying a list of functors as a paramter of a method:
    \code
// The functor :
class UnObject : public QObject, public Functor<Page>  {
    Q_OBJECT

public :
    UnObject(PageServicePtr pageSrv, QObject *parent = 0) :
        QObject(parent), _pageSrv(pageSrv) {}

    virtual void process (DoQuBook::PagePtr p) {
        qDebug() << p->getName();
        _pageSrv->removeTag(p, "mytag");
    }

public slots :
    void onPageTagged(DoQuBook::PagePtr p) {
        qDebug() << "onPageTagged";
        process(p);
    }

private :
    PageServicePtr _pageSrv;
};

//Create the functor :
UnObject object (pageService);

//Register the new functor
QList<Functor<Page>*> functors;
functors << object;
PageEvent *e = pageService->onPageTaggedEvent("mytag", functors);
    \endcode
  *
  */
  class PageEvent : public QObject , public Functor<Page> {

      Q_OBJECT

    public :
      PageEvent (QObject *parent);
      virtual void process(DoQuBook::PagePtr object) {
        postUnMarshall(object);
        emit processSignal(object);
      }

    protected :
      virtual void postUnMarshall(const PagePtr &/*object*/) {}

    signals :
      void processSignal(DoQuBook::PagePtr page);
  };




}

#endif // PAGEEVENT_HPP
