#ifndef PAGESERVICE_HPP
#define PAGESERVICE_HPP

#include <QSharedPointer>
#include <DoQuBook/Page.hpp>
#include <DoQuBook/CRUDService.hpp>
#include <QFileInfo>
#include <DoQuBook/ListRestFileAttachment.hpp>
#include <DoQuBook/Future.hpp>
#include <DoQuBook/PageEvent.hpp>

namespace DoQuBook
{

  class PageService;
  typedef QSharedPointer<PageService> PageServicePtr;

  /**
   * @brief The PageService class, manages pages on a specific repository.
   *
   * You can build a PageService using a repository :
   *  \code
DoQuBook::RepositoryPtr repo = DoQuBook::ApplicationContext::instance()->getDefaultRepository();
PageServicePtr pageSrv = repo->getPageService();
   *  \endcode
   *
   */
  class PageService : virtual public CRUDService<Page, int>
  {
    public:
      PageService();
      virtual ~PageService();

      /// Iterates over all pages stored on the repository
      virtual PageIteratorPtr getAll() = 0;

      /// \section pageSrv-att Attachments
      /// Returns attachments (files) of a specific Page (page must be sync with repo [id!=-1])
      virtual ListRestFileAttachmentPtr getAllAttachments (const PagePtr p) = 0;
      /// Adds an attachment (any kind of file) to a page (page must be sync).
      virtual void addAttachment (const PagePtr page, const QFileInfo & file) = 0;

      /// \section pageSrv-findpages Find pages
      /// Returns all
      virtual PageIteratorPtr getByName (const QString & name) = 0;
      virtual PagePtr findByName (const QString & name) = 0;

      virtual QStringList existingTags() = 0;
      virtual PageIteratorPtr findByTag (const QString & tag) = 0;
      virtual void addTag(const PagePtr &p, const QString &tag) = 0;
      virtual void removeTag(const PagePtr &p, const QString &tag) = 0;

      /// @see PageEvent
      virtual PageEvent* onPageTaggedEvent(const QString &tag, QList< Functor<Page>* > functors = QList< Functor<Page>* >()) = 0;
#ifdef HAVE_CPP11
      /**
       * @brief onPageTaggedEvent using C++11 lambdas.
       * see \file ../tests/testPageServiceEvents.cpp
       - Or the folowing quick example :
       - \code
pageService->onPageTaggedEvent("mytag", [=](PagePtr taggedPage) {
     qDebug() << "The page " << taggedPage->getId() << "is tagged";
} );
       * \endcode
       - Or with a variable :
        \code
std::function<void (PagePtr)> func = [=](PagePtr taggedPage) {
    qDebug() << "The page " << taggedPage->getId() << "is tagged";
};

pageService->onPageTaggedEvent("mytag", func);
        \endcode
        - Or with a function :
        \code
void onPageTaggedFunction(DoQuBook::PagePtr p) {
    qDebug() << "onPageTagged Function called with " << p->getId();
}
pageService->onPageTaggedEvent("mytag", onPageTaggedFunction);
        \endcode

        - Or an object and a method :
        \code
class UnAutreObject {
  public :
    void onPageTaggedMethod(DoQuBook::PagePtr p) {
        qDebug() << "onPageTagged Method called with " << p->getId();
    }
};

UnAutreObject o;
pageService->onPageTaggedEvent("mytag", std::bind(&UnAutreObject::onPageTaggedMethod, o, std::placeholders::_1));

        \endcode
       * @param tag
       * @param func
       */
      virtual void onPageTaggedEvent(const QString &tag, std::function<void (PagePtr)> func) = 0;
#endif

      virtual Future<Page>* onPageAdded (Functor<Page> *functor) = 0;

    public :
      static const QString TAG_BLURRED;
      static const QString TAG_WHITE_PAGE;
      static const QString TAG_BLEEDTHROUGH;
      static const QString TAG_BOOK_COVER;
      static const QString TAG_FOLDED;
      static const QString TAG_REAL_IMAGE;
      static const QString TAG_SYNTHETIC_IMAGE;

  };

}
#endif // PAGESERVICE_HPP
