#ifndef RESTOBJECTMANAGER_HPP
#define RESTOBJECTMANAGER_HPP

#include <QUrl>
#include <DoQuBook/WebClient.hpp>
#include <DoQuBook/MarshallableFormater.hpp>
#include <QDebug>
#include <DoQuBook/Config.hpp>
#include <DoQuBook/Exceptions.hpp>


/*!
  *
  * \page pageREST The DoQuBook REST Client api
  *
  *  DoQuBook uses REST to invoke the WebServices hosted with the JEE app named DoQuBookWeb.
  *  On this page we will show :
  *   - how objects are sent back to DoQuBook from DoQuBookWeb (ie : the server side part),
  *   - how to call a web service and get the objects sent from the server (ie : the cliend side part).
  *
  * \attention
  *    For now, DoQuBook  only support the JSon format for (Un)Marshalling of objects.
  *
  * \section section1PageRest Introduction to REST
  *
  *  REST (Representational State Transfer) is not a protocol nor a format, but a sofware architecture.
  *  The Web follows a REST architecture. Indeed, a REST service is defined by :
  *     - a resource + URL : a resource is what the server manages (ex : a collection of books).
  *     - a HTTP method : GET, POST, PUT, DEL, each of these method corespond to an operation of the
  *      CRUD (Create, Read, Update, Delete) acronym.
  *     - each objects are sent through the network in XMl, JSON, HTML, ... .
  *     - there is no client context being stored between REST operations (ie : REST is stateless)
  *
  *  Example : <br/>
  *  Lets say that a server manages a collection of Point (id, x, y)
  *  In order to "read" the point of id 1 we would use :
  * \code
  *    GET http://server.com/rest/point/1
  * \endcode
  *  This HTTP operation will answer (in JSON) :
  * \code
  *    "point" : {
  *         "id" : 1
  *         "x" : 5678,
  *         "y" : 1234,
  *     }
  * \endcode
  *  In XML :
  *  \code
  *     <point>
  *         <id>1</id>
  *         <x>5678</x>
  *         <y>1234</y>
  *     </point>
  *  \endcode
  *
  * With a post method the client will have to send XML or JSON representing an object instance to a specific URL :
  *
  *\code
  *    POST http://server.com/rest/point
  *    <point>
  *         <id>2</id>
  *         <x>34</y>
  *         <y>12</y>
  *    </point>
  * \endcode
  *
  * \section section2PageRest A Look at the server side code.
  *
  *  On the server side we use RESTeasy framework to create our resource. Folowing is a resource
  *  named "point" with two possible methods :
  *     - the first, of type GET that, returns a Point.
  *     - the second, unmarshall an XML or JSON Point and saves it to the data base.
  *
  *  \code
@Path("point")
public class BookResource {

        @GET
        @Path("/{id}")
        @Produces({MediaType.APPLICATION_JSON, MediaType.APPLICATION_XML})
        @Mapped(namespaceMap = { @XmlNsMap(namespace = "http://www.w3.org/2001/XMLSchema-instance", jsonName = "instance") })
        public Point getPoint(@PathParam("id") int id) {
            Point p = new Point(1, 1234, 5678);
            return p;
        }

        @POST
        @Consumes({MediaType.APPLICATION_JSON, MediaType.APPLICATION_XML})
        @Mapped(namespaceMap = { @XmlNsMap(namespace = "http://www.w3.org/2001/XMLSchema-instance", jsonName = "instance") })
        public void addPoint(final Point p) {
            // Save the new p to the data base.
        }
    // ....
}

  *  \endcode
  * Some more details :<br/>
  *  - The Point p can be retrieve from the server by the HTTP operation GET http://server/rest/point/1
  *  As you can see, the method getPoint is also annotated by "@Produces". This annotation alows to tell the application that the
  *  Point p can be send to the client either in XML or in JSON format. The format is chosen by the client when providing the "Accept"
  *  http header :
  *   - Accept:application/json : the server will send JSON.
  *   - Accept:application/xml : the server will send XML.
  *
  *  - A Point can be added to the server by sending it as an XML or JSON (depending on the "Content-Type" http header field) feed,
  *      to the POST method addPoint.
  *
  *
  *  In order to "Marshall" the type Point to XML or JSON we use JAX-B annotations. As you can see on the folowing example this is a pretty
  *  simple task :
  * \code
@XmlRootElement(name="point")
@XmlAccessorType(XmlAccessType.FIELD)
public class Point implements Serializable {

        @XmlElement
        private Integer id;

        @XmlElement
        private Double x;

        @XmlElement
        private Double y;


        public Point(Integer pId, Double px, Double py)
        {
                id = pId;
                x= px;
                y= py;
        }
        // ....
}
  * \endcode
  *
  *
  * \section section3PageRest The DoQuBook REST client api.
  *
  *  On the client side, we now want to receive the JSON streams and convert them into objects.
  *  This can be done using the class RESTObjectManager<T, TPtr> where T is the Marshallable type and TPtr
  *  the MarshallablePtr type (Point and PointPtr). <br/>
  *  Lets take a look at how this class can be used : <br/>
  * - First the creation and configuration of an instance of RESTObjectManager can be done using the folowing code :
  * \code
  * // Point extends Marshallable and typedef QSharedPointer<Point> PointPtr;
  * // Creation
  * RESTObjectManager<Point, PointPtr> pointManager (WebClient::JSON)
  * // Provide resource path to each HTTP method :
  * // note : urls can take parameters corresponding to the object property name
  * pointManager.addRessourcePath(WebClient::GET, "http://example.com/App/rest/point/(id)");
  * pointManager.addRessourcePath(WebClient::POST, "http://example.com/App/rest/point");
  * pointManager.addRessourcePath(WebClient::PUT, "http://example.com/App/rest/point");
  * pointManager.addRessourcePath(WebClient::DEL, "http://example.com/App/rest/point/(id)");
  * \endcode
  *  As you can see, each method has an URL. This means that for a get method, the RESTObjectManager will use the url
  *  http://example.com/App/rest/point/(id). The (id) part of the url will be replace by the value given as parameter
  *  in the get method, but it will also replaced by the object's property (property named "id") value when using POST,
  *  PUT or even DEL.
  *
  * - The get method :
  * \code
  *  QMap<QString, QString> params;
  *  params["id"] = QString::number (id);
  *
  *  PointPtr p = pointManager.get(params);
  * \endcode
  * - Post an object :
  * \code
  *  // let p a PointPtr
  *  pointManager.post (p);
  * \endcode
  * - Put an object :
  * \code
  *  // let p a PointPtr
  *  pointManager.put (p);
  * \endcode
  *
  * - Del an object :
  *  - Based on the object properties
  *  \code
  *  // let p a PointPtr
  *  pointManager.del (p);
  * \endcode
  *  - Or based on a parameter map :
  * \code
  *  QMap<QString, QString> params;
  *  params["id"] = QString::number (p->getId());
  *  pointManager.del (params);
  * \endcode
  *
  * Now that we have seen how to send or retreive objects to or from the server,
  *  we have to "annotate" the C++ Point class a bit like we did on the Java part
  * (with JAX-B). Sadly, standard C++ by it's own does not provide annotations, nor
  * reflection or introspection (witch is used to serialize the annotated java class to
  * XML or JSON). Thats where Qt comes and saves us. Qt has pre-compilation phase (moc),
  * that alows a "BASIC" reflection on what is called "properties". Qt Properties are mapped
  * to classes attributes using getters and setters.<br/>
  * Folowing, is the C++ Point class.
  * \code
  *
  *  class Point;
  *  typedef QSharedPointer<Point> PointPtr;
  *
  * class DoQuBookSHARED_EXPORT Point : public SimpleMarshallable
  * {
  *      Q_OBJECT
  *      Q_PROPERTY (int id READ getId WRITE setId)
  *      Q_PROPERTY (float x READ getX WRITE setX)
  *      Q_PROPERTY (float y READ getY WRITE setY)
  *
  *  public :
  *
  *    MARSHALLABLE_CREATE_FONCTIONS(Point)
  *
  *    float getX() const {return _x;}
  *    float getY() const {return _y;}
  *    float getId() const {return m_id;}
  *    void setX (float x) {_x = x;}
  *    void setY (float y) {_y = y;}
  *    void setId (float id) {m_id = id;}
  *
  *   // Constructors other methods
  *
  *  private :
  *       float _x;
  *       float _y;
  *       int m_id;
  * };
  * \endcode
  * You can think that in a way the macro Q_PROPERTY is a kind of annotation.
  * A complete page explaining the marshalling module of DoQuBook details how to create Marshallable object : <br/>
  *  see \ref pageMarshallable <br/>
  *
  * \section section4PageRest Conclusion.
  *
  * The class RESTObjectManager offers a create way to acces REST resources. But it may do not corespond to your needs.
  * If you have specific request please look at the folowing classes or even at the QtNetwork module :
  *   - MarshallableFormater
  *   - MarshallableFormaterJSon
  *   - WebClient
  */
namespace DoQuBook
{
    /*!
      * \brief This class can be used to manage a REST Object of type
      * Marshallable <br/>
      *
      * Tutorial on page \ref pageREST. <br/>
      *
      * This class takes two parameters T (Marshallable) and TPtr (MarshallablePtr) <br/>
      * <u>Usage : </u> <br/>
      * - Configuration :
      * \code
      * // Point extends Marshallable and typedef QSharedPointer<Point> PointPtr;
      * // Creation
      * RESTObjectManager<Point, PointPtr> pointManager (WebClient::JSON)
      * // Provide resource path to each HTTP method :
      * // note : urls can take parameters corresponding to the object property name
      * pointManager.addRessourcePath(WebClient::GET, "http://example.com/App/rest/point/(id)");
      * pointManager.addRessourcePath(WebClient::POST, "http://example.com/App/rest/point");
      * pointManager.addRessourcePath(WebClient::PUT, "http://example.com/App/rest/point");
      * pointManager.addRessourcePath(WebClient::DEL, "http://example.com/App/rest/point/(id)");
      * \endcode
      * - Getting an object :
      * \code
      *  QMap<QString, QString> params;
      *  params["id"] = QString::number (id);
      *
      *  PointPtr p = pointManager.get(params);
      * \endcode
      * - Post an object :
      * \code
      *  // let p a PointPtr
      *  pointManager.post (p);
      * \endcode
      * - Put an object :
      * \code
      *  // let p a PointPtr
      *  pointManager.put (p);
      * \endcode
      *
      * - Del an object :
      *  - Based on the object properties
      *  \code
      *  // let p a PointPtr
      *  pointManager.del (p);
      * \endcode
      *  - Or based on a parameter map :
      * \code
      *  QMap<QString, QString> params;
      *  params["id"] = QString::number (p->getId());
      *  pointManager.del (params);
      * \endcode
      */
    template <class T>
    class RESTObjectManager
    {
        public:
            RESTObjectManager (WebClient::ContentType type, MarshallableFormaterPtr marshaller)
                : m_formater (marshaller),
                  m_contentType (type) {}

            void addRessourcePath (WebClient::Method method, const QString & string) {
                m_routPaths[method] = string;
            }

            // Methods using the route table
            QSharedPointer<T> get (const QMap<QString, QString> &params) {
                WebClient wc;
                QUrl u = urlFromParams (params, WebClient::GET);

                wc.setContentType (m_contentType);
                QByteArray array = wc.sendSynchronous (u, WebClient::GET, "");

                if (array.isEmpty() ) {
                    return QSharedPointer<T>();
                }
                else {
                    return fromBytes<T> (array);
                }
            }

            template <class U>
            QSharedPointer<U> post (QSharedPointer<T> object) {
                QUrl url = urlFromObject (object, WebClient::POST);
                QByteArray jsonBytes = toBytes (object);

                WebClient wc;
                wc.setContentType (m_contentType);

                QByteArray array = wc.sendSynchronous (url, WebClient::POST, jsonBytes);

                return fromBytes<U> (array);
            }

            QSharedPointer<T> post (QSharedPointer<T> object) {
                return post<T> (object);
            }

            QSharedPointer<T> post (QSharedPointer<T> object, const QMap<QString, QString> &params) {
                QUrl url = urlFromParams (params, WebClient::POST);
                QByteArray jsonBytes = toBytes (object);

                WebClient wc;
                wc.setContentType (m_contentType);

                QByteArray array = wc.sendSynchronous (url, WebClient::POST, jsonBytes);

                return fromBytes<T> (array);

            }

            QByteArray postAndReturnByteArray (QSharedPointer<T> object, const QMap<QString, QString> &params) {
                QUrl url = urlFromParams (params, WebClient::POST);
                QByteArray jsonBytes = toBytes (object);

                WebClient wc;
                wc.setContentType (m_contentType);

                QByteArray array = wc.sendSynchronous (url, WebClient::POST, jsonBytes);
                return array;
            }

            template <class U>
            QSharedPointer<U> put (QSharedPointer<T> object) {
                QUrl url = urlFromObject (object, WebClient::PUT);
                QByteArray jsonBytes = toBytes (object);

                WebClient wc;
                wc.setContentType (m_contentType);
                QByteArray array = wc.sendSynchronous (url, WebClient::PUT, jsonBytes);

                return fromBytes<U> (array);
            }

            QSharedPointer<T> put (QSharedPointer<T> object) {
                return put<T> (object);
            }

            QSharedPointer<T> put (QSharedPointer<T> object, const QMap<QString, QString> &params) {
                QUrl url = urlFromParams (params, WebClient::POST);
                QByteArray jsonBytes = toBytes (object);

                WebClient wc;
                wc.setContentType (m_contentType);

                QByteArray array = wc.sendSynchronous (url, WebClient::PUT, jsonBytes);
                return fromBytes<T> (array);

            }

            void del (QSharedPointer<T> object) {
                QUrl url = urlFromObject (object, WebClient::DEL);
                WebClient wc;
                wc.setContentType (m_contentType);
                QByteArray jsonBytes = toBytes (object);
                QByteArray array = wc.sendSynchronous (url, WebClient::DEL, jsonBytes);
            }

            void del (const QMap<QString, QString> &params) {
                QUrl url = urlFromParams (params, WebClient::DEL);
                WebClient wc;
                wc.setContentType (m_contentType);
                QByteArray array = wc.sendSynchronous (url, WebClient::DEL, "");
            }

            // Methods with specific url
            /*
            T get (const QUrl & resourcePath);
            bool post (const QUrl & resourcePath, const T &object);
            bool put (const QUrl & resourcePath, const T &object);
            bool del (const QUrl & resourcePath, const T &object);
            bool del (const QUrl & resourcePath);
            */

        private :

            QUrl urlFromParams (const QMap<QString, QString> &params, WebClient::Method method) {
                if (!m_routPaths.contains (method) ) {
                    throw MessageException ("No such HTTP method registered");
                }

                QString url = m_routPaths [method];
                QMapIterator<QString, QString> it (params);
                while (it.hasNext() ) {
                    it.next();
                    url.replace ("(" + it.key() + ")", it.value() );
                }
                return QUrl (url);
            }
            QUrl urlFromObject (QSharedPointer<T> object, WebClient::Method method) {
                QString url = m_routPaths [method];
                for (int i = 0; i < object->metaObject()->propertyCount(); i++) {
                    QMetaProperty prop = object->metaObject()->property (i);
                    QString propName (prop.name() );
                    if (prop.type() != QVariant::UserType) {
                        url.replace ("(" + propName + ")", prop.read (object.data() ).toString() );
                    }
                }
                return QUrl (url);
            }
            QByteArray toBytes (QSharedPointer<T> object) {
                QString string;
                (*m_formater).marshall (object , &string);
                return string.trimmed().toUtf8();
            }

            template <class U>
            QSharedPointer<U> fromBytes (const QByteArray & bytes) {
                QSharedPointer<U> b = U::create();
                if (!bytes.isEmpty() ) {
                    m_formater->unMarshall (bytes, b);
                }
                return b;
            }

        private :
            QMap<WebClient::Method, QString> m_routPaths;
            MarshallableFormaterPtr m_formater;
            WebClient::ContentType m_contentType;
    };

}

#endif // RESTOBJECTMANAGER_HPP
