#ifndef REGION_HPP
#define REGION_HPP

#include <DoQuBook/Point.hpp>
#include <QImage>
#include <vector>
#include <DoQuBook/Config.hpp>
#if QT_VERSION >= 0x050000
#include <QtConcurrent/QtConcurrent>
#endif
#include <QtCore>
#include <assert.h>

namespace DoQuBook
{

  class RegionPointProcessor;
  typedef QSharedPointer<RegionPointProcessor> RegionPointProcessorPtr;

  class RegionPointProcessor
  {
    public :
      RegionPointProcessor() {}
      virtual ~RegionPointProcessor() {}

      // Called before any use
      virtual void init() {}
      // Called after any use
      virtual void clear() {}
      // Can be used for memoizing
      virtual QString name() {return QString();}


      virtual void processPoint (int x, int y) = 0;
      virtual void newRegion() = 0;
      virtual void endRegion() = 0;
      virtual void boundingBoxOfRegion (int x1, int y1, int x2, int y3) = 0;

      /// can be used as filters
      virtual void area (double a) = 0;
      virtual void perimeter (double p) = 0;
  };


  class Region;
  typedef QSharedPointer<Region> RegionPtr;
  typedef QList<RegionPtr> RegionPtrList;

  class Region : public Marshallable
  {
      Q_OBJECT
    public :
      struct ScanLine {
          QPointF start;
          QPointF end;
      };

      typedef QList<ScanLine> ScanLineList;
      typedef QSharedPointer<ScanLineList> ScanLineListPtr;

      enum DrawingMode { BBox, Contour, Domain };

    public :
      virtual QVariantMap serialize()  {
        QVariantMap region;

        QVariantList lines;
        foreach (const ScanLine &l, *(_scanLines)) {
            QVariantMap line;
            QVariantMap p1;
            p1["x"] = l.start.x();
            p1["y"] = l.start.y();

            QVariantMap p2;
            p2["x"] = l.end.x();
            p2["y"] = l.end.y();

            line["start"] = p1;
            line["end"] = p2;
            lines.append(line);
        }

        region["scanLines"] = lines;
        return region;
      }

      // TODO : back from QVariantMap
      virtual void from (QVariantMap /*map*/) {}
      virtual MarshallablePtr createFromVariantMap (QVariantMap /*m*/) {return MarshallablePtr();}

      MARSHALLABLE_CREATE_FONCTIONS (Region)

    public:
      Region();
      Region (const QString & label);
      Region (const Region::ScanLineListPtr &scanLines);

      void addScanLine(const ScanLine &line);


      QImage thumbnail (const QImage & image);

      void draw (QImage& image, const QRgb& color, DrawingMode drawingMode = Domain);

      void iterateOnDomain(const RegionPointProcessorPtr &processor);

      QString getLabel();
      void setLabel (const QString & label);



      //////// Will maybe memoized
      QList<PointPtr> getContour();
      // All Points
      QList<QPointF> domain();

      QRectF  boundingBox();
      void setBoundingBox (const PointPtr & p1, const PointPtr & p2);

      double area();
      void setArea (double a);

      double perimeter();
      void setPerimeter (double p);

    private :
      Region::ScanLineListPtr _scanLines;
      QString _label;

      /// Memoized field;
      QVariantMap memoizedPointProcessorResults;
      QList<DoQuBook::PointPtr> _domain;
      QRectF _bbox;
      double _area;
      double _perimeter;
  };

  class RegionList;
  typedef QSharedPointer<RegionList> RegionListPtr;

  class DoQuBookSHARED_EXPORT RegionList : public SimpleMarshallable
  {
          Q_OBJECT
          Q_PROPERTY (QList<RegionPtr> regions READ getRegions WRITE setRegions)

      public :

          MARSHALLABLE_CREATE_FONCTIONS (RegionList)

          RegionList()
          {}
          QList<RegionPtr>  getRegions() {
              return _regions;
          }
          void setRegions (QList<RegionPtr> pRegions) {
              _regions= pRegions;
          }

          void addRegion(RegionPtr region) {
            _regions << region;
          }

      private :
          MARSHALLABLE_MLIST (QList, Region, regions)
          MARSHALLABLE_ACTIV_1LIST (regions, _regions)
          QList<RegionPtr> _regions;
  };



  class BoundingBoxProcessor : public RegionPointProcessor
  {
    public :
      BoundingBoxProcessor() {}
      virtual ~BoundingBoxProcessor() {}

      virtual void processPoint (int /*x*/, int /*y*/)
      {}

      virtual void newRegion()  {}
      virtual void endRegion() {}

      virtual void boundingBoxOfRegion (int x1, int y1, int x2, int y2) {
        QPointF topLeft (x1, y1);
        QPointF bottomRight (x2, y2);
        QRectF bbox (topLeft, bottomRight);
        process (bbox);
      }

      virtual void process (QRectF & bbox) = 0;

      virtual void area (double ) {};
      virtual void perimeter (double ) {};

  };

  class RegionPainter : public RegionPointProcessor
  {
    public :
      RegionPainter (QImage & image, const QRgb &color)
        : _image (image), _color(color), regionAccepted (true) {}
      RegionPainter (QImage & image)
        : _image (image), regionAccepted (true) {
        _color = qRgb(255,0,0);
      }
      virtual ~RegionPainter() {}

      virtual void processPoint (int x, int y) {
          _image.setPixel (x, y, _color);
      }

      virtual void newRegion() {
        /*int r = qrand() % ((255 + 1) - 0) + 0;
        int g = qrand() % ((255 + 1) - 0) + 0;
        int b = qrand() % ((255 + 1) - 0) + 0;
        _color = qRgb(r,g,b);
        */
      }
      virtual void endRegion() {}
      virtual void boundingBoxOfRegion (int /*x1*/, int /*y1*/, int /*x2*/, int /*y3*/) {}

      /// can be used as filters
      virtual void area (double /*a*/) {
        //regionAccepted = a >= 20;
      }
      virtual void perimeter (double /*p*/) {}

      void setColor(const QRgb &color) {
        _color = color;
      }

    protected :
      QImage & _image;
      QRgb _color;
      bool regionAccepted;
  };


  class DomainBuilder;
  typedef QSharedPointer<DomainBuilder> DomainBuilderPtr;

  class DomainBuilder : public RegionPointProcessor {
    public :
      DomainBuilder (QList<QPointF> &points)
        : _points (points) {}

      virtual void processPoint (int x, int y)  {
        _points.append(QPointF(x,y));
      }

      virtual void newRegion() {}
      virtual void endRegion() {}
      virtual void boundingBoxOfRegion (int /*x1*/, int /*y1*/, int /*x2*/, int /*y3*/) {}

      /// can be used as filters
      virtual void area (double /*a*/) {}
      virtual void perimeter (double /*p*/) {}

    private :
      QList<QPointF> &_points;

  };

  class RegionsExtractor;
  typedef QSharedPointer<RegionsExtractor> ScanLineExtractorPtr;


  class RegionsExtractor : public DoQuBook::RegionPointProcessor {

    public :
      RegionsExtractor()
      {
        regions = RegionList::create();
      }
      virtual ~RegionsExtractor() {}
      virtual void processPoint(int x, int y) {
        if(currentScanLine.start == QPoint()) {
          startNewLine(QPoint (x,y));
        }
        else {
          if(x != currentScanLine.end.x()+1) {
            region->addScanLine(currentScanLine);
            startNewLine(QPoint (x,y));
          }
          currentScanLine.end = QPoint (x,y);
        }
      }

      virtual void newRegion() {
          region = QSharedPointer<Region> (new Region);
      }

      virtual void endRegion()  {
        // Add the last scanline
        region->addScanLine(currentScanLine);
        regions->addRegion(region);
        startNewLine(QPoint());
      }

      virtual void boundingBoxOfRegion(int , int , int , int ) {}

      /// can be used as filters
      virtual void area(double ) {}
      virtual void perimeter(double ) {}

      RegionListPtr getRegions () {
        return regions;
      }

    private :
      void startNewLine(const QPoint &p) {
        currentScanLine = Region::ScanLine();
        currentScanLine.start = p;
        currentScanLine.end = p;
      }

    private :
      Region::ScanLine currentScanLine;
      RegionPtr region;
      RegionListPtr regions;
  };



  class MappedRegionsProcessor {
    public :


      MappedRegionsProcessor (QList<RegionPtr> &regions)
        : _regions (regions) {}

      struct processRegion
      {
          processRegion(const RegionPointProcessorPtr& processor)
            : _processor(processor)
          {}

          void operator()(const RegionPtr & region)
          {
            region->iterateOnDomain(_processor);
          }

          const RegionPointProcessorPtr& _processor;
      };

      void runProcessors(const RegionPointProcessorPtr& processor) {
        processRegion functor (processor);
        QFuture<void> preps = QtConcurrent::map(_regions, functor);
        preps.waitForFinished();
      }
      QList<RegionPtr> &_regions;
  };

  class MappedRegionsProcessors {
    public :
      MappedRegionsProcessors (QList<RegionPtr> &regions)
        : _regions (regions) {}

      struct processRegion
      {
          processRegion(QList<RegionPtr> &regions)
            : _regions(regions)
          {}

          void operator()(RegionPointProcessorPtr processor)
          {
            foreach (const RegionPtr &region, _regions) {
                region->iterateOnDomain(processor);
            }
          }

          QList<RegionPtr> &_regions;
      };

      void runProcessors(const QList<RegionPointProcessorPtr> & processors) {
        processRegion functor (_regions);
        QFuture<void> preps = QtConcurrent::map(processors, functor);
        preps.waitForFinished();
      }
      QList<RegionPtr> &_regions;
  };



}
#endif // REGION_HPP
