/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "SimpleMarshallable.hpp"
#include <DoQuBook/Config.hpp>

SimpleMarshallable::SimpleMarshallable (QObject * parent)
    : Marshallable (parent) {}


QMap<QString, SimpleMarshallable::ConverterPtr> SimpleMarshallable::converteurs()
{
    return QMap<QString, SimpleMarshallable::ConverterPtr>();
}


MarshallablePtr SimpleMarshallable::createFromVariantMap (QVariantMap m)
{
    if (m.contains ("@instance.type") ) {
        QString name =  m["@instance.type"].toString().toLower();
        QString myName = QString (this->metaObject()->className() ).toLower();
        QStringList namespaces = myName.split ("::");
        myName = namespaces.at (namespaces.length() - 1);

        if (myName == name) {
            return createPtr();
        }
        else {

            QList<MarshallablePtr> seeAlsoMarshallables = seeAlso();
            if (!seeAlsoMarshallables.empty() ) {
                foreach (MarshallablePtr marshal, seeAlsoMarshallables) {
                    MarshallablePtr p = marshal->createFromVariantMap (m);
                    if (!p.isNull() ) {
                        return p;
                    }
                }
            }
            else {
                return MarshallablePtr();
            }

        }
    }
    else {
        MarshallablePtr ptr = createPtr();
        return ptr;
    }

    return MarshallablePtr();
}




QStringList SimpleMarshallable::customProperties()
{
    return QStringList();
}

QVariantMap SimpleMarshallable::serialize()
{
    QVariantMap result;

    const QMetaObject * metaobject = this->metaObject();
    int count = metaobject->propertyCount();


    if (!ignoredProperties().contains ("@property.type") ) {
        result["@property.type"] = metaobject->className();
    }

    for (int i = 0; i < count; ++i) {
        QMetaProperty metaproperty = metaobject->property (i);
        const char * name = metaproperty.name();
        QString propName = name;
        QString mapKeyName = name;

        if (ignoredProperties().contains (QLatin1String (name) ) || (!metaproperty.isReadable() ) ) {
            continue;
        }

        QString methodName = "convert_" + propName;

        QVariant value;
        QVariantList list;
        QMap<QString, ConverterPtr> convs = converteurs();
        if (convs.contains (propName) ) {
            convs[propName]->to (&list);
            value = list;
        }
        else if (customProperties().contains (propName) ) {
            /*bool ok =*/ metaobject->invokeMethod (this, methodName.toStdString().c_str(), Qt::DirectConnection, Q_ARG (QVariantList *, &list) );
            value = list;
        }
        else {
            if (propName == "type") {
                mapKeyName = "@instance.type";
            }
            value = this->property (name);
        }

        result[mapKeyName] = value;
    }

    return result;
}

void SimpleMarshallable::from (QVariantMap map)
{
    const QMetaObject * metaobject = this->metaObject();

    QVariantMap::const_iterator iter;
    for (iter = map.constBegin(); iter != map.constEnd(); ++iter) {
        QString propName = iter.key();

        int pIdx = metaobject->indexOfProperty ( iter.key().toUtf8().constData() );
        if ( pIdx < 0 ) {
            continue;
        }
        QMetaProperty metaproperty = metaobject->property ( pIdx );
        QVariant::Type type = metaproperty.type();
        QVariant v ( iter.value() );

        QString methodName = "from_" + propName;
        QMap<QString, ConverterPtr> convs = converteurs();

        if (convs.contains (propName) ) {
            convs[propName]->from (v);
        }
        else if (customProperties().contains (propName) ) {
            metaobject->invokeMethod (this, methodName.toUtf8().constData(), Qt::DirectConnection, Q_ARG (QVariant, v) );
        }
        else {
            QString typeName (metaproperty.typeName () );
            if ( typeName == "QVariant") {
                metaproperty.write ( this, v );
            }
            if ( v.canConvert ( type ) ) {
                v.convert ( type );
                metaproperty.write ( this, v );
            }
        }
    }
}

QString SimpleMarshallable::getType()
{
    QString type = this->metaObject()->className();
    QStringList namespaces = type.split ("::");
    QString className = namespaces[namespaces.length() - 1];
    QString firstLetterLower = className.at (0).toLower();

    return firstLetterLower.append (className.mid (1) );
}

void SimpleMarshallable::setType (QString t)
{
    _marshallableType = t;
}

/*bool SimpleMarshallable::merge(MarshallablePtr other)
{
    const QMetaObject* metaobject = this->metaObject();
    const QMetaObject* othermetaobject = other->metaObject();
    if(metaobject->className() == othermetaobject->className()) {
        int count = metaobject->propertyCount();

        for (int i=0; i<count; ++i) {
            QMetaProperty metaproperty = metaobject->property (i);
            QString name = metaproperty.name();
            if(metaproperty.isWritable()
                    && name != "id"
                    && name != "objectName") {

                if(metaproperty.type() == QVariant::UserType
                   && QString(metaproperty.enclosingMetaObject()->superClass()->className()) == "SimpleMarshallable") {
                   // QList<MarshallablePtr> list;
                    const QMetaObject *o = metaproperty.enclosingMetaObject();
                    int k = 0;
                    for(int i = 0; i < o->methodCount(); i++) {
                        qDebug() << o->method(i).signature();
                    }
                }


                metaproperty.write(this, othermetaobject->property(i).read(other.data()));

            }
        }
    }
    return false;
}*/

