/*!
 * \file SimpleMarshallable.hpp
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#ifndef SIMPLEMARSHALLABLE_HPP
#define SIMPLEMARSHALLABLE_HPP

#include <DoQuBook/Marshallable.hpp>
#include <DoQuBook/DoQuBook_global.hpp>
#include <QDebug>



/// Use this macro for auto-deletion of the marshallable
#define MARSHALLABLE_DECLARE_POINTER(T) \
    class DoQuBookSHARED_EXPORT T; \
    typedef QSharedPointer<T> T ## Ptr; \
 
/// Use this macro with MARSHALLABLE_DECLARE_POINTER it provides static function to replace a private constructors (disabling classic pointers)
#define MARSHALLABLE_CREATE_FONCTIONS(T) \
    MarshallablePtr createPtr() \
    { \
        return T::create(); \
    } \
    \
    static T ## Ptr create() \
    { \
        T ## Ptr ptr (new T); \
        ptr->_this = ptr.toWeakRef(); \
        return ptr; \
    } \
    private : \
        QWeakPointer<T> _this; \
    public : \
 

/*!
  * \brief Implements a basic Marshallable
  *
  * \attention A SimpleMarshallable can only have ONE list porperty if using MARSHALLABLE_LIST_PROPERTY macro.
  */
class DoQuBookSHARED_EXPORT SimpleMarshallable : public Marshallable
{
        Q_OBJECT
    public :

        //MARSHALLABLE_CREATE_FONCTIONS (SimpleMarshallable)

        SimpleMarshallable (QObject * parent = 0);
        virtual ~SimpleMarshallable () {}

        virtual QStringList customProperties();

        class Converter;
        typedef QSharedPointer<Converter> ConverterPtr;

        class DoQuBookSHARED_EXPORT Converter
        {
            public :
                virtual ~Converter() {}
                virtual void to (QVariantList * list) = 0;
                virtual void from (QVariant v) = 0;
        };


        virtual QMap<QString, ConverterPtr> converteurs();
        virtual MarshallablePtr createFromVariantMap (QVariantMap m);



        template <class MarshalType, class RealType>
        void listFrom (QVariant v, MarshallablePtr factory, QList<MarshalType> &outList) {
            outList.clear();
            QVariantList list = v.toList();

            if (list.isEmpty() ) {
                QVariant variantTmp = v.toMap();
                QVariantMap variant = variantTmp.toMap();
                MarshallablePtr marshal = factory->createFromVariantMap (variant);
                MarshalType p = marshal.dynamicCast<RealType>();
                connect (p.data(), SIGNAL (modified() ), this, SLOT (emitModified() ) );
                if (variant.values().size() > 0) {
                    p->from (variant);
                    outList.append (p);
                }
            }

            foreach (QVariant var, list) {
                QVariantMap variant = var.toMap();
                MarshallablePtr marshal = factory->createFromVariantMap (variant);
                MarshalType p = marshal.dynamicCast<RealType>();
                connect (p.data(), SIGNAL (modified() ), this, SLOT (emitModified() ) );
                p->from (variant);
                outList.append (p);
            }
        }

        virtual QVariantMap serialize();

        void from (QVariantMap map);

        QString getType();

        void setType (QString t);

        //virtual bool merge (MarshallablePtr other);

    signals:
        void modified();

    public slots:
        virtual void emitModified() {
            emit modified();
        }

    protected:
        QString _marshallableType;
};

/// Use this macro for QList<MarshallablePtr> properties
#define MARSHALLABLE_MLIST(TList, TElem, PropName) \
    class PropName ## Converter: public SimpleMarshallable::Converter { \
    public : \
            PropName ## Converter(SimpleMarshallable& marshallable, TList< TElem ## Ptr > &list) \
            : _marshallable(marshallable) \
            , _list(list) \
        {} \
        virtual void to(QVariantList *list) \
        { \
            list->clear(); \
            foreach(TElem ## Ptr marshallable, _list) \
            { \
                *list << marshallable->serialize(); \
            } \
        } \
        virtual void from(QVariant v) \
        { \
            MarshallablePtr factory = MarshallablePtr (new TElem); \
            _list.clear(); \
            _marshallable.listFrom<TElem ## Ptr, TElem>(v, factory, _list); \
        } \
    private:\
        SimpleMarshallable &_marshallable;\
        TList< TElem ## Ptr > &_list;\
    };\
 
/*! When a converteur has been defined, it calls this macro to register a list property
  * \attention A SimpleMarshallable can only have ONE list property if using this to this implementation !
  * \def MARSHALLABLE_DECLARE_LIST_PROPERTY
  */
#define MARSHALLABLE_ACTIV_1LIST(PropName, Attr) \
    virtual QMap<QString, SimpleMarshallable::ConverterPtr> converteurs() \
    { \
        QMap<QString, SimpleMarshallable::ConverterPtr> converts; \
        SimpleMarshallable::ConverterPtr p (new PropName ## Converter(*this, Attr)); \
        converts[#PropName] = p;\
        return converts; \
    } \
 
#define MARSHALLABLE_ACTIV_2LIST(FirstPropName, FirstAttr, SecondPropName, SecondAttr) \
    virtual QMap<QString, SimpleMarshallable::ConverterPtr> converteurs() \
    { \
        QMap<QString, SimpleMarshallable::ConverterPtr> converts; \
        SimpleMarshallable::ConverterPtr p1 (new FirstPropName ## Converter(*this, FirstAttr)); \
        converts[#FirstPropName] = p1;\
        SimpleMarshallable::ConverterPtr p2 (new SecondPropName ## Converter(*this, SecondAttr)); \
        converts[#SecondPropName] = p2;\
        return converts; \
    } \
 

/// Use this macro for QList<QString> properties
#define MARSHALLABLE_QTLIST(TList, TElem, PropName) \
    class PropName ## Converter: public SimpleMarshallable::Converter { \
        public : \
        PropName ## Converter(Marshallable &/*marshall*/, TList<TElem> &list) \
            : _list(list) \
        {} \
        virtual void to(QVariantList *list) \
        { \
            list->clear(); \
            foreach(TElem m, _list) \
            { \
                *list << m; \
            } \
        } \
        virtual void from(QVariant v) \
        { \
            _list.clear(); \
            if(v.canConvert <QString>())  \
            { \
                TElem e = v.value<TElem>(); \
                _list.append(e); \
            } \
            else { \
            QVariantList list = v.toList(); \
            foreach(QVariant variant, list) { \
                TElem e = variant.value<TElem>(); \
                _list.append(e); \
            } \
            } \
        } \
        private: \
        TList<TElem> &_list; \
    }; \
 

/*! Used for polymorphic types see DoQuBook::Annotation and MARSHALLABLE_IMPL_SEE_ALSO
  * Put this in .hpp file
  */
#define MARSHALLABLE_DECLARE_SEE_ALSO() \
    virtual QList<MarshallablePtr> seeAlso(); \
 
/*! Used for polymorphic types see DoQuBook::Annotation and MARSHALLABLE_DECLARE_SEE_ALSO
  * Put this in .cpp file
  */
#define MARSHALLABLE_IMPL_SEE_ALSO(T, TA, TB) \
    QList<MarshallablePtr> T::seeAlso() \
{ \
    QList<MarshallablePtr> factos; \
    factos.append(TA::create()); \
    factos.append(TB::create()); \
    return factos; \
    }\
 
/*! Used for polymorphic types see DoQuBook::Annotation and MARSHALLABLE_DECLARE_SEE_ALSO
  * Put this in .hpp of a sub-class
  */
#define MARSHALLABLE_EXTENDS(T) \
    virtual QList<MarshallablePtr> seeAlso() \
{ \
    return QList<MarshallablePtr>(); \
} \
 




#endif // SIMPLEMARSHALLABLE_HPP
