/*!
 * \brief This file is part of DoQuBook Framework
 *
 * Main author :  Rabeux Vincent <rabeux@labri.fr>
 * from LaBRI, University of Bordeaux 1 - Sud Ouest
 *
 * \details DoQuBook is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * \version RC-0.0.2
 * \date    2011
 *
 * Authors and main contributors :
 * \author Rabeux Vincent <rabeux@labri.fr>
 * \author Maxime Daymard <mr.xyz@hotmail.fr>
 * \author Jean-Philippe Domenger <domenger@labri.fr>
 * \author Nicholas Journet <journet@labri.fr>
 *
 * \copyright GNU LGPL.
 * Copyright (C) 2003-2011 <LaBRI>. All Rights Reserved.
 * Web: <www.labri.fr>
 *
 * DoQuBook is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#ifndef SINGLETON_HPP
#define SINGLETON_HPP
#include <DoQuBook/DoQuBook_global.hpp>

namespace DoQuBook
{

    /**
      * \brief This class defines a templated Singleton.
      *  http://fr.wikipedia.org/wiki/Singleton_(patron_de_conception)
      *
      * \author Vincent Rabeux
      *
      * Usage :
      * \code
        class UniqueObject : public DoQuMeasures::Singleton<UniqueObject> {
            friend class DoQuMeasures::Singleton<UniqueObject>;
            public :
                void function(){ ... }
            private :
                UniqueObject(){}
        };
        \endcode
      *
      * How to use UniqueObject ?
      * \code
        ...
            UniqueObject::instance()->function();
        ...
        \endcode
      */
    template <class T>
    class Singleton
    {
        public:

            static T * instance() {
                if (! m_instance) {
                    m_instance = new T;
                }
                return m_instance;
            }

            static T & instanceRef() {
                if (! m_instance) {
                    m_instance = new T;
                }
                return (*m_instance);
            }

            static void destroy() {
                if (m_instance) {
                    delete m_instance;
                    m_instance = 0;
                }
            }

        protected:
            Singleton() {}
            virtual ~Singleton() {}

        private :
            static T * m_instance;
    };

}
template <class T>
T * DoQuBook::Singleton<T>::m_instance = 0;


#endif // SINGLETON_HPP
