#include "WebClient.hpp"
#include <QtNetwork>
#include <DoQuBook/Logger.hpp>
#include <assert.h>
#include <DoQuBook/ApplicationContext.hpp>

#ifdef ZLIB
#include <zlib.h>
#endif // ZLIB


using namespace DoQuBook;

WebClient::WebClient (QObject * parent) :
    QObject (parent)
    , m_contentType (JSON)
{
}

QByteArray WebClient::sendSynchronous  (const QUrl & url, Method m, const QByteArray & content)
{
    return sendSynchronous (url, m, content, "");
}

QByteArray DoQuBook::WebClient::postJpegFile (const QUrl & url, const QByteArray & content, const QString & fileName)
{
    //setContentType(JPEG_MULTIPART);
    return sendSynchronous (url, POST, content, fileName);
}


QByteArray WebClient::sendSynchronous (const QUrl & url, WebClient::Method method, const QByteArray & content, const QString & fileName)
{
    //TODO: ALL DATAS ARE LOADED IN MEMORY !!!
    // We should stream it with QIODevice http://qt-project.org/doc/qt-4.8/qnetworkaccessmanager.html#post
    SERVER_INFO ("");
    SERVER_INFO ("----------------------------------------------------------------------------------------------");

    QByteArray datasToBeSent;
    QNetworkRequest requete (url);

    if (m_contentType == JPEG_MULTIPART
        || m_contentType == PNG_MULTIPART
        || m_contentType == TIFF_MULTIPART
        || m_contentType == XML_MULTIPART
        || m_contentType == ZIP_MULTIPART) {
        if (method == POST || method == PUT ) {
            QByteArray bound = boundary();
            datasToBeSent = createFormDataBytes (content, bound, fileName);
            prepareRequestForMultiPart (requete, bound, datasToBeSent);
        }
    }
    else {
        requete.setHeader (QNetworkRequest::ContentTypeHeader, contentTypeToString() );
        requete.setRawHeader ("Accept", contentTypeToString().toUtf8() );
#ifdef ZLIB
        requete.setRawHeader ("Accept-Encoding", "gzip");
        if (content.size() > 0) {
            requete.setRawHeader ("Content-Encoding", "gzip");
            datasToBeSent = gCompress (content);
        }
        else {
            datasToBeSent = content;
        }
#endif
#ifndef ZLIB
        datasToBeSent = content;
#endif
    }

    QNetworkAccessManager m;

    QEventLoop loop;
    connect (&m, SIGNAL (finished (QNetworkReply *) ), &loop, SLOT (quit() ) );

    QNetworkReply * r = NULL;
    if (method == POST) {
        SERVER_INFO ("POST : " + url.toString() );
        r = m.post (requete, datasToBeSent);
    }
    if (method == GET) {
        if (ApplicationContext::instance()->cacheActivated() ) {
            QNetworkDiskCache * cache = ApplicationContext::instance()->getCache();
            m.setCache (cache);
        }
        SERVER_INFO ("GET : " + url.toString() );
        r = m.get (requete);
    }
    if (method == PUT) {
        SERVER_INFO ("PUT : " + url.toString() );
        r = m.put (requete, datasToBeSent);
    }
    if (method == DEL) {
        SERVER_INFO ("DEL : " + url.toString() );
        r = m.deleteResource (requete);
    }

    SERVER_INFO ("Content : " + content);

    if (!r->bytesAvailable() ) {
        loop.exec();
    }

    QVariant statusCode = r->attribute ( QNetworkRequest::HttpStatusCodeAttribute );
    if ( !statusCode.isValid() ) {
        SERVER_INFO ("statusCode not valid");
    }
    else {

        int status = statusCode.toInt();
        SERVER_INFO ("Response Status Code : " + QString::number (status) );
        if ( status != 200 ) {
            QString reason = r->attribute ( QNetworkRequest::HttpReasonPhraseAttribute ).toString();
            SERVER_INFO ("Error reason :" + reason);
            SERVER_INFO ("Response : " + r->readAll());
        }
        if (status == 503 || r->error() == QNetworkReply::ProtocolUnknownError) {
            QString reason = r->attribute ( QNetworkRequest::HttpReasonPhraseAttribute ).toString();
            SERVER_INFO ("Error reason :" + reason);
        }
        if (status == 597) {
            SERVER_INFO ("Server responded with code 597");
            SERVER_INFO ("Need to register ");
            throw NeedToRegister (url);
        }
        else {
            if (r->error() == QNetworkReply::HostNotFoundError) {
                throw InternetConnectionException (url);
            }

            if (r->error() == QNetworkReply::ContentNotFoundError) {
                throw ContentNotFoundException (url);
            }

            if (r->error() != QNetworkReply::NoError) {
                throw NetworkConnectionException (r->errorString() + " on url :" + url.toString() );
            }
        }
    }

    QByteArray array = r->readAll();
    if (statusCode.toInt() == 201 && array.size() <= 0) {
        array = r->header (QNetworkRequest::LocationHeader).toByteArray();
    }

    QVariant fromCache = r->attribute (QNetworkRequest::SourceIsFromCacheAttribute);
    SERVER_INFO ("USING CACHE = " + QString::number (fromCache.toBool() ) );
    SERVER_INFO ("RESPONSE =" + array);

#ifdef ZLIB
    if (r->rawHeader ("Content-Encoding") == "gzip") {
        QByteArray unCompressed = gUncompress (array);

        delete r;

        return unCompressed;
    }
#endif

    delete r;


    SERVER_INFO ( "----------------------------------------------------------------------------------------------");

    return array;
}

QString WebClient::contentTypeToString()
{
    if (m_contentType == JSON) {
        return "application/json";
    }
    else if (m_contentType == XML) {
        return "application/xml";
    }
    else {
        assert (false);
        return "";
    }
}

void DoQuBook::WebClient::setContentType (DoQuBook::WebClient::ContentType t)
{
    m_contentType = t;
}

void DoQuBook::WebClient::prepareRequestForMultiPart (QNetworkRequest & request,
        const QByteArray & boundary,
        const QByteArray & data)
{
    request.setRawHeader ("Content-Type", "multipart/form-data; boundary=" + boundary);
    request.setRawHeader ("Content-Length", QString::number (data.size() ).toUtf8() );
}

QByteArray DoQuBook::WebClient::boundary()
{
    QByteArray boundary = "87142694621188";
    return boundary;
}

QByteArray DoQuBook::WebClient::createFormDataBytes (const QByteArray & bytes,
        const QByteArray & boundary,
        const QString & fileName)
{
    QByteArray data = "--" + boundary + "\r\n";
    data += "Content-Disposition: form-data; name=\"upload\"; filename=\"" + fileName + "\";\r\n";
    if (m_contentType == JPEG_MULTIPART) {
        data += "Content-Type: image/jpeg\r\n\r\n" + bytes + "\r\n";
    }
    else  if (m_contentType == PNG_MULTIPART) {
        data += "Content-Type: image/png\r\n\r\n" + bytes + "\r\n";
    }
    else if (m_contentType == TIFF_MULTIPART) {
        data += "Content-Type: image/tiff\r\n\r\n" + bytes + "\r\n";
    }
    else if (m_contentType == XML_MULTIPART) {
        data += "Content-Type: application/xml\r\n\r\n" + bytes + "\r\n";
    }
    else if (m_contentType == ZIP_MULTIPART) {
        data += "Content-Type: application/zip\r\n\r\n" + bytes + "\r\n";
    }
    else { // default is JPEG
        data += "Content-Type: image/jpeg\r\n\r\n" + bytes + "\r\n";
    }

    data += "--" + boundary + "--\r\n";
    return data;
}

// TODO : fix
bool DoQuBook::WebClient::isAvailable (const QUrl & /*urlbroken*/)
{
    /*qDebug() << urlbroken;
    QUrl url (urlbroken.toString() );
    QNetworkAccessManager * manager = new QNetworkAccessManager;

    QEventLoop * loop = new QEventLoop;
    connect (manager, SIGNAL (finished (QNetworkReply *) ), loop, SLOT(quit()) );

    QNetworkReply * r = manager->head (QNetworkRequest (url) );

    loop->exec();

    bool ret = (r->error() == QNetworkReply::NoError);
    delete manager;
    delete loop;

    return ret;
    */
    return true;
}



#ifdef ZLIB
QByteArray WebClient::gUncompress (const QByteArray & d)
{
    if (data.size() <= 4) {
        qWarning ("gUncompress: Input data is truncated");
        return QByteArray();
    }

    QByteArray result;

    int ret;
    z_stream strm;
    static const int CHUNK_SIZE = 1024;
    char out[CHUNK_SIZE];

    /* allocate inflate state */
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    strm.avail_in = data.size();
    strm.next_in = (Bytef *) (data.data() );

    ret = inflateInit2 (&strm, 15 +  32); // gzip decoding
    if (ret != Z_OK) {
        return QByteArray();
    }

    // run inflate()
    do {
        strm.avail_out = CHUNK_SIZE;
        strm.next_out = (Bytef *) (out);

        ret = inflate (&strm, Z_NO_FLUSH);
        Q_ASSERT (ret != Z_STREAM_ERROR); // state not clobbered

        switch (ret) {
            case Z_NEED_DICT:
                ret = Z_DATA_ERROR;     // and fall through
            case Z_DATA_ERROR:
            case Z_MEM_ERROR:
                (void) inflateEnd (&strm);
                return QByteArray();
        }

        result.append (out, CHUNK_SIZE - strm.avail_out);
    }
    while (strm.avail_out == 0);

    // clean up and return
    inflateEnd (&strm);
    return result;
}

QByteArray WebClient::gCompress (const QByteArray & data)
{
    if (data.size() <= 1) {
        qWarning ("gCompress: Input data is truncated");
        return QByteArray();
    }

    QByteArray result;

    int ret;
    //unsigned have;
    z_stream strm;
    static const int CHUNK_SIZE = 1024;

    char out[CHUNK_SIZE];

    /* allocate inflate state */
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    strm.total_out = 0;
    strm.avail_in = data.size();
    strm.next_in = (Bytef *) (data.data() );

    ret = deflateInit2 (&strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED, (15 + 16), 8, Z_DEFAULT_STRATEGY); // gzip encoding
    //ret = deflateInit(&strm, Z_DEFAULT_COMPRESSION);
    if (ret != Z_OK) {
        qCritical() << "Error compressing data";
        return QByteArray();
    }

    /* run inflate() on input until output buffer not full */
    int deflateStatus;
    do {
        strm.avail_out = CHUNK_SIZE;
        strm.next_out = (Bytef *) (out);

        deflateStatus = deflate (&strm, Z_FINISH);

        Q_ASSERT (ret != Z_STREAM_ERROR);
        result.append (out, CHUNK_SIZE - strm.avail_out);

    }
    while ( deflateStatus == Z_OK );

    /* clean up and return */
    deflateEnd (&strm);
    return result;
}

#endif


void HttpCallManager::run()
{
    if (!m) {
        initAll();
    }


    SERVER_INFO("Listen to event on url"+_url);
    QNetworkRequest * requete = new QNetworkRequest (_url);
    requete->setHeader (QNetworkRequest::ContentTypeHeader, "application/json");
    requete->setRawHeader ("Accept", "application/json");

    m->get (*requete);
}

#ifdef HAVE_CPP11
#include <QtConcurrent/QtConcurrent>

std::future<QByteArray> AsyncWebClientHelper::get(const QString& url, std::function<void (const QByteArray&)> onSuccess )
{
  std::future<QByteArray>  f = p.get_future();

  QtConcurrent::run([=]() {
    WebClient wc;
    QByteArray result = wc.sendSynchronous(url, WebClient::GET, "");
    onSuccess(result);
    //p.set_value(result);
  });

  return f;
}
#endif
