#include <QtTest>
#include <DoQuBook/AnnotationLocal.hpp>
#include <DoQuBook/Page.hpp>
#include "testAnnotation.hpp"
#include <assert.h>
#include <QDebug>
#include <QObject>
#include <QVariant>
#include <QPolygon>
#include <DoQuBook/Book.hpp>
#include <QTextStream>
#include <DoQuBook/MarshallableFormater.hpp>
#include <Utils.hpp>
#include <DoQuBook/RESTObjectManager.hpp>
#include <DoQuBook/PluginLoader.hpp>

using namespace DoQuBook;

namespace Tests {

class TestAnnotationLocal : public QObject {

    Q_OBJECT

public :
    static const QString ANNOTATION_NAME;
    static const QString REAL_JSON_PATH;

    TestAnnotationLocal ()
        :_annotation(NULL)
    {}

    ~TestAnnotationLocal()
    {
        if(!_annotation.isNull())
            _annotation.clear();
    }


private slots:

    void testAddAnnotatedPolygons ();
    void testPolygonNoDuplication ();


    void testEqual();
    void testMarshalling ();
    void testMerge ();
    void testMerge_matchingPolygons();
    void testMerge_mergePolygonsFromOtherAnnotation();


    /**
         * Called before each test function is executed.
        */
    void init();

    /**
         * Called after every test function.
        */
    void cleanup();



protected :
    AnnotationLocalPtr _annotation;
};

}

using namespace Tests;

const QString TestAnnotationLocal::ANNOTATION_NAME = "local bleed through";
const QString TestAnnotationLocal::REAL_JSON_PATH = ":/testsdatas/annlocal";

void TestAnnotationLocal::init()
{
    _annotation = AnnotationLocal::create();
    _annotation->setName(ANNOTATION_NAME);
    assert(!_annotation.isNull());
}

void TestAnnotationLocal::cleanup()
{
    _annotation.clear();
}


void TestAnnotationLocal::testAddAnnotatedPolygons()
{
    AnnotatedPolygonPtr polygon = createPolygon();

    _annotation->addAnnotatedPolygon(polygon);


    /// TEST FOR ADD (DIFF points, same value, DIFF Ptr, DIFF ID != -1) = YES ADD
    for(int i = _annotation->getAnnotatedPolygons().size(); i <= 3; i++) {
        AnnotatedPolygonPtr poly2 = createPolygon();
        PointPtr p = poly2->getPoint(0);
        p->setX(p->getX()+i);
        poly2->setId(i);

        int oldSize = _annotation->getAnnotatedPolygons().size();
        _annotation->addAnnotatedPolygon(poly2);

        QVERIFY (_annotation->getAnnotatedPolygons().size() == oldSize+1);
    }

    /// TEST FOR ADD (DIFF points, same value, DIFF Ptr, same ID == -1) = YES ADD
    for(int i = _annotation->getAnnotatedPolygons().size(); i <= 3; i++) {
        AnnotatedPolygonPtr poly2 = createPolygon();
        PointPtr p = poly2->getPoint(0);
        p->setX(p->getX()+i);

        int oldSize = _annotation->getAnnotatedPolygons().size();
        _annotation->addAnnotatedPolygon(poly2);

        QVERIFY (_annotation->getAnnotatedPolygons().size() == oldSize+1);
    }

    /// TEST FOR ADD (DIFF points, same value, same Ptr, DIFF ID != -1) = YES ADD
    for(int i = _annotation->getAnnotatedPolygons().size(); i <= 3; i++) {;
        polygon->setId(i);
        int oldSize = _annotation->getAnnotatedPolygons().size();
        _annotation->addAnnotatedPolygon(polygon);

        QVERIFY (_annotation->getAnnotatedPolygons().size() == oldSize+1);
    }
}

void TestAnnotationLocal::testPolygonNoDuplication()
{
    AnnotatedPolygonPtr polygon = createPolygon();

    _annotation->addAnnotatedPolygon(polygon);

    /// TEST FOR DUPLICATION (same points, same value, Same ID) = No Add
    for(int i = 0; i <= 3; i++) {
        _annotation->addAnnotatedPolygon(polygon);

        QVERIFY(_annotation->getAnnotatedPolygons().size() == 1);
    }

    foreach (AnnotatedPolygonPtr annotatedPolygon, _annotation->getAnnotatedPolygons()) {
        QVERIFY(annotatedPolygon == polygon);
    }

    /// TEST FOR DUPLICATION (same points, diff value, Same ID) = NO ADD
    for(int i = 0; i <= 3; i++) {
        polygon->setValue(polygon->getValue().toInt()+1);
        _annotation->addAnnotatedPolygon(polygon);

        QVERIFY(_annotation->getAnnotatedPolygons().size() == 1);
    }

    /// TEST FOR DUPLICATION (diff points, diff value, same Ptr, Same ID) = NO ADD
    for(int i = 0; i <= 3; i++) {
        PointPtr p = polygon->getPoint(0);
        p->setX(p->getX()+1);
        _annotation->addAnnotatedPolygon(polygon);
        QVERIFY (_annotation->getAnnotatedPolygons().size() == 1);
        /// Check if the point is modified;
        QVERIFY (p->getX() == polygon->getPoint(0)->getX());
        /// Check if the polygon is modified;
        QVERIFY(p->getX() == _annotation->getAnnotatedPolygon(0)->getPoint(0)->getX());
    }

    /// TEST FOR DUPLICATION (DIFF points, same value, DIFF Ptr, Same ID != -1) = No ADD
    AnnotatedPolygonPtr polyX = createPolygon();
    polyX->setId(1);
    _annotation->addAnnotatedPolygon(polyX);
    QVERIFY(!_annotation->findAnnotatedPolygonById(1).isNull());
    for(int i = _annotation->getAnnotatedPolygons().size(); i <= 3; i++) {
        AnnotatedPolygonPtr poly2 = createPolygon();
        PointPtr p = poly2->getPoint(0);
        p->setX(p->getX()+i);
        poly2->setId(1);

        int oldSize = _annotation->getAnnotatedPolygons().size();
        _annotation->addAnnotatedPolygon(poly2);

        QVERIFY (_annotation->getAnnotatedPolygons().size() == oldSize);
    }

    /// TEST FOR DUPLICATION (same points, same value, DIFF Ptr, DIFF ID == -1) = NO ADD
    // Add a clean (no points modification polygon) with id 1
    polyX = createPolygon();
    polyX->setId(1);
    _annotation->addAnnotatedPolygon(polyX);
    for(int i = _annotation->getAnnotatedPolygons().size(); i <= 3; i++) {
        AnnotatedPolygonPtr poly2 = createPolygon();

        int oldSize = _annotation->getAnnotatedPolygons().size();
        _annotation->addAnnotatedPolygon(poly2);

        // Chack if poly2 was replaced with the id of the existing polygon
        QVERIFY (poly2->getId() == 1);

        QVERIFY (_annotation->getAnnotatedPolygons().size() == oldSize);
    }
}

void TestAnnotationLocal::testEqual()
{
    AnnotationLocalPtr annLocal1 = createAnnotationLocal();
    AnnotationLocalPtr annLocal2 = createAnnotationLocal();
    QVERIFY(annLocal1->equals(*annLocal2));

    // Test that if we change name it is not equal
    {
        AnnotationLocalPtr annLocal3 = createAnnotationLocal();
        annLocal3->setName("DifferentName");
        QVERIFY(!annLocal1->equals(*annLocal3));
    }

    // Test that if we add a polygon point it is not equal
    {
        AnnotationLocalPtr annLocal3 = createAnnotationLocal();
        AnnotatedPolygonPtr poly = annLocal3->getAnnotatedPolygon(0);
        PointPtr p = Point::create();
        p->setX(1); p->setY(3);
        poly->addPoint(p);
        QVERIFY(!annLocal1->equals(*annLocal3));
    }

    // Test that if we change polygon point location it is not equal
    {
        AnnotationLocalPtr annLocal3 = createAnnotationLocal();
        AnnotatedPolygonPtr poly = annLocal3->getAnnotatedPolygon(0);
        poly->getPoint(0)->setX(32);
        QVERIFY(!annLocal1->equals(*annLocal3));
    }

    // Test that if we change polygon value it is not equal
    {
        AnnotationLocalPtr annLocal3 = createAnnotationLocal();
        AnnotatedPolygonPtr poly = annLocal3->getAnnotatedPolygon(0);
        poly->setValue(23);

        QVERIFY(!annLocal1->equals(*annLocal3));
    }

}

void TestAnnotationLocal::testMarshalling()
{
    AnnotationLocalPtr annLocal = createAnnotationLocal();

    MarshallableFormaterPtr jsonFormater = PluginLoader::instance()->getPlugin<MarshallableFormater>("labri.QJSONMarshallableFormater.0.0.1");
    QString marshalled;
    jsonFormater->marshall(annLocal, &marshalled);

    DoQuBook::AnnotationLocalPtr annLocal2 = AnnotationLocal::create();
    jsonFormater->unMarshall(QByteArray(marshalled.toStdString().c_str()), annLocal2);

    QVERIFY(annLocal->equals(*annLocal2));
}

void TestAnnotationLocal::testMerge()
{

    // Create an annLocal id = -1 and polygon id = -1;
    AnnotationLocalPtr annLocal = AnnotationLocal::create();
    annLocal->setName("annLocal");
    AnnotatedPolygonPtr polygon = createPolygon();
    polygon->setId(1);
    annLocal->addAnnotatedPolygon(polygon);

    // Create an 2nd annLocal with id = 1 and polygon and points too
    AnnotationLocalPtr annLocal2 = AnnotationLocal::create();
    AnnotatedPolygonPtr polygon2 = createPolygon();
    annLocal2->setName("annLocal"); polygon2->setId(1);
    foreach (PointPtr p, polygon2->getPoints()) {
        p->setId(1);
    }
    annLocal2->addAnnotatedPolygon(polygon2);

    PointPtr pAutre = Point::create();
    pAutre->setX(89), pAutre->setY(98);
    pAutre->setId(78);

    annLocal2->getAnnotatedPolygon(0)->addPoint(pAutre);

    QVERIFY (annLocal->getAnnotatedPolygon(0)->getPoints().size() != annLocal2->getAnnotatedPolygon(0)->getPoints().size());
    qDebug() << "S1 = " << annLocal->getAnnotatedPolygon(0)->getPoints().size() << "S2 =" <<  annLocal2->getAnnotatedPolygon(0)->getPoints().size();
    annLocal->merge(annLocal2);
    qDebug() << "S1 = " << annLocal->getAnnotatedPolygon(0)->getPoints().size() << "S2 =" <<  annLocal2->getAnnotatedPolygon(0)->getPoints().size();
    QVERIFY (annLocal->getAnnotatedPolygon(0)->getPoints().size() == annLocal2->getAnnotatedPolygon(0)->getPoints().size());

    // Check that all id are equal
    qDebug() << annLocal->getName() << annLocal2->getName();
    QVERIFY (annLocal->getName() == annLocal2->getName());
    QVERIFY (annLocal->getAnnotatedPolygons().size() == annLocal2->getAnnotatedPolygons().size());
    for(int i = 0; i < annLocal->getAnnotatedPolygons().size(); i++) {
        AnnotatedPolygonPtr polyAnnLocal = annLocal->getAnnotatedPolygon(i);
        AnnotatedPolygonPtr polyAnnLocal2 = annLocal2->getAnnotatedPolygon(i);

        QVERIFY (polyAnnLocal->getId() == polyAnnLocal2->getId());
        QVERIFY (polyAnnLocal->getPoints().size() == polyAnnLocal2->getPoints().size());
        qDebug() << "Points ";
        for(int j = 0; j < polygon->getPoints().size();j++) {
            qDebug() << "p1 " << polygon->getPoint(j)->getId() << "p2 " << polygon2->getPoint(j)->getId();
            QVERIFY (polygon->getPoint(j)->getId() == polygon2->getPoint(j)->getId());
        }
    }

}


void TestAnnotationLocal::testMerge_matchingPolygons()
{
    // Create an annLocal id = -1 and polygon id = -1;
    AnnotationLocalPtr annLocal = AnnotationLocal::create();
    annLocal->setName("annLocal");
    AnnotatedPolygonPtr polygon = createPolygon();
    annLocal->addAnnotatedPolygon(polygon);

    // Create an 2nd annLocal with id = 1 and polygon and points too
    AnnotationLocalPtr annLocal2 = AnnotationLocal::create();
    AnnotatedPolygonPtr polygon2 = createPolygon();
    annLocal2->setName("annLocal"); polygon2->setId(1);
    foreach (PointPtr p, polygon2->getPoints()) {
        p->setId(1);
    }
    annLocal2->addAnnotatedPolygon(polygon2);

    //check that all ids are different :
    {
        for(int i = 0; i < annLocal->getAnnotatedPolygons().size(); i++) {
            AnnotatedPolygonPtr polyAnnLocal = annLocal->getAnnotatedPolygon(i);
            AnnotatedPolygonPtr polyAnnLocal2 = annLocal2->getAnnotatedPolygon(i);
            QVERIFY (polyAnnLocal->getId() != polyAnnLocal2->getId());
            for(int j = 0; j < polygon->getPoints().size();j++)
                QVERIFY (polygon->getPoint(j)->getId() != polygon2->getPoint(j)->getId());
        }
    }

    // Merge
    annLocal->merge(annLocal2);

    // Check that all id are equal
    {
        QVERIFY(annLocal->getName() == annLocal2->getName());
        for(int i = 0; i < annLocal->getAnnotatedPolygons().size(); i++) {
            AnnotatedPolygonPtr polyAnnLocal = annLocal->getAnnotatedPolygon(i);
            AnnotatedPolygonPtr polyAnnLocal2 = annLocal2->getAnnotatedPolygon(i);

            QVERIFY (polyAnnLocal->getId() == polyAnnLocal2->getId());
            for(int j = 0; j < polygon->getPoints().size();j++) {
                QVERIFY (polygon->getPoint(j)->getId() == polygon2->getPoint(j)->getId());
            }
        }
    }
    //= Every polygon should be match with an id from annLocal2
    // So nothing should have an id of -1
    {
        for(int i = 0; i < annLocal->getAnnotatedPolygons().size(); i++) {
            AnnotatedPolygonPtr polyAnnLocal = annLocal->getAnnotatedPolygon(i);

            QVERIFY (polyAnnLocal->getId() != -1);
        }

    }
}

void TestAnnotationLocal::testMerge_mergePolygonsFromOtherAnnotation()
{
    // Create an annLocal id = -1 and polygon id = -1;
    // annLocal has only one polygon
    AnnotationLocalPtr annLocal = AnnotationLocal::create();
    AnnotatedPolygonPtr polygon = createPolygon();
    annLocal->addAnnotatedPolygon(polygon);

    // Create an 2nd annLocal with id = 1 and polygon and points too
    AnnotationLocalPtr annLocal2 = AnnotationLocal::create();
    AnnotatedPolygonPtr polygon2 = createPolygon();
    annLocal2->setName("annLocal"); polygon2->setId(1);
    foreach (PointPtr p, polygon2->getPoints()) {
        p->setId(1);
    }

    annLocal2->addAnnotatedPolygon(polygon2);

    // Both annotation have the same size
    QVERIFY (annLocal->getAnnotatedPolygons().size() == annLocal2->getAnnotatedPolygons().size());
    // witch is 1
    QVERIFY (annLocal->getAnnotatedPolygons().size() == 1);


    // Create a polygon with moved points
    AnnotatedPolygonPtr polygon3 = createPolygon();
    polygon3->getPoint(0)->setX(34);
    polygon3->getPoint(1)->setX(34);

    annLocal2->addAnnotatedPolygon(polygon3);
    // Size of annLocal + 1= annlocal2
    QVERIFY (annLocal->getAnnotatedPolygons().size() + 1 == annLocal2->getAnnotatedPolygons().size());
    // which is 2
    QVERIFY (annLocal2->getAnnotatedPolygons().size() == 2);
}

QTEST_MAIN(TestAnnotationLocal)

#if QT_VERSION >= 0x050000
  #include "testAnnotationLocal.moc"
#else
  #include "moc_testAnnotationLocal.cxx"
#endif

