#include <QDebug>
#include <QFile>
#include <QTextStream>
#include <DoQuBook/Page.hpp>
#include <QtTest>
#include <assert.h>
#include <QObject>
#include <DoQuBook/AnnotationLocal.hpp>
#include <Utils.hpp>
#include <DoQuBook/MarshallableFormater.hpp>


using namespace DoQuBook;

namespace Tests {

class TestPage : public QObject {

    Q_OBJECT

public :
    static const QString pageName;

    TestPage():_page(NULL){}

    ~TestPage()
    {
        if(!_page.isNull())
            _page.toWeakRef();
    }

private slots:

    void testGetSetName();

    void testAddLocalAnnotation();
    void testAddAnnotatedPolygon();

    void testEqual();
    void testMarshalling();


    void testPageService();
    void testPageAttachments();

    /**
         * Called before each test function is executed.
        */
    void init();

    /**
         * Called after every test function.
        */
    void cleanup();

protected :
    PagePtr _page;
};

}

using namespace Tests;

const QString TestPage::pageName = "page001";

void TestPage::init()
{
    _page = Page::create();
    assert(_page);
}

void TestPage::cleanup()
{
    if(!_page.isNull())
        _page.toWeakRef();
}

void TestPage::testGetSetName()
{
    _page->setName(pageName);
    QString name = _page->getName();

    QVERIFY(name == pageName);
}

void TestPage::testAddLocalAnnotation()
{
    PagePtr page = Page::create();
    AnnotationLocalPtr annLocal = createAnnotationLocal();
    page->addLocalAnnotation(annLocal);


    try {
        // Check if it as been ADDED !
        QVERIFY(!page->getAnnotationLocal(annLocal->getName()).isNull());

        // Check if the first polygons match
        AnnotatedPolygonPtr polyAdded = page->getAnnotationLocal(annLocal->getName())->getAnnotatedPolygon(0);
        AnnotatedPolygonPtr polyCreated = annLocal->getAnnotatedPolygon(0);
        QVERIFY(polyAdded->equals(*polyCreated));

        // Check if the Annotations Match
        AnnotationLocalPtr addedAnn = page->getAnnotationLocal(annLocal->getName());
        QVERIFY(addedAnn->equals(*annLocal));

    } catch (AnnotationNotFoundException e) {
        QVERIFY(false);
    } catch (AnnotationCastException e) {
        QVERIFY(false);
    }
}

void TestPage::testAddAnnotatedPolygon()
{
    //page->getAnnotationLocal("annLocal")->addAnnotatedPolygon(createPolygon());
    PagePtr page = Page::create();
    AnnotationLocalPtr annLocal = createAnnotationLocal();
    page->addLocalAnnotation(annLocal);

    PagePtr page2 = Page::create();
    page2->addAnnotatedPolygon("annLocal", createPolygon());

    QVERIFY(page->getAnnotations().size() ==  page2->getAnnotations().size());
    for(int i = 0; i < page->getAnnotations().size(); i++) {
        AnnotationLocalPtr l1 = page->getAnnotations().at(i).dynamicCast<AnnotationLocal>();
        AnnotationLocalPtr l2 = page2->getAnnotations().at(i).dynamicCast<AnnotationLocal>();
        QVERIFY (l1->getName() == l2->getName());
        QVERIFY (l1->getAnnotatedPolygons().size() == l2->getAnnotatedPolygons().size());
    }
    QVERIFY(page->equals(*page2));

    /// Adding the same Polygon must not add it
    page2->addAnnotatedPolygon("annLocal", createPolygon());
    QVERIFY(page->equals(*page2));

    /// if we changed the polygon value : They must be different !
    AnnotatedPolygonPtr poly = createPolygon();
    poly->setValue(poly->getValue().toInt()+1);
    page2->addAnnotatedPolygon ("annLocal", poly);
    QVERIFY (!page->equals(*page2));

    /// Add it to the original page with an other method (from AnnotationLocal)
    /// The two pages must be equal
    page->getAnnotationLocal("annLocal")->addAnnotatedPolygon(poly);
    QVERIFY (page->equals(*page2));
}

void TestPage::testEqual()
{
    PagePtr page1 = createPage();
    PagePtr page2 = createPage();

    QVERIFY(page1->equals(*page2));

    page2->setName("OtherName");
    QVERIFY(!page1->equals(*page2));
}


void TestPage::testMarshalling()
{

    PagePtr page = createPage();

    MarshallableFormaterJSon<Page> jsonFormater ;
    QString marshalled;
    jsonFormater.marshall(page, &marshalled);
    qDebug() << marshalled;

    MarshallableFormaterJSon<Page> jsonFormater2;
    PagePtr page2  = jsonFormater2.unMarshall(QByteArray(marshalled.toStdString().c_str()));

    QVERIFY(page->equals(*page2));
}

#include <DoQuBook/Repository.hpp>
#include <DoQuBook/ApplicationContext.hpp>

void TestPage::testPageService()
{
    RepositoryPtr repo = ApplicationContext::instance()->getRepository("LocalhostDev");
    PageServicePtr srv = repo->getPageService();
    PagePtr p = srv->get(268);

    int version = p->getVersion();

    p->addGlobalAnnotation("test", "test");
    srv->update(p);

    QCOMPARE(p->getVersion(), version+1);
}

void TestPage::testPageAttachments()
{
    RepositoryPtr repo = ApplicationContext::instance()->getRepository("digidoc");
    PageCollectionServicePtr collectionSrv = repo->getPageCollectionService();
    PageServicePtr pageSrv = repo->getPageService();

    qDebug() << "Retreiving all images ... please wait";
    QList<PagePtr> pages = collectionSrv->getPages("Syntheticimages-17Oct2012");
    foreach (PagePtr page, pages) {
        qDebug() << "Saving image of page " << page->getName();
        QImage img = page->getImage();
        img.save(page->getName()+".png");
        ListRestFileAttachmentPtr attachments = pageSrv->getAllAttachments(page);
        foreach(RestFileAttachmentPtr att, attachments->getAttachments()) {
            qDebug() << "Saving attachment =" << att->getFileName();
            att->save();
        }
    }


}



QTEST_MAIN(TestPage)



#include "moc_testPage.cxx"
