#include <QtTest>

#include <assert.h>
#include <QDebug>
#include <QObject>
#include <Utils.hpp>


#include <DoQuBook/Binarizer.hpp>
#include <DoQuBook/PluginLoader.hpp>
#include <DoQuBook/DoQuBookApp.hpp>
#include <DoQuBook/ApplicationContext.hpp>
#include <DoQuBook/RegionExtractor.hpp>

#include <DoQuBook/Region.hpp>

#include <DoQuBook/MarshallableFormater.hpp>

#include <DoQuBook/WebClient.hpp>

#include <DoQuBook/PluginLoader.hpp>

using namespace DoQuBook;

namespace Tests {

  class MeanIntensityRegionPointProcessor : public RegionPointProcessor {
    public :
      MeanIntensityRegionPointProcessor (const QImage &image)
        : _image (image), sum (0), nbPixel (0) {}

      virtual QString name() {return "mean_intenity";}


      virtual void processPoint (int x, int y)  {
        sum += qGray(_image.pixel(x,y));
        nbPixel++;
      }

      virtual void newRegion() {}
      virtual void endRegion() {}
      virtual void boundingBoxOfRegion (int /*x1*/, int /*y1*/, int /*x2*/, int /*y3*/) {}

      /// can be used as filters
      virtual void area (double /*a*/) {}
      virtual void perimeter (double /*p*/) {}

      float meanIntensity() {return sum/nbPixel;}

    private :
      const QImage &_image;
      int sum;
      int nbPixel;

  };

  class TestRegionProcessing : public QObject {

      Q_OBJECT
    public :

      TestRegionProcessing(){

        DoQuBook::RepositoryPtr repo = DoQuBook::ApplicationContext::instance()->getDefaultRepository();
        //PageCollectionServicePtr pageCollectionSrv = repo->getPageCollectionService();
        PageServicePtr pageSrv = repo->getPageService();
        PagePtr page = pageSrv->get(1);

        PluginLoader *loader = PluginLoader::instance();
        //BinarizerPtr binarizer = loader->getPlugin<Binarizer>("labri.OtsuBinarization.0.1.1");
        binarizer = loader->getPlugin<Binarizer>("labri.BinarizationOtsu2T.0.1.1");

        //regionExtractor = loader->getPlugin<RegionExtractor>("labri.RegionExtractorOpenCVForegroundPlugin.0.1.1");
        regionExtractor = loader->getPlugin<RegionExtractor>("labri.RegionExtractorISEWhiteCCPlugin.0.1.1");

        //img_damier = binarizer->binarize(page->getImage());
        img_damier = damier();
      }

      ~TestRegionProcessing(){}

    private slots:

      void init(){}

      void cleanup(){}

      void testSplitRegions() {
        // Segment all regions
        QBENCHMARK {
          regionExtractor->regions(img_damier);
        }
      }

      void testExtractRegions() {
        regionExtractor->regions(img_damier);

        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;

        QVariantMap params;
        params["color"] = QColor(255, 255, 255);

        QBENCHMARK {
          regionExtractor->runProcesors(processors2, params);
        }
      }

      void testPainterUsingRegionExtractor() {
        regionExtractor->regions(img_damier);
        QSharedPointer<RegionPainter> painter (new RegionPainter(img_damier));
        painter->setColor(qRgb(255, 0, 0));
        QList<RegionPointProcessorPtr> processors1; processors1 << painter;
        QVariantMap params;
        params["color"] = QColor(255, 255, 255);

        QBENCHMARK {
          regionExtractor->runProcesors(processors1, params);
        }
      }


      void testPainterOnExtractedRegions() {
        img_damier = damier();
        regionExtractor->regions(img_damier);

        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;

        QVariantMap params;
        params["color"] = QColor(255, 255, 255);


        regionExtractor->runProcesors(processors2, params);

        QList<RegionPtr> scanLines = scan->getRegions()->getRegions();
        if(regionExtractor->algorithmName().contains("ISE"))
          QCOMPARE(scanLines.size(), img_damier.width()*img_damier.height()/2);
        else
          QVERIFY(scanLines.size() == 1);

        QSharedPointer<RegionPainter> painter (new RegionPainter(img_damier));
        painter->setColor(qRgb(255,255,0));
        QBENCHMARK {
          foreach (const RegionPtr &r, scanLines) {
            r->iterateOnDomain(painter);
          }
        }

        img_damier.save("damier_ISE.png");
      }

      void testMeanIntenityProcessorOnOneExtractedRegion() {
        img_damier = damier();
        regionExtractor->regions(img_damier);

        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;

        QVariantMap params;
        params["color"] = QColor(255, 255, 255);


        regionExtractor->runProcesors(processors2, params);

        QList<RegionPtr> scanLines = scan->getRegions()->getRegions();
        QVERIFY(scanLines.size()>0);
        RegionPtr region = scanLines.at(0);

        QSharedPointer<MeanIntensityRegionPointProcessor> p (new MeanIntensityRegionPointProcessor (img_damier));
        QBENCHMARK {
          region->iterateOnDomain(p);
        }
        p->meanIntensity();

      }

      void testMeanIntenityWithoutProcessorOnOneExtractedRegion() {
        regionExtractor->regions(img_damier);

        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;

        QVariantMap params;
        params["color"] = QColor(255, 255, 255);


        regionExtractor->runProcesors(processors2, params);

        QList<RegionPtr> scanLines = scan->getRegions()->getRegions();
        RegionPtr region = scanLines.at(0);

        QList<QPointF> regionDomain = region->domain();
        int sum = 0;
        QBENCHMARK {
          foreach (const QPointF &point, regionDomain) {
            sum+= qGray(img_damier.pixel(point.x(), point.y()));
          }
        }
        /*int mean  = */sum/regionDomain.size();
      }

      // Every regions is processed in a diffrent thread
      void testMappedRegionsProcessor() {
        regionExtractor->regions(img_damier);

        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;

        QVariantMap params;
        params["color"] = QColor(255, 255, 255);


        regionExtractor->runProcesors(processors2, params);

        QList<RegionPtr> scanLines = scan->getRegions()->getRegions();

        MappedRegionsProcessor regionsProcessor (scanLines);
        QSharedPointer<RegionPainter> painter (new RegionPainter(img_damier));
        painter->setColor(qRgb(255,255,0));
        QBENCHMARK {
          regionsProcessor.runProcessors(painter);
        }
      }

      // Every processor runs on a Thread
      void testSeveralProcessorsMappedRegionsProcessors() {
        regionExtractor->regions(img_damier);
        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;
        QVariantMap params;
        params["color"] = QColor(255, 255, 255);
        regionExtractor->runProcesors(processors2, params);
        QList<RegionPtr> scanLines = scan->getRegions()->getRegions();


        MappedRegionsProcessors regionsProcessors (scanLines);
        QSharedPointer<RegionPainter> painter1 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter2 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter3 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter4 (new RegionPainter(img_damier));
        QList<RegionPointProcessorPtr> processors;
        processors << painter1 << painter2 << painter3 << painter4;

        QBENCHMARK {
          regionsProcessors.runProcessors(processors);
        }
      }

      // Every processor is run sequentially
      void testSeveralProcessorsLoopedProcessors() {
        regionExtractor->regions(img_damier);
        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;
        QVariantMap params;
        params["color"] = QColor(255, 255, 255);
        regionExtractor->runProcesors(processors2, params);
        QList<RegionPtr> scanLines = scan->getRegions()->getRegions();


        QSharedPointer<RegionPainter> painter1 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter2 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter3 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter4 (new RegionPainter(img_damier));
        QList<RegionPointProcessorPtr> processors;
        processors << painter1 << painter2 << painter3 << painter4;

        QBENCHMARK {
          foreach (const RegionPointProcessorPtr &proc, processors) {
            foreach (const RegionPtr &r, scanLines) {
              r->iterateOnDomain(proc);
            }
          }
        }
      }

      void testSeveralProcessorsOnSegmentation() {
        regionExtractor->regions(img_damier);

        QSharedPointer<RegionPainter> painter1 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter2 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter3 (new RegionPainter(img_damier));
        QSharedPointer<RegionPainter> painter4 (new RegionPainter(img_damier));
        QList<RegionPointProcessorPtr> processors;
        processors << painter1 << painter2 << painter3 << painter4;

        QVariantMap params;
        params["color"] = QColor(255, 255, 255);

        QBENCHMARK {
          regionExtractor->runProcesors(processors, params);
        }

      }

      void testSave() {
        img_damier.save("sav.png");
      }

      ///////////////////////////////////////////////////////////////////////////
      ///////////////////////////////////////////////////////////////////////////
      /// Serialization
      ///

      void testSerialize() {
        // UNCOMMENT WHEN COMPRESSION=ON
       /* regionExtractor->regions(img_damier);

        QSharedPointer<RegionsExtractor> scan (new RegionsExtractor);
        QList<RegionPointProcessorPtr> processors2; processors2 << scan;

        QVariantMap params;
        params["color"] = QColor(255, 255, 255);


        regionExtractor->runProcesors(processors2, params);

        QList<RegionPtr> scanLines = scan->getRegions()->getRegions();
        RegionListPtr regions = RegionList::create();
        regions->setRegions(scanLines);

        MarshallableFormaterPtr jsonFormater = PluginLoader::instance()->getPlugin<MarshallableFormater>("labri.QJSONMarshallableFormater.0.0.1");
        QString json;
        jsonFormater->marshall(regions, &json);

        WebClient w;
        QByteArray compressedJson = w.gCompress(json.toUtf8());

        QFile file("regions.json");
        file.open(QIODevice::WriteOnly | QIODevice::Text);
        QTextStream out(&file);
        out << json;
        file.close();

        QFile file2("regions.json.zip");
        file2.open(QIODevice::WriteOnly | QIODevice::Text);
        QTextStream out2(&file2);
        out2 << compressedJson;
        file2.close();

        img_damier.save("damier.png");
        */
      }

    private :
      QImage damier() {
        QImage img(100, 100, QImage::Format_RGB888);
        int nbWhitePixels = 0;
        bool lastWhite = true;
        for(int i = 0; i< img.width(); i++) {
          for(int j = 0; j< img.height(); j++) {
            if(lastWhite)
              img.setPixel(i,j, qRgb(0, 0, 0));
            else {
              img.setPixel(i,j, qRgb(255, 255, 255));
              nbWhitePixels++;
            }

            lastWhite = !lastWhite;
          }
          lastWhite = !lastWhite;
        }
        return img;
      }
    private :
      BinarizerPtr binarizer;
      RegionExtractorPtr regionExtractor;

      QImage img_damier;

  };

}
using namespace Tests;


QTEST_MAIN(TestRegionProcessing)

#if QT_VERSION >= 0x050000
#include "testRegionProcessing.moc"
#else
#include "moc_testRegionProcessing.cxx"
#endif

