#include "IseCooccurrencePixelFeaturesComputer.hpp"

#include <cv.h>
#include <highgui.h>


#include <DoQuBook/DoQuBookOpenCVConverter.hpp>

#include <ise/iseSegmentationContext.hpp>
#include <ise/iseSegmentationClustering.hpp>
#include <ise/iseSplitAlgorithm.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>
#include <ise/iseQuantificationThresholderOtsu2T.hpp>
#include <ise/iseHistogram1D.hpp>
#include <ise/iseSegmentationClusteringLocalThresholds.hpp>
#include <ise/iseSegmentationClusteringLocalThresholderSauvola.hpp>



#include <ise/iseGrayLevelCooccurrenceMatrix.hpp>

#include <girl/girlCommon.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>

#include <QtPlugin>
#include <QColor>

using namespace DoQuBook;


class ComputeGLCMOfRegion
{
public:
  ComputeGLCMOfRegion
  (const RegionFeaturesPtr& feats,
   const girl::region &id,
   OrientationGLCM orientation,
   ise::GrayLevelCooccurrenceMatrix &featureMat)
    : _feats (feats)
    , _id(id)
    , _orientatio (orientation)
    , _featureMat (featureMat)
  {}

  void compute() {
    _featureMat.compute(_id, _orientatio);
    _energy = _featureMat.energy();
    _entropy = _featureMat.entropy();
    _mean = _featureMat.mean();
    _stdd = _featureMat.stdd();
    _sumEntropy = _featureMat.sumEntropy();
    _diffEntropy = _featureMat.diffEntropy();

    QList<double> regionFeats;
    regionFeats << _energy << _entropy << _mean
               << _stdd << _sumEntropy << _diffEntropy;

    _feats->setFeatures(regionFeats);
  }


protected:
  const RegionFeaturesPtr& _feats;
  const girl::region &_id;
  OrientationGLCM _orientatio;
  ise::GrayLevelCooccurrenceMatrix &_featureMat;

protected :
  double _energy;
  double _entropy;
  double _mean;
  double _stdd;
  double _sumEntropy;
  double _diffEntropy;
};


class ToDoQuBookRegionConverter : public girl::toger::pointIterator
{
public:
    ToDoQuBookRegionConverter(DoQuBook::ScanLineExtractorPtr &processor)
        : _processor (processor)
    {}

    virtual void processPoint(const girl::point &pt)
    {
        _processor->processPoint(pt.x(), pt.y());
    }

protected:

    DoQuBook::ScanLineExtractorPtr &_processor;
};

class SegClustSameColorPixel : public ise::SegmentationClustering
{

public :
    SegClustSameColorPixel(ise::SegmentationContext *segContext
                                         ,QImage &image)
        : ise::SegmentationClustering (segContext)
        , m_image (image)
    {
    }

    virtual void init(const girl::regionDeque &/*regions*/, int /*level*/=1) {}

    virtual void init(girl::region , int ) {}


    virtual bool same(const girl::point &p, girl::direction dir) {
        girl::point p2 = p;
        p2.move(dir);

        int pixelP1 = qGray(m_image.pixel(p.x(),p.y()));
        int pixelP2 = qGray(m_image.pixel(p2.x(),p2.y()));

        return pixelP1 == pixelP2;
    }

protected :
    QImage &m_image;
};


QList<DoQuBook::ExecutorParameterPtr>  IseCooccurrencePixelFeaturesComputer::parameters ()
{
  QList<DoQuBook::ExecutorParameterPtr> params;
  //params << DoQuBook::ExecutorParameter::create("thd", "Double", true);
  return params;
}

DoQuBook::RegionFeaturesListPtr IseCooccurrencePixelFeaturesComputer::extract (const QImage &qimage, RegionListPtr& regions, const QVariantMap & /*parameters*/)
{
  DoQuBookOpenCVConverter converter;
  cv::Mat src = converter.getCvMat(qimage);


  girl::tools::Image image (src);

  girl::toger tog(image.width(), image.height());


  ise::SegmentationContext segContext;
  segContext.setToger(&tog);
  segContext.setImage(&image);


  // Convert DoQuBookRegions to girlRegions using an image
  QImage regionsImage(qimage.size(), QImage::Format_RGB32);
  regionsImage.fill(QColor(0,0,0));
  QSharedPointer<RegionPointProcessor> painter (new RegionPainter(regionsImage, qRgb(255,255,255)));
  QList<RegionPtr> rs = regions->getRegions();
  MappedRegionsProcessor mrp (rs);
  mrp.runProcessors(painter);

  // get the girl::regionDeque
  girl::regionDeque toBeSplittedRegions;
  /// On choisis la premiere region à spliter
  toBeSplittedRegions.push_back(1);

  ise::SplitAlgorithmDefault splitAlgorithm(&segContext);
  ise::SplitDriverDefault defaultDriver(&segContext);

  girl::regionDeque createdRegions;

  SegClustSameColorPixel segCluster (&segContext, regionsImage);
  splitAlgorithm.apply(toBeSplittedRegions, defaultDriver, segCluster , 0, createdRegions);


  ise::GrayLevelCooccurrenceMatrix featureMat (&segContext);

  RegionFeaturesListPtr regionFeatsList = RegionFeaturesList::create();
  foreach (const girl::region &r, createdRegions) {

      girl::point p = tog.anyPoint(r);
      QRgb pixel = regionsImage.pixel(p.x(), p.y());
      if(pixel == qRgb(255,255,255)) {
          // Compute features on r
          DoQuBook::RegionFeaturesPtr  feats = RegionFeatures::create();
          ComputeGLCMOfRegion computer(feats, r, D0, featureMat);
          computer.compute();

          // Extract the DoQuBook region
          ScanLineExtractorPtr regionExtractor (new RegionsExtractor);
          ToDoQuBookRegionConverter converter(regionExtractor);
          regionExtractor->newRegion();
          tog.iterateOnRegionDomain(r, converter);
          regionExtractor->endRegion();

          // There is only one region
          RegionPtr region = regionExtractor->getRegions()->getRegions().at(0);
          feats->setRegion(region);

          regionFeatsList->add(feats);
      }
  }

  return regionFeatsList;

}




DQB_EXPORT_PLUGIN(IseCooccurrencePixelFeaturesComputer)
