#include "ClassBuilder.hpp"

#include <QDebug>

ClassBuilder::ClassBuilder(Entity &e, QString lang, QString declarationExtention, QString implementationExtention)
    : GenericBuilder(e, lang, declarationExtention, implementationExtention)
{
}



QString ClassBuilder::getAnnotationsTemplate()
{
    QString templ;
    QFile file (":/" + lang() + "/entity/annotations");
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qCritical() << lang() << ": Could not open file :/entity/annotations";
        return templ;
    }

    QTextStream in(&file);
    templ.append(in.readAll());
    return templ.trimmed();
}

QString ClassBuilder::getConstructorTemplate()
{
    QString templ = openTemplate(":/"+lang()+"/constructor");
    templ.replace("${CONSTRUCTOR_DECLARATION}", getConstructorDeclarationTemplate());
    templ.replace("${CONSTRUCTOR_IMPLEMENTATION}", getConstructorImplementationTemplate());

    QString inits;
    foreach (Field f, entity().fields()) {
        if(f.isList()) {
            QString listInitTemplate = getInitArrayListTemplate();
            listInitTemplate.replace("${NAME}", f.name());
            listInitTemplate.replace("${TYPE}", f.type());
            inits.append(listInitTemplate);
        }
    }
    templ.replace("${LIST_INIT}", inits);

    return templ;
}

QString ClassBuilder::getInitArrayListTemplate()
{
    QString templ;
    QFile file (":/" + lang() + "/init_arraylist");
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qCritical() << lang() << ": Could not open file :/init_arraylist";
        return templ;
    }

    QTextStream in(&file);
    templ.append(in.readAll());
    return templ.trimmed();
}

QString ClassBuilder::getFieldDeclarationTemplate()
{
    return openTemplate(":/"+lang()+"/basicfield_declaration");
}

QString ClassBuilder::getBasicFieldAnnotation()
{
    return openTemplate(":/"+lang()+"/basicfield_annotation").trimmed();
}

QString ClassBuilder::getIdFieldAnnotation()
{
    return openTemplate(":/"+lang()+"/idfield_annotation").trimmed();
}

QString ClassBuilder::getIntIdFieldAnnotation()
{
    return openTemplate(":/"+lang()+"/intidfield_annotation").trimmed();
}

/**
  * Setup the annotations attached to a field of type List
  * by reading in the "listfield_annotations.txt" file
  */
QString ClassBuilder::getListFieldAnnotations()
{
    return openTemplate(":/"+lang()+"/listfield_annotations").trimmed();
}

QString ClassBuilder::getGetterTemplate()
{
    return openTemplate(":/"+lang()+"/getter");
}

QString ClassBuilder::getSetterTemplate()
{
    return openTemplate(":/"+lang()+"/setter");
}

QString ClassBuilder::getGetterDeclarationTemplate()
{
    return openTemplate(":/"+lang()+"/getter_declaration").trimmed();
}

QString ClassBuilder::getGetterImplementationTemplate()
{
    return openTemplate(":/"+lang()+"/getter_implementation");
}

QString ClassBuilder::getSetterDeclarationTemplate()
{
    return openTemplate(":/"+lang()+"/setter_declaration").trimmed();
}


QString ClassBuilder::getSetterImplementationTemplate()
{
    return openTemplate(":/"+lang()+"/setter_implementation");
}


// retrieve the Constructor Implementation Template
QString ClassBuilder::getConstructorImplementationTemplate()
{
    return openTemplate(":/"+lang()+"/constructor_implementation");
}

// retrieve the Constructor Declaration Template
QString ClassBuilder::getConstructorDeclarationTemplate()
{
    return openTemplate(":/"+lang()+"/constructor_declaration").trimmed();
}

QString ClassBuilder::getImplementationFileTemplate()
{
    return openTemplate(":/"+lang()+"/entity/implementation_file").trimmed();
}

QString ClassBuilder::getFieldType(const Field &f)
{
    if(f.isList())
        return "List<"+f.type()+">";
    else if (f.isMap())
        return "Map<"+f.key()+","+f.type()+">";
    else
        return f.type();
}
/**
  * Build annotations attached to a field
  * Setup differents annotations according to the kind (list or single)
  * Id field is a special one
  */
QString ClassBuilder::getFieldAnnotations(const Field &f)
{
    if(f.isList()) {
        return getListFieldAnnotations();
    } else {
        if(f.isId()) {
            if(f.type() == "int")
                return getIntIdFieldAnnotation();
            else
                return getIdFieldAnnotation();
        } else {
            return getBasicFieldAnnotation();
        }
    }
}

/**
  * Build fields declarations
  * Set up the type, name and annotations attached to each field
  */
QString ClassBuilder::fieldDeclaration(const Field &f)
{
    QString type = getFieldType(f);
    QString fieldDeclarationTempl = getFieldDeclarationTemplate();
    QString fieldAnnotations = getFieldAnnotations(f);

    fieldDeclarationTempl.replace("${FIELD_ANNOTATIONS}", fieldAnnotations);
    fieldDeclarationTempl.replace("${TYPE}", type);
    fieldDeclarationTempl.replace("${NAME}", f.name());
    QString nameUp = f.name().at(0).toUpper()+f.name().mid(1);
    fieldDeclarationTempl.replace("${NAME_FIRST_LETTER_UP}", nameUp);
    return fieldDeclarationTempl;
}

QString ClassBuilder::fieldDeclarationGettersAndSetters(const Field &f)
{
    QString getAndSet;
    QString getterTempl = getGetterTemplate();
    getterTempl.replace("${GETTER_DECLARATION}", getGetterDeclarationTemplate());
    getterTempl.replace("${GETTER_IMPLEMENTATION}", getGetterImplementationTemplate());

    getterTempl.replace("${TYPE}", getFieldType(f));

    getterTempl.replace("${NAME}", f.name());

    QString setterTempl = getSetterTemplate();
    setterTempl.replace("${SETTER_DECLARATION}", getSetterDeclarationTemplate());
    setterTempl.replace("${SETTER_IMPLEMENTATION}", getSetterImplementationTemplate());

    setterTempl.replace("${TYPE}", getFieldType(f));

    setterTempl.replace("${NAME}", f.name());
    getAndSet.append(getterTempl);
    getAndSet.append(setterTempl);
    QString nameUp = f.name().at(0).toUpper()+f.name().mid(1);
    getAndSet.replace("${NAME_FIRST_LETTER_UP}", nameUp);
    return getAndSet;
}

QString ClassBuilder::fieldImplementationGettersAndSetters(const Field &f)
{
    QString getAndSet;
    QString getterImplTempl = getGetterImplementationTemplate();
    getterImplTempl.replace("${TYPE}", getFieldType(f));
    getterImplTempl.replace("${NAME}", f.name());
    QString nameUp = f.name().at(0).toUpper()+f.name().mid(1);
    getterImplTempl.replace("${NAME_FIRST_LETTER_UP}", nameUp);

    QString setterImplTempl = getSetterImplementationTemplate();
    setterImplTempl.replace("${TYPE}", getFieldType(f));
    setterImplTempl.replace("${NAME}", f.name());
    setterImplTempl.replace("${NAME_FIRST_LETTER_UP}", nameUp);

    getAndSet.append(getterImplTempl);
    getAndSet.append(setterImplTempl);
    return getAndSet;
}

/**
  * Build and write the implementation file
  */
void ClassBuilder::writeImplementationFile()
{
    // Open the implementation file
    QFile * file = implementationFile();
    if (!file->open(QIODevice::WriteOnly | QIODevice::Text))
        return;

    // get a stream to the file
    QTextStream out2(file);

    // retrieve the classe implementation template
    QString implTemplate = getImplementationFileTemplate();

    // Set the differents getters/setters
    QString gettersAndsettersImpl;
    foreach (Field f, entity().fields()) {
        gettersAndsettersImpl.append(fieldImplementationGettersAndSetters(f));
    }

    // set the classe name
    gettersAndsettersImpl.replace("${CLASS_NAME}", entity().name());

    // set the etters and setters section
    implTemplate.replace("${GETTERS_AND_SETTERS}", gettersAndsettersImpl);

    // retrieve and set the Ctor template
    QString constructorImpl = getConstructorImplementationTemplate();
    implTemplate.replace("${CONSTRUCTOR}", constructorImpl);
    implTemplate.replace("${CLASS_NAME}", entity().name());

    // Do more replacements if needed
    moreTemplatesOnImplementationFile(implTemplate);

    // push the stream content to the file
    out2 << implTemplate;

    if(file) delete file; file = NULL;
}

void ClassBuilder::writeDeclarationFile()
{
    // create and retrieve the declaration file
    QFile * file = declarationFile();

    // error checkin'
    if (!file->open(QIODevice::WriteOnly | QIODevice::Text))
        return;

    // stream to the outfile
    QTextStream out(file);

    // retrieve the class declaration template
    QString templ = getTemplate();

    // set inheritence
    setInheritence(templ);

    // replace values according to each specific lgg
    templ.replace("${IMPORTS}",     imports());                 // imports
    templ.replace("${ANNOTATIONS}", getAnnotationsTemplate());  // classe annotations
    templ.replace("${CONSTRUCTOR}", getConstructorTemplate());  // ctor

    templ.replace("${CLASS_NAME}",          entity().name());    // classe name
    templ.replace("${CLASS_NAME_UPPER}",    classNameUpper());
    templ.replace("${CLASS_NAME_LOWER}",    classNameLower());

    // set up field declarations
    QString fields;
    foreach (Field f, entity().fields()) {
        fields.append(fieldDeclaration(f));
    }
    templ.replace("${FIELDS}", fields);

    // set up setters/getters
    QString getsAndSets;
    foreach (Field f, entity().fields()) {
           getsAndSets.append(fieldDeclarationGettersAndSetters(f));
    }
    templ.replace("${GETTERS_AND_SETTERS}", getsAndSets);

    // more replacement if specificly needed
    moreTemplatesOnDeclarationFile(templ);

    // pushing the stream to the file
    out << templ;

    if(file) delete file; file = NULL;
}

void ClassBuilder::setInheritence(QString & templ)
{
    // only if a parent class has been setup
    if( ! entity().parent().isEmpty() ) {
        templ.replace("${INHERIT_KEYWORD}", getInheritKeyword());  // mot clef héritage
        templ.replace("${INHERIT_PARENT}",  getInheritParent() );  // parent
    }
    else {
        templ.replace("${INHERIT_KEYWORD}", getInheritDefaultKeyword()); // mot clef héritage
        templ.replace("${INHERIT_PARENT}",  getInheritDefaultParent());  // parent
    }
}

QString ClassBuilder::getInheritKeyword()
{
    return ": ";
}

QString ClassBuilder::getInheritParent()
{
    return entity().parent();
}

QString ClassBuilder::getInheritDefaultKeyword()
{
    return "";
}

QString ClassBuilder::getInheritDefaultParent()
{
    return "";
}

QString ClassBuilder::getDeclarationFileTemplate()
{
    return ":/" + lang() + "/entity/class";
}


QString ClassBuilder::getterTemplate()
{
    return ":/" + lang() + "/getter";
}

QString ClassBuilder::declarationFileSuffix()
{
    QString f ( entity().name() + langExtDecl() );
    return f;
}

QString ClassBuilder::implementationFileSuffix()
{
    QString f ( entity().name() + langExtImpl() );
    return f;
}

QString ClassBuilder::getImportTemplate()
{
    return ":/" + lang() + "/imports";
}

QString ClassBuilder::imports()
{
    QString import;
    QFile file(getImportTemplate());
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qCritical() << lang() << ": Could not open file " << getImportTemplate();
        return import;
    }

    QTextStream in(&file);
    while (!in.atEnd()) {
        import.append(in.readLine()+"\n");
    }
    return import;
}

