#include <DoQuBook/PluginLoader.hpp>
#include <DoQuBook/Book.hpp>
#include <DoQuBook/DoQuBookApp.hpp>
#include <DoQuBook/RectoVersoRegistration.hpp>
#include <QPainter>
#include <DoQuBook/AttachmentServices.hpp>
#include <DoQuBook/ApplicationContext.hpp>
#include <DoQuBook/Binarizer.hpp>
#include <QDebug>

#include <QtGui>
#include <QtXml/QtXml>

#include "JavaClassBuilder.hpp"
#include "CppClassBuilder.hpp"
#include "CSharpClassBuilder.hpp"
#include "JavaInterfaceServiceBuilder.hpp"
#include "JavaImplServiceBuilder.hpp"
#include "JavaResourceBuilder.hpp"
#include "CppServiceBuilder.hpp"
#include "CSharpServiceBuilder.hpp"

using namespace DoQuBook;

QString orgName ("labri");

/// Must be the name of the target in CMake
QString appName ("DIGIDOCTest");


/**
  * ./XMLRESTGen.app/Contents/MacOS/XMLRESTGen  -xml ../Personne.xml
  */
class XMLRESTGen : public DoQuBook::DoQuBookApp
{
public :
    static const int HEIGHT = 600;
    static const int WIDTH = 800;
    static const int NB_IMAGE = 3;

    XMLRESTGen(QCoreApplication &app)
        : DoQuBook::DoQuBookApp(app, orgName, appName)

    {}



protected :

    // Lists different program parameters
    virtual QList<DoQuBookApp::ProgramParameter> paramsAndDescriptions()
    {
        QList<DoQuBookApp::ProgramParameter> paramsAndDesc;
        paramsAndDesc.push_back(ProgramParameter("xml",false, "XML file (ex : Personne.xml)"));
        paramsAndDesc.push_back(ProgramParameter("rep",false, "XML files in a rep (ex : myRep/*.xml)"));
        return paramsAndDesc;
    }

    // The main's run
    virtual void run(QMap<QString, QString> args)
    {
        // if the "-xml" argument is used
        if( args.find("xml") != args.end() )
            ParseXMLCmd(args);

        // find the "-rep" argument is used
        if( args.find("rep") != args.end() )
            ParseRepCmd(args);

    }

    // if an arg is -xml
    void ParseXMLCmd(QMap<QString, QString> args)
    {
        QString xmlfile = args["xml"];
        WriteEntityFiles(getDomDocument(xmlfile));
    }

    // if an arg is -rep
    void ParseRepCmd(QMap<QString, QString> args)
    {
        // get the pointed rep
        QString xmlDir = QDir::currentPath () + QDir::separator () + args["rep"];

        // set filters to xml files only
        QStringList filter( QStringList() << "*.xml" );

        // fileiterator
        QDirIterator dirIterator(xmlDir, filter ,QDir::Files);

        //
        while(dirIterator.hasNext())
        {
            QString xmlFile = dirIterator.next();
            qDebug() << "Generate from " << xmlFile;
            WriteEntityFiles( getDomDocument( xmlFile ) );
        }
    }

    QDomDocument getDomDocument(QString filename)
    {
        QDomDocument doc("mydocument");
        QFile file( filename );

        if (!file.open(QIODevice::ReadOnly)) {
            qCritical() << "could not openFile";
            throw new std::exception();
        }

        if (!doc.setContent(&file))
        {
            file.close();
            qCritical() << "Error on XML";
            throw new std::exception();
        }
        file.close();
        return doc;
    }

    void WriteEntityFiles(QDomDocument doc)
    {
        Entity e;
        // entities builder declarations - pass it the Entity ref
        JavaClassBuilder javabuilder (e);
        CppClassBuilder cppbuilder (e);
        CSharpClassBuilder csharpbuilder (e);

        // xml document node
        QDomElement docElem = doc.documentElement();

        // entity node
        QDomElement entity = docElem.elementsByTagName("entity").at(0).toElement();

        // entity name
        QString entityName = entity.attributes().namedItem("name").nodeValue();
        e.setName(entityName);

        // entity parent
        QString entityParent = entity.attributes().namedItem("extends").nodeValue();
        e.extends(entityParent);

        // simple fields parsing
        QDomNodeList fields = entity.elementsByTagName("field");
        for(int i = 0; i < fields.count(); i++) {

            // field type
            QString type = fields.at(i).attributes().namedItem("type").nodeValue();

            // field name
            QString name = fields.at(i).attributes().namedItem("name").nodeValue();

            // set params and attach field to the Entity Obj
            bool isId = fields.at(i).attributes().namedItem("id").nodeValue() == "true";
            Field f (type, name);
            f.setId(isId);      // this field is an Id one
            e.addField(f);
        }

        // list fields parsing
        QDomNodeList listfields = entity.elementsByTagName("list");

        // for each list found
        for(int i = 0; i < listfields.count(); i++) {

            // list type
            QString type = listfields.at(i).attributes().namedItem("type").nodeValue();

            // list name
            QString name = listfields.at(i).attributes().namedItem("name").nodeValue();

            // setup and attach the list field to the Entity obj
            bool isMarshallable = listfields.at(i).attributes().namedItem("marshallable").nodeValue() == "true";
            Field f (type, name);
            f.setIsList(true);                      // this field is a List
            f.setIsMarshallable(isMarshallable);    // this field is serializable
            e.addField(f);
        }

        // map fields parsing
        QDomNodeList mapfields = entity.elementsByTagName("map");

        // for each map found
        for(int i = 0; i < mapfields.count(); i++) {

            // map key type
            QString key = mapfields.at(i).attributes().namedItem("key").nodeValue();

            // map value type
            QString value = mapfields.at(i).attributes().namedItem("value").nodeValue();

            // map name
            QString name = mapfields.at(i).attributes().namedItem("name").nodeValue();

            // setup and attach the list field to the Entity obj
            bool isMarshallable = mapfields.at(i).attributes().namedItem("marshallable").nodeValue() == "true";
            Field f (key, value, name);
            f.setIsMap(true);                       // this field is a Map
            f.setIsMarshallable(isMarshallable);    // this field is serializable
            e.addField(f);
        }

        javabuilder.toFile();
        cppbuilder.toFile();
        csharpbuilder.toFile();

        qDebug() << "Entity created";




        // services builders
        JavaInterfaceServiceBuilder jInterfServiceBuilder(e);
        JavaImplServiceBuilder jServiceImplBuilder(e);
        JavaResourceBuilder jResourceBuilder (e);
        CppServiceBuilder cppServiceBuilder (e);
        CSharpServiceBuilder csharpServiceBuilder (e);

        // service node
        QDomElement service = docElem.elementsByTagName("service").at(0).toElement();

        // service/entity names
        QString name = service.attributes().namedItem("name").nodeValue();
        QString entityname = service.attributes().namedItem("entityname").nodeValue();

        // parse the CRUD method needs
        QDomElement crudmethods = service.elementsByTagName("crudmethods").at(0).toElement();
        bool create = crudmethods.attributes().namedItem("create").nodeValue() == "true";
        bool read = crudmethods.attributes().namedItem("read").nodeValue() == "true";
        bool update = crudmethods.attributes().namedItem("update").nodeValue() == "true";
        bool del = crudmethods.attributes().namedItem("delete").nodeValue() == "true";
        CRUDMethods crudmeths (create, read, update, del);

        // pass them to the service builders
        jInterfServiceBuilder.setNamesAndCrudMethods(name, entityname, crudmeths);
        jServiceImplBuilder.setNamesAndCrudMethods(name, entityname, crudmeths);
        jResourceBuilder.setNamesAndCrudMethods(name, entityname, crudmeths);
        cppServiceBuilder.setNamesAndCrudMethods(name, entityname, crudmeths);
        csharpServiceBuilder.setNamesAndCrudMethods(name, entityname, crudmeths);

        //file creations
        jInterfServiceBuilder.toFile();
        jServiceImplBuilder.toFile();
        jResourceBuilder.toFile();
        cppServiceBuilder.toFile();
        csharpServiceBuilder.toFile();

        qDebug() << "Entity Service created";
    }
};


int main(int argc, char *argv[])
{
    QCoreApplication app(argc, argv);

    XMLRESTGen test (app);

    return test.exec();
}
